/*
 * ChessFileChooser.java --- Custom file chooser class.
 *
 * Copyright (c) 2001, 2002, 2003 by Pascal Wassong All Rights Reserved.
 *
 * This file is part of iNatch.
 *
 * iNatch is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * iNatch is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

package iNatch;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Graphics2D;
import java.awt.Graphics;
import java.awt.Image;
import java.awt.Toolkit;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.io.Reader;
import java.util.MissingResourceException;
import javax.swing.Box;
import javax.swing.JFileChooser;
import javax.swing.JPanel;
import javax.swing.UIManager;

import wassong.gui.tools.Resources;

public class ChessFileChooser extends JFileChooser {

    // --- Forsythe previewer --------------------------------------------------

    class ForsythePreviewer extends JPanel {

	public ForsythePreviewer( Resources resources ) {

	    if ( images[ 0 ] == null ) {
		Toolkit	defaultToolkit = Toolkit.getDefaultToolkit();
		Class	c              = this.getClass();
		images[  0 ] = defaultToolkit.getImage(
		    c.getResource( "resources/wKing16.gif" ) );
		images[  1 ] = defaultToolkit.getImage(
		    c.getResource( "resources/wPawn16.gif" ) );
		images[  2 ] = defaultToolkit.getImage(
		    c.getResource( "resources/wQueen16.gif" ) );
		images[  3 ] = defaultToolkit.getImage(
		    c.getResource( "resources/wRook16.gif" ) );
		images[  4 ] = defaultToolkit.getImage(
		    c.getResource( "resources/wBishop16.gif" ) );
		images[  5 ] = defaultToolkit.getImage(
		    c.getResource( "resources/wKnight16.gif" ) );
		images[  6 ] = defaultToolkit.getImage(
		    c.getResource( "resources/bKing16.gif" ) );
		images[  7 ] = defaultToolkit.getImage(
		    c.getResource( "resources/bPawn16.gif" ) );
		images[  8 ] = defaultToolkit.getImage(
		    c.getResource( "resources/bQueen16.gif" ) );
		images[  9 ] = defaultToolkit.getImage(
		    c.getResource( "resources/bRook16.gif" ) );
		images[ 10 ] = defaultToolkit.getImage(
		    c.getResource( "resources/bBishop16.gif" ) );
		images[ 11 ] = defaultToolkit.getImage(
		    c.getResource( "resources/bKnight16.gif" ) );
		setLocalisedResources( resources );
	    }

	    Dimension d = new Dimension( squareSize * 8, squareSize * 8 );
	    this.setPreferredSize( d );
	    this.setMinimumSize  ( d );
	    //	    this.setBorder( BorderFactory.createEtchedBorder() );

	    emptyDiagramTooltip = resources.getString( "Open-no-position" );

	    this.setToolTipText( emptyDiagramTooltip );

	}

	public void initPosition( File f ) {

	    if ( f != null ) {
		try {
		    Reader		r  = new FileReader( f );
		    BufferedReader	in = new BufferedReader( r );
		    String		s;

		    int		counter = 50;
		    while ( counter > 0 && ( s = in.readLine() ) != null ) {

			counter--;
			s = s.trim();
			PositionFactory.make( new StandardBoard(),
					      new OrthodoxCondition(),
					      s );
			if ( PositionFactory.status()
			     == PositionFactory.CORRECT_POSITION ) {

			    position = PositionFactory.position();
			    positionLoaded = true;
			    this.setToolTipText( position.toString() );
			    this.repaint();
			    return;
			}
		    }
		} catch ( FileNotFoundException fnfe ) {
		    // Nothing to do
		} catch ( IOException ioe ) {
		    // Nothing to do
		}
	    }
	    if ( positionLoaded ) {
		positionLoaded = false;
		this.setToolTipText( emptyDiagramTooltip );
		this.repaint();
	    }
	}

	public void paintComponent( Graphics g ) {

	    super.paintComponent( g );

	    Graphics2D	g2d = (Graphics2D)g;
	    Color	blackSquareColor = new Color( 192, 192, 192 );

	    int[][]	chessboard = null;
	    if ( positionLoaded ) {
		chessboard = position.chessboard();
	    }

	    for ( int j = 0; j < 8; j++ ) {
		for ( int i = 0; i < 8; i++ ) {
		    boolean	isBlackSquare = ( ( i + j ) % 2 ) == 0;
		    if ( isBlackSquare ) {
			g2d.setColor( blackSquareColor );
		    } else {
			g2d.setColor( Color.white );
		    }
		    g2d.fillRect(
			i * squareSize ,
			( 8 - j - 1 ) * squareSize ,
			squareSize, squareSize );
		    g2d.setColor( Color.black );
		    if ( positionLoaded ) {
			int	p = chessboard[ i ][ j ];
			if ( p != Position.EMPTY_SQUARE ) {
			    Piece	pi    = position.piece( p );
			    int	index = pi.type();
			    if ( pi.colour() == Piece.BLACK ) {
				index += 6;
			    }
			    g2d.drawImage(
				images[ index ],
				i * squareSize ,
				( 8 - j - 1 ) * squareSize ,
				this );
			}
		    }
		}
	    }
	    if ( positionLoaded ) {
		g2d.drawString(
		    ( "(" + position.numberOfWhitePieces()
		      + " + " + position.numberOfBlackPieces() + ')' ),
		    0,
		    9 * squareSize );
	    }
	}

	private Position	position       = null  ;
	private boolean		positionLoaded = false ;
	private String		emptyDiagramTooltip;

	private  final int	squareSize     = 16 ;
	private  final Image[]	images         = new Image[ 12 ];

    }

    // --- Constructor -------------------------------------------------------

    public ChessFileChooser( Resources resources ) {

	JPanel	pane = new JPanel();
	pane.setLayout( new BorderLayout() );

	previewer = new ForsythePreviewer( resources );
	pane.add( previewer, BorderLayout.CENTER );
	pane.add( Box.createHorizontalGlue(), BorderLayout.SOUTH );

	setAccessory( pane );
	addPropertyChangeListener( new PropertyChangeListener() {
		public void propertyChange( PropertyChangeEvent e ) {
		    if ( e.getPropertyName().equals(
			JFileChooser.SELECTED_FILE_CHANGED_PROPERTY ) ) {
			File f = (File)e.getNewValue();
			showFileContents( f );
		    }
		}
	    } );
    }

    protected void showFileContents( File f ) {
	previewer.initPosition( f );
    }

    //--------------------------------------------------------------------------

    public static void setLocalisedResources( Resources resources ) {
	String[]	ui = {
	    "FileChooser.acceptAllFileFilterText",
	    "FileChooser.ancestorInputMap",
	    "FileChooser.cancelButtonText",
	    "FileChooser.cancelButtonMnemonic",
	    "FileChooser.cancelButtonToolTipText",
	    "FileChooser.detailsViewButtonAccessibleName",
	    "FileChooser.detailsViewButtonToolTipText",
	    "FileChooser.detailsViewIcon",
	    "FileChooser.directoryDescriptionText",
	    "FileChooser.fileDescriptionText",
	    "FileChooser.fileNameLabelMnemonic",
	    "FileChooser.fileNameLabelText",
	    "FileChooser.filesOfTypeLabelMnemonic",
	    "FileChooser.filesOfTypeLabelText",
	    "FileChooser.helpButtonMnemonic",
	    "FileChooser.helpButtonText",
	    "FileChooser.helpButtonToolTipText",
	    "FileChooser.homeFolderAccessibleName",
	    "FileChooser.homeFolderIcon",
	    "FileChooser.homeFolderToolTipText",
	    "FileChooser.listViewButtonAccessibleName",
	    "FileChooser.listViewButtonToolTipText",
	    "FileChooser.listViewIcon",
	    "FileChooser.lookInLabelMnemonic",
	    "FileChooser.lookInLabelText",
	    "FileChooser.newFolderAccessibleNam",
	    "FileChooser.newFolderErrorSeparator",
	    "FileChooser.newFolderErrorText",
	    "FileChooser.newFolderIcon",
	    "FileChooser.newFolderToolTipText",
	    "FileChooser.openButtonMnemonic",
	    "FileChooser.openButtonText",
	    "FileChooser.openButtonToolTipText",
	    "FileChooser.saveButtonMnemonic",
	    "FileChooser.saveButtonText",
	    "FileChooser.saveButtonToolTipText",
	    "FileChooser.upFolderAccessibleName",
	    "FileChooser.upFolderIcon",
	    "FileChooser.upFolderToolTipText",
	    "FileChooser.updateButtonMnemonic",
	    "FileChooser.updateButtonText",
	    "FileChooser.updateButtonToolTipText"
	};

	for ( int i = 0; i < ui.length; i++ ) {
	    try {
		String	s = resources.getString( ui[ i ] );
		if ( ui[ i ].endsWith( "Mnemonic" ) ) {
		    UIManager.put( ui[ i ],
				   new Integer( s.charAt( 0 ) ) );
		} else {
		    UIManager.put( ui[ i ], s );
		}
	    } catch ( MissingResourceException mre ) {
		// Nothing to do
	    }
	}
    }

    //--------------------------------------------------------------------------

    private ForsythePreviewer	previewer;

}
