/*
 * ChessGame.java --- Main class to store the chess stuff.
 *
 * Copyright (c) 2001, 2002, 2003, 2004 by Pascal Wassong All Rights Reserved.
 *
 * This file is part of iNatch.
 *
 * iNatch is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * iNatch is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

package iNatch;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.Reader;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.StringTokenizer;

import wassong.gui.tools.Resources;

public class ChessGame {

    public ChessGame ( Resources theResources, MyProperties theProperties ) {
	resources  = theResources;
	properties = theProperties;
	newGame( new StandardBoard(), new OrthodoxCondition() );
    }

    public void newGame( Board board, Condition condition ) {
	position		= new Position( board, condition, true );
	moves			= new Move[ NB_MAX_MOVES ];
	comments		= new ArrayList( 50 );
	file			= null;
	standardInitialPosition = true;

	nbMoves     = 0;
	currentMove = 0;

	finalPosition	   = (Position)position.clone();
	finalNumberOfMoves = 0;

	hasChanged = false;

	fireChessGameChanged();
    }

    //------------------------------------------------------------------------
    // Selectors
    //------------------------------------------------------------------------

    public Position position() {
	return this.position;
    }

    public int totalNumberOfMoves() {
	return nbMoves;
    }

    public int currentNumberOfMoves() {
	return currentMove;
    }

    public Move move( int i ) {
	if ( i < 0 || i >= nbMoves ) {
	    throw new IndexOutOfBoundsException();
	}
	return moves[ i ];
    }

    public List comments() {
	return this.comments;
    }

    public File file() {
	return this.file;
    }

    public String filename() {
	if ( this.file == null ) {
	    return null;
	}
	return this.file.getName();
    }

    public boolean hasChanged() {
	return this.hasChanged;
    }

    public Position finalPosition() {
	return this.finalPosition;
    }

    public int finalNumberOfMoves() {
	return this.finalNumberOfMoves;
    }

    //------------------------------------------------------------------------
    // Changing the game
    //------------------------------------------------------------------------

    public void	play( Move m ) {
	moves[ currentMove ] = m;
	nbMoves = currentMove + 1;
	position.play( m );
	currentMove++;
	hasChanged = true;
	fireChessPositionChanged();
    }

    public void killEndOfGame() {
	nbMoves = currentMove;
	hasChanged = true;
	fireChessPositionChanged();
    }

    public void replaceMove( int index, Move m ) {
	if ( index < 0 || index >= nbMoves ) {
	    throw new IllegalArgumentException();
	}

	int	n = currentMove;
	internalGotoMove( index );
	moves[ index ] = m;
	internalGotoMove( n );

	fireChessPositionChanged();
    }

    public void useFinalPositionAsInitialPosition() {
	position    = (Position)finalPosition.clone();
	nbMoves     = 0;
	currentMove = 0;

	standardInitialPosition = false;

	fireChessPositionChanged();
    }

    public void setBoard( Board board ) {
	position.setBoard( board );
	finalPosition.setBoard( board );
	fireChessPositionChanged();
    }

    public void setCondition( Condition condition ) {
	position.setCondition( condition );
	finalPosition.setCondition( condition );
	fireChessPositionChanged();
    }

    //------------------------------------------------------------------------
    // Changing the final position
    //------------------------------------------------------------------------

    public void clearFinalBoard() {
	finalPosition = new Position( new StandardBoard(),
				      new OrthodoxCondition(),
				      false );
	finalPosition.addPieceAt(
	    new Piece( Piece.KING, Piece.WHITE, "e1" ) );
	finalPosition.addPieceAt(
	    new Piece( Piece.KING, Piece.BLACK, "e8" ) );
	finalNumberOfMoves = 0;
	if ( nbMoves == 0 && standardInitialPosition ) {
	    hasChanged = true;
	}
	fireChessPositionChanged();
    }

    public void setInitialPositionInFinalBoard() {
	finalPosition = new Position( new StandardBoard(),
				      new OrthodoxCondition(),
				      true );
	finalNumberOfMoves = 0;
	fireChessPositionChanged();
    }

    public void clearSquare( int column, int row ) {
	if ( finalPosition.clearSquare( column, row ) ) {
	    if ( nbMoves == 0 && standardInitialPosition ) {
		hasChanged = true;
	    }
	    fireChessPositionChanged();
	}
    }

    public void addPieceAt( Piece piece ) {
	if ( finalPosition.addPieceAt( piece ) ) {
	    if ( nbMoves == 0 && standardInitialPosition ) {
		hasChanged = true;
	    }
	    fireChessPositionChanged();
	}
    }

    public void setFinalNumberOfMoves( int nb ) {
	if ( nb < 0 ) {
	    throw new IllegalArgumentException();
	}
	finalNumberOfMoves = nb;
	if ( nbMoves == 0 && standardInitialPosition ) {
	    hasChanged = true;
	}
	fireChessPositionChanged();
    }

    public void copyPositionOfGameToFinalPosition() {
	finalPosition      = (Position)position.clone();
	finalNumberOfMoves = currentMove;
	fireChessPositionChanged();
    }

    //------------------------------------------------------------------------
    // Moving in the game
    //------------------------------------------------------------------------

    public void forwardMove() {
	if ( currentMove < nbMoves ) {
	    position.play( moves[ currentMove ] );
	    currentMove++;
	    fireChessPositionChanged();
	}
    }

    public void backwardMove() {
	if ( currentMove > 0 ) {
	    currentMove--;
	    position.unplay( moves[ currentMove ] );
	    fireChessPositionChanged();
	}
    }

    public void beginningOfGame() {
	if ( currentMove > 0 ) {
	    while ( currentMove > 0 ) {
		currentMove--;
		position.unplay( moves[ currentMove ] );
	    }
	    fireChessPositionChanged();
	}
    }

    public void endOfGame() {
	if ( currentMove < nbMoves ) {
	    while ( currentMove < nbMoves ) {
		position.play( moves[ currentMove ] );
		currentMove++;
	    }
	    fireChessPositionChanged();
	}
    }

    public void gotoMove( int n ) {
	internalGotoMove( n );
	fireChessPositionChanged();
    }

    private void internalGotoMove( int n ) {
	if ( n < 0 ) {
	    n = 0;
	} else if ( n > nbMoves ) {
	    n = nbMoves;
	}
	if ( n == currentMove ) {
	    return;
	}

	if ( n < currentMove ) {
	    while ( currentMove > n ) {
		currentMove--;
		position.unplay( moves[ currentMove ] );
	    }
	} else {
	    while ( currentMove < n ) {
		position.play( moves[ currentMove ] );
		currentMove++;
	    }
	}
    }

    //------------------------------------------------------------------------
    // Experimental part
    //------------------------------------------------------------------------

    public Move[] alternateMoves( int column, int row ) {

	int	index = position.chessboard()[ column ][ row ];
	if ( index == Position.EMPTY_SQUARE ) {
	    return new Move[ 0 ];
	}

	Position pos = new Position( new StandardBoard(),
				     new OrthodoxCondition(),
				     true );
	for ( int i = 0; i < nbMoves; i++ ) {
	    pos.play( moves[ i ] );
	}

	boolean	pieceFound  = false;
	int	movesIndex  = nbMoves;
	int	x           = 0;
	int	y           = 0;
	int	colour      = 0;
	while ( !pieceFound && movesIndex > 0 ) {
	    movesIndex--;
	    Move	m = moves[ movesIndex ];
	    Piece	p = m.destination();
	    if ( p.column() == column && p.row() == row ) {
		colour = p.colour();
		x = m.origin().column();
		y = m.origin().row();
		pieceFound = true;
	    }
	    pos.unplay( m );
	}
	if ( !pieceFound ) {
	    // This may happen in case of a castling
	    return new Move[ 0 ];
	}

	Move[]	allMoves        = pos.moves( colour, null );
	Move[]	possibleMoves   = new Move[ allMoves.length ];
	int	nbPossibleMoves = 0;
	for ( int i = 0; i < allMoves.length; i++ ) {
	    Move	m = allMoves[ i ];
	    Piece	p = m.origin();
	    if ( p.column() == x && p.row() == y ) {
		int j = 0;
		try {
		    Position	tempPos = (Position)pos.clone();
		    tempPos.play( m );
		    for ( j = movesIndex + 1; j < nbMoves; j++ ) {
			tempPos.play( moves[ j ] );
		    }
		    possibleMoves[ nbPossibleMoves ] = m;
		    nbPossibleMoves++;
		} catch ( IllegalArgumentException e ) {
		    // Nothing to do
		}
	    }
	}

	allMoves = new Move[ nbPossibleMoves ];
	for ( int i = 0; i < nbPossibleMoves; i++ ) {
	    allMoves[ i ] = possibleMoves[ i ];
	}
	return allMoves;
    }

    //------------------------------------------------------------------------
    // load & save part
    //------------------------------------------------------------------------

    public void loadGame( File f ) {
	try {
	    Reader r = new FileReader( f );
	    newGame( new StandardBoard(), new OrthodoxCondition() );
	    this.file = f;
	    loadGame( r );
	    hasChanged = false;
	    fireChessGameChanged();
	} catch ( FileNotFoundException fnfe ) {
	    String[]	args = { f.getName() };
	    Error.errorMessage( "load-game-error-title",
				"file-not-found",
				f.getName() );
	}
    }

    public void loadGame( Reader r ) {
	try {
	    BufferedReader in = new BufferedReader( r );

	    boolean	positionFound     = false;
	    boolean	positionJustFound = false;

	    boolean	addToComments = comments.isEmpty();
	    String	s;

	    Position	loadedInitialPosition = null;

	    while ( ( s = in.readLine() ) != null ) {

		if ( addToComments ) {
		    comments.add( s );
		}
		s = s.trim();

		if ( positionJustFound ) {
		    int index = 0;
		    while ( index < s.length()
			    && Character.isDigit( s.charAt( index ) ) ) {
			index++;
		    }
		    if ( index > 0 ) {
			finalNumberOfMoves
			    = Integer.parseInt( s.substring( 0, index ) );
			if ( index < s.length() && s.charAt( index ) == '.' ) {
			    finalNumberOfMoves *= 2;
			    if ( index + 1 < s.length()
				 && s.charAt( index + 1 ) == '5' ) {
				finalNumberOfMoves++;
			    }
			}
		    }
		    positionJustFound = false;
		}

		if ( loadedInitialPosition == null ) {
		    PositionFactory.make( new StandardBoard(),
					  new OrthodoxCondition(),
					  s );
		    if ( PositionFactory.status()
			 == PositionFactory.CORRECT_POSITION ) {

			loadedInitialPosition = PositionFactory.position();
			if ( !positionFound ) {
			    finalPosition	  = loadedInitialPosition;
			    positionFound	  = true;
			    positionJustFound	  = true;
			    finalNumberOfMoves	  = 0;
			    loadedInitialPosition = null;
			}
		    }
		}

		if ( s.startsWith( "1." ) || s.startsWith( "1 " ) ) {
		    StringBuffer allMoves = new StringBuffer( s );
		    while ( ( s = in.readLine() ) != null ) {
			if ( addToComments ) {
			    comments.add( s );
			}
			if ( s.length() > 0 ) {
			    allMoves.append( " " + s );
			} else {
			    break;
			}
		    }

		    Move[] initialPositionMoves = null;
		    if ( loadedInitialPosition != null ) {
			initialPositionMoves = loadMoves(
			    allMoves.toString(), loadedInitialPosition );
		    }

		    Position p;
		    if ( nbMoves > 0 ) {
			p = new Position( new StandardBoard(),
					  new OrthodoxCondition(),
					  true );
		    } else {
			// This is to allow `Paste' to a non standard initial
			// position.
			p = (Position)position.clone();
		    }

		    Move[] originMoves = loadMoves( allMoves.toString(), p );

		    if ( originMoves != null || initialPositionMoves != null ) {
			Move[] movesToTake;
			if ( initialPositionMoves == null
			     || ( originMoves != null
				  && ( originMoves.length
				       > initialPositionMoves.length ) ) ) {
			    movesToTake = originMoves;
			    position	= p;
			} else {
			    movesToTake = initialPositionMoves;
			    position	= loadedInitialPosition;
			}

			nbMoves = movesToTake.length;
			for ( int i = 0; i < nbMoves; i++ ) {
			    moves[ i ] = movesToTake[ i ];
			    position.play( moves[ i ] );
			}
			currentMove = nbMoves;
			break;
		    }
		}
	    }
	    if ( addToComments ) {
		while ( ( s = in.readLine() ) != null ) {
		    comments.add( s );
		}
	    }
	    hasChanged = true;
	    fireChessGameChanged();
	} catch ( IOException e ) {
	    System.err.println( "IOException catched : " + e );
	}
    }

    public void setComments( List newComments ) {

	comments.clear();
	Iterator it = newComments.iterator();
	while ( it.hasNext() ) {
	    comments.add( it.next() );
	}
	fireChessGameChanged();
    }

    public void saveGame( File f ) {
	FileWriter	out;
	try {
	    out = new FileWriter( f );
	} catch ( IOException e ) {
	    Error.errorMessage(
		"save-error-title",
		"unable-to-create",
		f.getPath() );
	    return;
	}

	try {
	    StringTokenizer	t;

	    if ( standardInitialPosition ) {
		if ( nbMoves == 0 ) {
		    t = new StringTokenizer(
			this.formatOutput(
			    "final-board-diagram;final-board-forsythe;"
			    + "final-stipulation" ),
			"\n",
			true );
		} else {
		    t = new StringTokenizer(
			formatOutput(
			    "board-diagram;board-forsythe;"
			    + "stipulation;empty-line;moves-long" ),
			"\n",
			true );
		}
	    } else {
		t = new StringTokenizer(
		    this.formatOutput(
			"board-diagram;board-forsythe;stipulation;empty-line;"
			+ "board-initial-position;empty-line;"
			+ "moves-long" ),
		    "\n",
		    true );
	    }
	    while ( t.hasMoreTokens() ) {
		// TODO : use StringBuffer to optimise
		String	s = t.nextToken();
		while ( s.length() > 72 ) {
		    int	pos = s.lastIndexOf( ' ', 72 );
		    if ( pos == -1 ) {
			pos = s.indexOf( ' ', 71 );
		    }
		    if ( pos != -1 ) {
			out.write( s.substring( 0, pos ) + '\n' );
			s = s.substring( pos + 1 );
		    } else {
			out.write( s );
			s = "";
		    }
		}
		out.write( s );
	    }

	    hasChanged = false;
	    this.file  = f;
	    fireChessGameChanged(); // For people using game name
	} catch ( IOException e ) {
	    Error.errorMessage(
		"save-error-title",
		"error-during-save",
		f.getPath() );
	} finally {
	    try {
		out.close();
	    } catch ( IOException e ) {
		// Nothing to do
	    }
	}
    }

    public String formatOutput( String format ) {

	StringBuffer	s = new StringBuffer();
	StringTokenizer	t = new StringTokenizer( format, ";" );
	String[]	listOfTypes = formatTypes();
	while ( t.hasMoreTokens() ) {
	    String	type = t.nextToken();
	    int		i;
	    for ( i = 0; i < listOfTypes.length; i++ ) {
		if ( type.equals( listOfTypes[ i ] ) ) {
		    break;
		}
	    }

	    switch ( i ) {
	    case FORMAT_BOARD_DIAGRAM:
		s.append( position().diagram() );
		break;
	    case FORMAT_BOARD_FORSYTHE:
		s.append( position().toString() )
		    .append( '\n' );
		break;
	    case FORMAT_BOARD_TRUE_TYPE:
		s.append( position().font1Echecs() );
		break;
	    case FORMAT_FINAL_BOARD_DIAGRAM:
		s.append( finalPosition.diagram() );
		break;
	    case FORMAT_FINAL_BOARD_FORSYTHE:
		s.append( finalPosition().toString() )
		    .append( '\n' );
		break;
	    case FORMAT_FINAL_BOARD_TRUE_TYPE:
		s.append( finalPosition().font1Echecs() );
		break;
	    case FORMAT_BOARD_INITIAL_POSITION:
		int n = currentMove;
		internalGotoMove( 0 );
		s.append( resources.getString( "initial-position" ) )
		    .append( '\n' )
		    .append( position().diagram() )
		    .append( position().toString() )
		    .append( "\n(" )
		    .append( position().numberOfWhitePieces() )
		    .append( " + " )
		    .append( position().numberOfBlackPieces() )
		    .append( ")\n" );
		internalGotoMove( n );
		break;
	    case FORMAT_MOVES_SHORT:
		s.append( shortMoves() )
		    .append( '\n' );
		break;
	    case FORMAT_MOVES_LONG:
		s.append( longMoves() )
		    .append( '\n' );
		break;
	    case FORMAT_MOVES_REVERSE:
		s.append( reverseMoves() );
		break;
	    case FORMAT_STIPULATION:
		s.append( nbMovesToString( totalNumberOfMoves() ) )
		    .append( "\t(" )
		    .append( position().numberOfWhitePieces() )
		    .append( " + " )
		    .append( position().numberOfBlackPieces() )
		    .append( ")\n" );
		break;
	    case FORMAT_FINAL_STIPULATION:
		s.append( nbMovesToString( finalNumberOfMoves() ) )
		    .append( "\t(" )
		    .append( finalPosition().numberOfWhitePieces() )
		    .append( " + " )
		    .append( finalPosition().numberOfBlackPieces() )
		    .append( ")\n" );
		break;
	    case FORMAT_EMPTY_LINE:
		s.append( '\n' );
		break;
	    default:
		s.append( type );
		break;
	    }
	}
	return s.toString();
    }

    //--------------------------------------------------------------------------
    // loadMoves
    //--------------------------------------------------------------------------

    /**
     * Transforms the string into moves. Tries all known languages and takes
     * the one where the most moves have been decoded.
     *
     * @param allMoves A string containing the moves to decode.
     */
    private Move[] loadMoves( String allMoves, Position initialPosition ) {
	String	forsytheInitials = resources.getString( "forsytheInitials" );
	StringTokenizer t = new StringTokenizer( forsytheInitials );

	Move[]	longestMoves   = null;
	int	nbLongestMoves = 0;
	while ( t.hasMoreTokens() ) {
	    Move[] theMoves
		= loadMoves( initialPosition, allMoves, t.nextToken() );
	    if ( theMoves.length > nbLongestMoves ) {
		nbLongestMoves = theMoves.length;
		longestMoves = theMoves;
	    }
	}

	return longestMoves;
    }

    private Move[] loadMoves( Position	initialPosition  ,
			      String	allMoves         ,
			      String	forsytheInitials ) {

	Position	pos           = (Position)initialPosition.clone();
	int		colour        = Piece.WHITE;
	int		nbLoadedMoves = 0;
	Move[]		loadedMoves   = new Move[ NB_MAX_MOVES ];

	StringTokenizer t = new StringTokenizer( allMoves );
	while ( t.hasMoreTokens() ) {
	    MoveFactory.make( t.nextToken(),
			      colour,
			      forsytheInitials,
			      pos );
	    int status = MoveFactory.status();
	    if ( status == MoveFactory.CORRECT_MOVE ) {
		loadedMoves[ nbLoadedMoves ] = MoveFactory.move();
		try {
		    pos.play( loadedMoves[ nbLoadedMoves ] );
		    nbLoadedMoves++;
		    colour = ( colour == Piece.WHITE )
			? Piece.BLACK : Piece.WHITE;
		} catch ( IllegalArgumentException e ) {
		    break;
		}
	    } else if ( status == MoveFactory.EMPTY_MOVE ) {
		continue;
	    } else {
		break;
	    }
	}
	Move[]	theMoves = new Move[ nbLoadedMoves ];
	for ( int i = 0; i < nbLoadedMoves; i++ ) {
	    theMoves[ i ] = loadedMoves[ i ];
	}
	return theMoves;
    }

    //------------------------------------------------------------------------
    // Listeners part
    //------------------------------------------------------------------------

    private List	gameChangedListeners	 = new LinkedList();
    private List	positionChangedListeners = new LinkedList();

    public void addChessGameListener( ChessGameListener l ) {
	gameChangedListeners.add( l );
    }

    private void fireChessGameChanged() {
	Iterator it = gameChangedListeners.iterator();
	while ( it.hasNext() ) {
	    ( (ChessGameListener)it.next() ).chessGameChanged();
	}
	fireChessPositionChanged();
    }

    public void addChessPositionListener( ChessGameListener l ) {
	positionChangedListeners.add( l );
    }

    private void fireChessPositionChanged() {
	Iterator it = positionChangedListeners.iterator();
	while ( it.hasNext() ) {
	    ( (ChessGameListener)it.next() ).chessPositionChanged();
	}
    }

    //--------------------------------------------------------------------------
    // Conversion to strings
    //--------------------------------------------------------------------------

    private static final String[]	CHECK_STRINGS = { "", "+", "++", "#" };
    public String shortMoves() {
	StringBuffer	s = new StringBuffer();

	int		n = currentMove;
	internalGotoMove( 0 );
	Position	p = (Position)position.clone();
	internalGotoMove( n );


	for ( int i = 0; i < nbMoves; i++ ) {
	    int	colour;
	    if ( ( i & 1 ) == 0 ) {
		colour = Piece.BLACK;
		if ( i != 0 ) {
		    s.append( ' ' );
		}
		s.append( String.valueOf( i / 2 + 1 )  + '.' );
	    } else {
		colour = Piece.WHITE;
		s.append( ' ' );
	    }
	    s.append( moves[ i ].toShortString( p ) );
	    p.play( moves[ i ] );

	    int	check = p.checkStatus( colour , moves[ i ] );
	    s.append( CHECK_STRINGS[ check ] );
	}
	return s.toString();
    }

    public String longMoves() {
	StringBuffer	s = new StringBuffer();

	int		n = currentMove;
	internalGotoMove( 0 );
	Position	p = (Position)position.clone();
	internalGotoMove( n );

	for ( int i = 0; i < nbMoves; i++ ) {
	    int	colour;
	    if ( ( i & 1 ) == 0 ) {
		colour = Piece.BLACK;
		if ( i != 0 ) {
		    s.append( ' ' );
		}
		s.append( String.valueOf( i / 2 + 1 )  + '.' );
	    } else {
		colour = Piece.WHITE;
		s.append( ' ' );
	    }
	    s.append( moves[ i ].toBeautifullString().trim() );
	    p.play( moves[ i ] );

	    int	check = p.checkStatus( colour , moves[ i ] );
	    s.append( CHECK_STRINGS[ check ] );
	}
	return s.toString();
    }

    public String longMove( int i ) {

	int	n = currentMove;
	internalGotoMove( i + 1 );

	int	colour = ( ( i & 1 ) == 0 ) ? Piece.BLACK : Piece.WHITE;
	int	check = position.checkStatus( colour , moves[ i ] );

	String	s = moves[ i ].toBeautifullString() + CHECK_STRINGS[ check ];

	internalGotoMove( n );

	return s;
    }

    public String reverseMoves() {

	StringBuffer	s = new StringBuffer();

	Position	p = (Position)position.clone();

	if ( ( currentMove & 1 ) == 0 ) {
	    s.append( "1. ..." );
	}
	for ( int i = currentMove - 1; i >= 0; i-- ) {
	    int	colour;
	    if ( ( i & 1 ) == 0 ) {
		colour = Piece.BLACK;
		if ( i != currentMove - 1 ) {
		    s.append( ' ' );
		}
		s.append(
		    String.valueOf( ( currentMove - i ) / 2 + 1 )  + '.' );
	    } else {
		colour = Piece.WHITE;
		s.append( ' ' );
	    }
	    s.append( moves[ i ].toStringWithCapture().trim() );

	    int	check = p.checkStatus( colour , moves[ i ] );
	    s.append( CHECK_STRINGS[ check ] );

	    p.unplay( moves[ i ] );
	}
	return s.toString();
    }

    //------------------------------------------------------------------------
    // I don't think that the following is at it's best place here.  But where
    // is it's best place ?
    //------------------------------------------------------------------------

    // This is the sole function that uses `properties'.  So if this function is
    // moved somewhere else, please remove this parameter from the constructor.
    // The test functions will be gratefull.
    public String nbMovesToString( int nb ) {
	String s;
	if ( properties.getBooleanProperty(
		 MyProperties.NOTATION_SHOW_SINGLE_MOVES ) ) {
	    s = String.valueOf( nb )
		+ ' '
		+ resources.getString(
		    nb < 2 ? "single-move" : "single-moves" );
	} else {
	    if ( nb != 0 ) {
		s = String.valueOf( nb / 2.0 );
	    } else {
		s = "0";
	    }
	    s += ' ' + resources.getString(
		nb < 3 ? "move" : "moves" );
	}
	return s;
    }

    public static String[] formatTypes() {
	return LIST_OF_TYPES;
    }

    private static final String[] LIST_OF_TYPES = {
	"board-diagram",
	"board-forsythe",
	"board-true-type",
	"final-board-diagram",
	"final-board-forsythe",
	"final-board-true-type",
	"board-initial-position",
	"moves-short",
	"moves-long",
	"moves-reverse",
	"stipulation",
	"final-stipulation",
	"empty-line"
    };

    private static final int	FORMAT_BOARD_DIAGRAM	      = 0;
    private static final int	FORMAT_BOARD_FORSYTHE	      = 1;
    private static final int	FORMAT_BOARD_TRUE_TYPE        = 2;
    private static final int	FORMAT_FINAL_BOARD_DIAGRAM    = 3;
    private static final int	FORMAT_FINAL_BOARD_FORSYTHE   = 4;
    private static final int	FORMAT_FINAL_BOARD_TRUE_TYPE  = 5;
    private static final int	FORMAT_BOARD_INITIAL_POSITION = 6;
    private static final int	FORMAT_MOVES_SHORT	      = 7;
    private static final int	FORMAT_MOVES_LONG	      = 8;
    private static final int	FORMAT_MOVES_REVERSE	      = 9;
    private static final int	FORMAT_STIPULATION	      = 10;
    private static final int	FORMAT_FINAL_STIPULATION      = 11;
    private static final int	FORMAT_EMPTY_LINE	      = 12;

    //------------------------------------------------------------------------
    // Attributes
    //------------------------------------------------------------------------

    private Resources		resources;
    private MyProperties	properties;

    private File		file ;

    private boolean		hasChanged ;

    private Position		position ;
    private Move[]		moves ;
    private int			nbMoves     = 0 ;
    private int			currentMove = 0 ;
    private List		comments ;

    private boolean		standardInitialPosition ;

    private Position		finalPosition ;
    private int			finalNumberOfMoves ;

    private static final int	NB_MAX_MOVES = 200;

}
