/*
 * ContentPanel.java --- The top level content panel of the application.
 *
 * Copyright (c) 2001, 2002, 2003, 2004 by Pascal Wassong All Rights Reserved.
 *
 * This file is part of iNatch.
 *
 * iNatch is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * iNatch is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

package iNatch;

import java.awt.BorderLayout;
import java.awt.Dimension;
import java.awt.FileDialog;
import java.awt.Toolkit;
import java.awt.datatransfer.DataFlavor;
import java.awt.datatransfer.Transferable;
import java.awt.event.ActionEvent;
import java.awt.event.ComponentAdapter;
import java.awt.event.ComponentEvent;
import java.io.File;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.Reader;
import java.io.StringReader;
import java.net.URL;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.Map;
import java.util.Properties;
import java.util.StringTokenizer;
import javax.swing.AbstractAction;
import javax.swing.Action;
import javax.swing.BorderFactory;
import javax.swing.ImageIcon;
import javax.swing.JComponent;
import javax.swing.JFileChooser;
import javax.swing.JFrame;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JProgressBar;
import javax.swing.JSplitPane;
import javax.swing.UIManager;
import javax.swing.text.BadLocationException;
import javax.swing.text.PlainDocument;

import nanoxml.XMLElement;
import wassong.gui.tools.MyMenuBar;
import wassong.gui.tools.Resources;
import wassong.gui.tools.ToolBar;

public class ContentPanel extends JPanel {

    ContentPanel( JFrame aFrame, String name ) {
	super( true );

	frame		= aFrame;
	applicationName = name;

	// Force SwingSet to come up in the Cross Platform L&F
	try {
	    UIManager.setLookAndFeel(
		UIManager.getCrossPlatformLookAndFeelClassName() );
	    // If you want the System L&F instead, comment out the above
	    // lines and uncomment the following:
	    //	    UIManager.setLookAndFeel(
	    //		UIManager.getSystemLookAndFeelClassName() );
	} catch ( Exception exc ) {
	    System.err.println( "Error loading L&F: " + exc );
	}

	setBorder( BorderFactory.createEtchedBorder() );
	setLayout( new BorderLayout() );


	// Resources
	resources = new Resources( RESOURCE_FILE );


	// Properties
	try {
	    InputStream	is =
		this.getClass().getResourceAsStream( DEFAULT_PROPERTIES_FILE );
	    Properties defaultProperties = new Properties();
	    defaultProperties.load( is );
	    // Default language depends on system language
	    defaultProperties.setProperty(
		MyProperties.OUTPUT_LANGUAGE,
		resources.getString( "defaultLanguage" ) );
	    properties = new MyProperties( defaultProperties, PROPERTIES_FILE );
	} catch ( IOException ioe ) {
	    System.err.println( "Unable to load default properties file : "
				+ DEFAULT_PROPERTIES_FILE );
	    System.exit( 0 );
	}


	// Various initialisations
	new PieceProperties( resources, properties );
	new MoveProperties( resources, properties );

	PositionFactory.setForsytheInitials(
	    resources.getString( "forsytheInitials" ) );

	chessGame = new ChessGame( resources, properties );
	copyPaste = new CopyPaste( chessGame, properties );

	setTitle();

	Error.initialise( resources, frame, name() );


	// All actions for menu and toolbar
	Map allActions = new HashMap();
	Action[] actions = getActions();
	for ( int i = 0; i < actions.length; i++ ) {
	    Action a = actions[i];
	    allActions.put( a.getValue( Action.NAME ), a );
	}

	// All components, only used by toolbar
	Map components = new HashMap();
	components.put( "change-language",
			new LanguageCombo( resources, properties ) );

	// Create menu toolbar from XML
	JPanel toolbarPanel = new JPanel( new BorderLayout() );
	XMLElement xml = new XMLElement();
	try {
	    xml.parseFromReader(
		new InputStreamReader(
		    this.getClass().getClassLoader().getResourceAsStream(
			GUI_FILE ) ) );
	} catch ( IOException e ) {
	    System.err.println( "IOException : " + e );
	    e.printStackTrace();
	}

	Enumeration childs = xml.enumerateChildren();
	while ( childs.hasMoreElements() ) {
	    XMLElement child = (XMLElement)childs.nextElement();
	    if ( child.getName().equals( "menubar" ) ) {
		aFrame.setJMenuBar(
		    new MyMenuBar( child, allActions, resources ) );
	    } else if ( child.getName().equals( "toolbar" ) ) {
		toolbarPanel.add( new ToolBar( name() + " toolbar",
					       child,
					       allActions,
					       components,
					       resources ),
				  BorderLayout.NORTH );
	    }
	}

	// Places the diagram part
	JPanel centerPanel = new JPanel( new BorderLayout() );

	JComponent finalDiagram = new DiagramPanel(
	    new FinalDiagram( resources, properties, chessGame ),
	    properties );
	JComponent mainDiagram = new DiagramPanel(
	    new MainDiagram ( resources, properties, chessGame ), properties );
	JSplitPane diagrams = new JSplitPane(
	    JSplitPane.HORIZONTAL_SPLIT, finalDiagram, mainDiagram );
	handleDividerLocation(
	    diagrams, finalDiagram, MyProperties.SIZE_RIGHT_DIAGRAM );
	centerPanel.add( diagrams, BorderLayout.NORTH );

	CommentPanel comments = new CommentPanel( chessGame );
	centerPanel.add( comments, BorderLayout.CENTER );

	// Places the moves part
	JComponent movesPanel
	    = new MovesPanel( resources, properties, chessGame );
	movesPanel.setBorder( BorderFactory.createEtchedBorder() );

	JSplitPane globalPanel = new JSplitPane(
	    JSplitPane.HORIZONTAL_SPLIT, centerPanel, movesPanel );
	handleDividerLocation(
	    globalPanel, centerPanel, MyProperties.SIZE_MOVES );

	toolbarPanel.add( globalPanel, BorderLayout.CENTER );

	add( toolbarPanel, BorderLayout.CENTER );

	chessGame.addChessGameListener( new ChessGameAdapter() {
		public void chessGameChanged() {
		    setTitle();
		}
	    } );

	handleSize( this, MyProperties.WINDOW_SIZE );

    }

    //--------------------------------------------------------------------------

    private void handleSize( final JComponent c, final String property ) {
	String size = properties.getProperty( property );
	if ( !size.equals( "default" ) ) {
	    StringTokenizer t = new StringTokenizer( size, "," );
	    int x = Integer.parseInt( t.nextToken().trim() );
	    int y = Integer.parseInt( t.nextToken().trim() );
	    c.setPreferredSize( new Dimension( x, y ) );
	}
	c.addComponentListener( new ComponentAdapter() {
		public void componentResized( ComponentEvent e ) {
		    properties.setProperty( property,
					    Integer.toString( c.getWidth() )
					    + ", " + c.getHeight() );
		}
	    } );
    }

    private void handleDividerLocation( final JSplitPane pane,
					JComponent c,
					final String property ) {

	pane.setOneTouchExpandable( true );

	int size = properties.getIntProperty( property );
	pane.setDividerLocation( size );
	c.addComponentListener( new ComponentAdapter() {
		public void componentResized( ComponentEvent e ) {
		    properties.setProperty(
			property,
			Integer.toString( pane.getDividerLocation() ) );
		}
	    } );
    }

    //--------------------------------------------------------------------------

    public Action[] getActions() {
	Action[] defaultActions = {
	    new AboutAction(),

	    new ForwardMoveAction(),
	    new BackwardMoveAction(),
	    new BeginningOfGameAction(),
	    new EndOfGameAction(),
	    new InsertNullMoveAction(),
	    new KillEndOfGameAction(),

	    new NewAction(),
	    new CopyAction(),
	    new PasteAction(),
	    new CopyBoardForsytheAction(),
	    new CopyBoardDiagramAction(),
	    new CopyBoardBothAction(),
	    new CopyBoard1EchecsFontAction(),
	    new CopyFinalBoardForsytheAction(),
	    new CopyFinalBoardDiagramAction(),
	    new CopyFinalBoardBothAction(),
	    new CopyFinalBoard1EchecsFontAction(),
	    new CopyMovesShortAction(),
	    new CopyMovesLongAction(),
	    new CopyMovesReverseAction(),

	    new ClearFinalBoardAction(),
	    new SetInitialPositionAction(),
	    new UseFinalPositionAsInitialPosition(),
	    new GetPositionOfGameAction(),
	    new SetOrthodoxBoardAction(),
	    new SetVerticalCylinderBoardAction(),
	    new SetOrhthodoxConditionAction(),
	    new SetMonochromaticConditionAction(),
	    new SetEinsteinConditionAction(),

	    new OpenAction(),
	    new SaveAction(),
	    new SaveAsAction(),
	    new PreferencesAction(),
	    new QuitAction(),

	    new VerifyAction(),
	};
	return defaultActions;
    }

    public void quit() {
	if ( gameDiscarded() ) {
	    try {
		properties.store(
		    new FileOutputStream( PROPERTIES_FILE ),
		    " " + name() + ' ' + version() );
	    } catch ( IOException ioe ) {
		// Nothing to do
	    }
	    System.exit( 0 );
	}
    }

    //--------------------------------------------------------------------------

    // This function is only used by New and Quit.  Maybe it should have a more
    // protected place ??
    private boolean gameDiscarded() {
	if ( chessGame.hasChanged() ) {
	    String[] options = new String[ 3 ];
	    options[ 0 ] = resources.getString( "yes-string" );
	    options[ 1 ] = resources.getString( "no-string" );
	    options[ 2 ] = resources.getString( "cancel-string" );
	    int	answer =
		JOptionPane.showOptionDialog(
		    getFrame(),
		    resources.getString( "not-saved-question" ),
		    name() + " - " + resources.getString(
			"not-saved-title" ),
		    JOptionPane.YES_NO_CANCEL_OPTION ,
		    JOptionPane.QUESTION_MESSAGE,
		    null,
		    options,
		    null );
	    if ( answer == JOptionPane.CANCEL_OPTION ) {
		return false;
	    } else if ( answer == JOptionPane.YES_OPTION ) {
		if ( !saveGame( false ) ) {
		    return false;
		}
	    }
	}
	return true;
    }

    //--------------------------------------------------------------------------
    // New, Open, Save, Quit ... actions
    //--------------------------------------------------------------------------

    public String name() {
	return applicationName;
    }

    public String version() {
	return CURRENT_VERSION;
    }

    private JFrame getFrame() {
	return frame;
    }

    private void setTitle() {
	String	filename = chessGame.filename();
	String  title;
	if ( filename == null || filename.length() == 0 ) {
	    title = name();
	} else {
	    title = name() + " - " + filename;
	}
	getFrame().setTitle( title );
    }

    private boolean saveGame( boolean saveAsAction ) {
	File f = chessGame.file();
	if ( f == null || saveAsAction ) {
	    if ( fileDialog == null ) {
		fileDialog = new FileDialog( getFrame() );
	    }
	    fileDialog.setMode( FileDialog.SAVE );
	    fileDialog.setDirectory(
		properties.getProperty( MyProperties.SAVE_DIRECTORY ) );
	    fileDialog.show();

	    String file = fileDialog.getFile();
	    if ( file == null ) {
		return false;
	    }
	    String directory = fileDialog.getDirectory();
	    properties.setProperty( MyProperties.SAVE_DIRECTORY, directory );
	    f = new File( directory, file );
	}

	if ( f.exists() ) {
	    boolean	deleted;
	    try {
		deleted = f.delete();
	    } catch ( SecurityException se ) {
		deleted = false;
	    }
	    if ( !deleted ) {
		Error.errorMessage(
		    "save-error-title",
		    "unable-to-delete",
		    f.getPath() );
		return false;
	    }
	}
	chessGame.saveGame( f );
	return true;
    }

    /**
     * Resets the undo manager.
     */
    protected void resetUndoManager() {
	/*
	  undo.discardAllEdits();
	  undoAction.update();
	  redoAction.update();
	*/
    }


    /*
      protected UndoableEditListener undoHandler = new UndoHandler();
      protected UndoManager undo = new UndoManager();
      class UndoHandler implements UndoableEditListener {
      public void undoableEditHappened( UndoableEditEvent e ) {
      undo.addEdit( e.getEdit() );
      undoAction.update();
      redoAction.update();
      }
      }
    */

    // --- action implementations -----------------------------------

    private abstract class MyAbstractAction extends AbstractAction {

	MyAbstractAction( String key ) {
	    super( key );
	    URL url = resources.getResource( key + "-Image" );
	    if ( url != null ) {
		putValue( SMALL_ICON, new ImageIcon( url ) );
	    }
	}
    }

    private abstract class ChessAbstractAction extends MyAbstractAction {
	ChessAbstractAction( String key ) {
	    super( key );
	    chessGame.addChessPositionListener( new ChessGameAdapter() {
		    public void chessPositionChanged() {
			enable();
		    }
		} );
	    enable();
	}

	abstract void enable();
    }

    private class AboutAction extends MyAbstractAction {
	public AboutAction() {
	    super( "about" );
	}

	public void actionPerformed( ActionEvent e ) {
	    String about = resources.getString( "aboutStr" );
	    JOptionPane.showMessageDialog(
		getFrame(),
		name() + ' ' + version() + '\n' + COPYRIGHT + '\n'
		+ "http://Natch.free.fr/iNatch.html",
		about + ' ' + name(),
		JOptionPane.INFORMATION_MESSAGE );
	}
    }

    //--------------------------------------------------------------------------
    // Move actions
    //--------------------------------------------------------------------------

    private class ForwardMoveAction extends ChessAbstractAction {
	public ForwardMoveAction() {
	    super ( "forward-move" );
	}

	public void actionPerformed( ActionEvent e ) {
	    chessGame.forwardMove();
	}

	void enable() {
	    this.setEnabled( chessGame.currentNumberOfMoves()
			     < chessGame.totalNumberOfMoves() );
	}
    }

    private class BackwardMoveAction extends ChessAbstractAction {

	public BackwardMoveAction() {
	    super( "backward-move" );
	}

	public void actionPerformed( ActionEvent e ) {
	    chessGame.backwardMove();
	}

	void enable() {
	    this.setEnabled( chessGame.currentNumberOfMoves() > 0 );
	}
    }

    private class BeginningOfGameAction extends ChessAbstractAction {

	public BeginningOfGameAction() {
	    super( "beginning-of-game" );
	}

	public void actionPerformed( ActionEvent e ) {
	    chessGame.beginningOfGame();
	}

	void enable() {
	    this.setEnabled( chessGame.currentNumberOfMoves() > 0 );
	}
    }

    private class EndOfGameAction extends ChessAbstractAction {

	public EndOfGameAction() {
	    super( "end-of-game" );
	}

	public void actionPerformed( ActionEvent e ) {
	    chessGame.endOfGame();
	}

	void enable() {
	    this.setEnabled( chessGame.currentNumberOfMoves()
			     < chessGame.totalNumberOfMoves() );
	}
    }

    private class InsertNullMoveAction extends ChessAbstractAction {

	public InsertNullMoveAction() {
	    super( "insert-null-move" );
	}

	public void actionPerformed( ActionEvent e ) {
	    chessGame.play( Move.NULL_WHITE_MOVE );
	}

	void enable() {
	    this.setEnabled( chessGame.currentNumberOfMoves() == 0
			     || !( chessGame.move(
				       chessGame.currentNumberOfMoves() - 1 )
				   .isNullMove() ) );
	}
    }

    private class KillEndOfGameAction extends ChessAbstractAction {

	public KillEndOfGameAction() {
	    super( "kill-end-of-game" );
	}

	public void actionPerformed( ActionEvent e ) {
	    chessGame.killEndOfGame();
	}

	void enable() {
	    this.setEnabled( chessGame.currentNumberOfMoves()
			     < chessGame.totalNumberOfMoves() );
	}
    }

    //--------------------------------------------------------------------------
    // New, Open, Save, Quit ... actions
    //--------------------------------------------------------------------------

    class NewAction extends MyAbstractAction {

	public NewAction() {
	    super( "new" );
	}

	public NewAction( String nm ) {
	    super( nm );
	}

	public void actionPerformed( ActionEvent e ) {

	    if ( gameDiscarded() ) {
		chessGame.newGame( new StandardBoard(),
				   new OrthodoxCondition() );
	    }
	}

    }

    class OpenAction extends NewAction {

	public OpenAction() {
	    super( "open" );
	}

	public void actionPerformed( ActionEvent e ) {
	    if ( fileChooser == null ) {
		fileChooser = new ChessFileChooser( resources );
		fileChooser.setDialogType( JFileChooser.OPEN_DIALOG );
		String currentDir = System.getProperty( "user.dir" );
		fileChooser.setCurrentDirectory( new File( currentDir ) );
	    }
	    fileChooser.setCurrentDirectory(
		new File( properties.getProperty(
			      MyProperties.SAVE_DIRECTORY ) ) );
	    if ( fileChooser.showOpenDialog( getFrame() )
		 == JFileChooser.APPROVE_OPTION ) {

		properties.setProperty(
		    MyProperties.SAVE_DIRECTORY,
		    fileChooser.getCurrentDirectory().getPath() );
		if ( gameDiscarded() ) {
		    File file = fileChooser.getSelectedFile();
		    chessGame.loadGame( file );
		    //		  Thread loader = new FileLoader( f );
		    //		  loader.start();
		}
	    }
	}
    }

    class SaveAction extends ChessAbstractAction {

	public SaveAction() {
	    super( "save" );
	}

	public void actionPerformed( ActionEvent e ) {
	    boolean unused = saveGame( false );
	}

	void enable() {
	    this.setEnabled( chessGame.hasChanged() );
	}
    }

    class SaveAsAction extends MyAbstractAction {

	public SaveAsAction() {
	    super( "save-as" );
	    this.setEnabled( true );
	}

	public void actionPerformed( ActionEvent e ) {
	    boolean unused = saveGame( true );
	}
    }

    class PreferencesAction extends MyAbstractAction {

	public PreferencesAction() {
	    super( "preferences" );
	    this.setEnabled( true );
	}

	public void actionPerformed( ActionEvent e ) {
	    boolean usused = Preferences.showPreferences(
		resources, properties, getFrame() );
	}
    }

    class QuitAction extends NewAction {

	QuitAction() {
	    super( "quit" );
	}

	public void actionPerformed( ActionEvent e ) {
	    quit();
	}
    }

    //--------------------------------------------------------------------------
    // Clipboard actions
    //--------------------------------------------------------------------------

    class CopyAction extends MyAbstractAction {

	public CopyAction() {
	    super( "copy-to-clipboard" );
	    this.setEnabled( true );
	}

	public void actionPerformed( ActionEvent e ) {
	    copyPaste.copy();
	}
    }

    class PasteAction extends MyAbstractAction {

	public PasteAction() {
	    super( "paste-from-clipboard" );
	    this.setEnabled( true );
	}

	public void actionPerformed( ActionEvent e ) {
	    Transferable	tr = Toolkit.getDefaultToolkit().
		getSystemClipboard().getContents( null );
	    try {
		Reader		r = new StringReader(
		    (String)tr.getTransferData( DataFlavor.stringFlavor ) );
		chessGame.loadGame( r );
	    } catch ( Exception ex ) {
		Error.errorMessage(
		    "paste-error-title",
		    "unable-to-paste",
		    "" );
	    }
	}
    }

    class CopyBoardForsytheAction extends MyAbstractAction {

	public CopyBoardForsytheAction() {
	    super( "copy-board-forsythe" );
	    this.setEnabled( true );
	}

	public void actionPerformed( ActionEvent e ) {
	    copyPaste.copy( "board-forsythe" );
	}
    }

    class CopyBoardDiagramAction extends MyAbstractAction {

	public CopyBoardDiagramAction() {
	    super( "copy-board-diagram" );
	    this.setEnabled( true );
	}

	public void actionPerformed( ActionEvent e ) {
	    copyPaste.copy( "board-diagram" );
	}
    }

    class CopyBoardBothAction extends MyAbstractAction {

	public CopyBoardBothAction() {
	    super( "copy-board-both" );
	    this.setEnabled( true );
	}

	public void actionPerformed( ActionEvent e ) {
	    copyPaste.copy( "board-diagram;board-forsythe" );
	}
    }

    class CopyBoard1EchecsFontAction extends MyAbstractAction {

	public CopyBoard1EchecsFontAction() {
	    super( "copy-board-1echecs" );
	    this.setEnabled( true );
	}

	public void actionPerformed( ActionEvent e ) {
	    copyPaste.copy( "board-true-type" );
	}
    }

    class CopyFinalBoardForsytheAction extends MyAbstractAction {

	public CopyFinalBoardForsytheAction() {
	    super( "copy-final-board-forsythe" );
	    this.setEnabled( true );
	}

	public void actionPerformed( ActionEvent e ) {
	    copyPaste.copy( "final-board-forsythe" );
	}
    }

    class CopyFinalBoardDiagramAction extends MyAbstractAction {

	public CopyFinalBoardDiagramAction() {
	    super( "copy-final-board-diagram" );
	    this.setEnabled( true );
	}

	public void actionPerformed( ActionEvent e ) {
	    copyPaste.copy( "final-board-diagram" );
	}
    }

    class CopyFinalBoardBothAction extends MyAbstractAction {

	public CopyFinalBoardBothAction() {
	    super( "copy-final-board-both" );
	    this.setEnabled( true );
	}

	public void actionPerformed( ActionEvent e ) {
	    copyPaste.copy( "final-board-diagram;final-board-forsythe" );
	}
    }

    class CopyFinalBoard1EchecsFontAction extends MyAbstractAction {

	public CopyFinalBoard1EchecsFontAction() {
	    super( "copy-final-board-1echecs" );
	    this.setEnabled( true );
	}

	public void actionPerformed( ActionEvent e ) {
	    copyPaste.copy( "final-board-true-type" );
	}
    }

    class CopyMovesShortAction extends MyAbstractAction {

	public CopyMovesShortAction() {
	    super( "copy-moves-short" );
	    this.setEnabled( true );
	}

	public void actionPerformed( ActionEvent e ) {
	    copyPaste.copy( "moves-short" );
	}
    }

    class CopyMovesLongAction extends MyAbstractAction {

	public CopyMovesLongAction() {
	    super( "copy-moves-long" );
	    this.setEnabled( true );
	}

	public void actionPerformed( ActionEvent e ) {
	    copyPaste.copy( "moves-long" );
	}
    }

    class CopyMovesReverseAction extends MyAbstractAction {

	public CopyMovesReverseAction() {
	    super( "copy-moves-reverse" );
	    this.setEnabled( true );
	}

	public void actionPerformed( ActionEvent e ) {
	    copyPaste.copy( "moves-reverse" );
	}
    }

    //--------------------------------------------------------------------------
    // Final board actions
    //--------------------------------------------------------------------------

    class ClearFinalBoardAction extends MyAbstractAction {
	public ClearFinalBoardAction() {
	    super( "clear-final-board" );
	    this.setEnabled( true );
	}

	public void actionPerformed( ActionEvent e ) {
	    chessGame.clearFinalBoard();
	}
    }

    class SetInitialPositionAction extends MyAbstractAction {
	public SetInitialPositionAction() {
	    super( "set-initial-position" );
	    this.setEnabled( true );
	}

	public void actionPerformed( ActionEvent e ) {
	    chessGame.setInitialPositionInFinalBoard();
	}
    }

    class UseFinalPositionAsInitialPosition extends MyAbstractAction {
	public UseFinalPositionAsInitialPosition() {
	    super( "use-as-initial-position" );
	    this.setEnabled( true );
	}

	public void actionPerformed( ActionEvent e ) {
	    chessGame.useFinalPositionAsInitialPosition();
	}
    }

    class GetPositionOfGameAction extends MyAbstractAction {
	public GetPositionOfGameAction() {
	    super( "get-position-of-game" );
	    this.setEnabled( true );
	}

	public void actionPerformed( ActionEvent e ) {
	    chessGame.copyPositionOfGameToFinalPosition();
	}
    }

    class SetOrthodoxBoardAction extends MyAbstractAction {
	public SetOrthodoxBoardAction() {
	    super( "set-orthodox-board" );
	    this.setEnabled( true );
	}

	public void actionPerformed( ActionEvent e ) {
	    chessGame.setBoard( new StandardBoard() );
	}
    }

    class SetVerticalCylinderBoardAction extends MyAbstractAction {
	public SetVerticalCylinderBoardAction() {
	    super( "set-vertical-cylinder-board" );
	    this.setEnabled( true );
	}

	public void actionPerformed( ActionEvent e ) {
	    chessGame.setBoard( new VerticalCylinderBoard() );
	}
    }

    class SetOrhthodoxConditionAction extends MyAbstractAction {
	public SetOrhthodoxConditionAction() {
	    super( "set-orthodox-condition" );
	    this.setEnabled( true );
	}

	public void actionPerformed( ActionEvent e ) {
	    chessGame.setCondition( new OrthodoxCondition() );
	}
    }

    class SetMonochromaticConditionAction extends MyAbstractAction {
	public SetMonochromaticConditionAction() {
	    super( "set-monochromatic-condition" );
	    this.setEnabled( true );
	}

	public void actionPerformed( ActionEvent e ) {
	    chessGame.setCondition( new MonochromaticCondition() );
	}
    }

    class SetEinsteinConditionAction extends MyAbstractAction {
	public SetEinsteinConditionAction() {
	    super( "set-einstein-condition" );
	    this.setEnabled( true );
	}

	public void actionPerformed( ActionEvent e ) {
	    chessGame.setCondition( new EinsteinCondition() );
	}
    }

    //--------------------------------------------------------------------------
    // Natch actions
    //--------------------------------------------------------------------------

    class VerifyAction extends ChessAbstractAction {
	public VerifyAction() {
	    super( "verify" );
	}

	public void actionPerformed( ActionEvent e ) {
	    int		nb = chessGame.currentNumberOfMoves();
	    Position	position;
	    if ( nb > 0 ) {
		position = chessGame.position();
	    } else {
		position = chessGame.finalPosition();
		nb = chessGame.finalNumberOfMoves();
	    }
	    java.util.List comments
		= Natch.execute( properties, position , nb );
	    if ( comments != null ) {
		chessGame.setComments( comments );
	    }
	}

	void enable() {
	    Position position = chessGame.position();
	    setEnabled( "orthodox-condition".equals( position.conditionName() )
			&& "standard".equals( position.board().name() ) );
	}
    }

    //--------------------------------------------------------------------------
    //--------------------------------------------------------------------------
    //--------------------------------------------------------------------------
    //--------------------------------------------------------------------------
    //--------------------------------------------------------------------------
    //--------------------------------------------------------------------------
    //--------------------------------------------------------------------------
    //--------------------------------------------------------------------------

    class UndoAction extends AbstractAction {
	public UndoAction() {
	    super( "Undo" );
	    this.setEnabled( false );
	}

	public void actionPerformed( ActionEvent e ) {
	    /*
	      try {
	      undo.undo();
	      } catch ( CannotUndoException ex ) {
	      System .out .println( "Unable to undo: " + ex );
	      ex.printStackTrace();
	      }
	      update();
	      redoAction.update();
	    */
	}

	protected void update() {
	    /*
	      if ( undo.canUndo() ) {
	      setEnabled( true );
	      putValue( Action.NAME, undo.getUndoPresentationName() );
	      }
	      else {
	      setEnabled( false );
	      putValue( Action.NAME, "Undo" );
	      }
	    */
	}
    }

    class RedoAction extends AbstractAction {
	public RedoAction() {
	    super( "Redo" );
	    this.setEnabled( false );
	}

	public void actionPerformed( ActionEvent e ) {
	    /*
	      try {
	      undo.redo();
	      } catch ( CannotRedoException ex ) {
	      System .out .println( "Unable to redo: " + ex );
	      ex.printStackTrace();
	      }
	      update();
	      undoAction.update();
	    */
	}

	protected void update() {
	    /*
	      if ( undo.canRedo() ) {
	      setEnabled( true );
	      putValue( Action.NAME, undo.getRedoPresentationName() );
	      }
	      else {
	      setEnabled( false );
	      putValue( Action.NAME, "Redo" );
	      }
	    */
	}
    }




    /**
     * Thread to load a file into the text storage model
     */
    class FileLoader extends Thread {

	FileLoader( File f ) {
	    setPriority( 4 );
	    file = f;
	}

	public void run() {
	    JComponent status = null; // This is the status bar
	    try {
		PlainDocument doc = new PlainDocument();
		// initialize the statusbar
		status.removeAll();
		JProgressBar progress = new JProgressBar();
		progress.setMinimum( 0 );
		progress.setMaximum( (int)file.length() );
		status.add( progress );
		status.revalidate();

		// try to start reading
		Reader in = new FileReader( file );
		char[] buff = new char[4096];
		int nch;
		while ( ( nch = in.read( buff, 0, buff.length ) ) != -1 ) {
		    doc.insertString( doc.getLength(),
				      new String( buff, 0, nch ), null );
		    progress.setValue( progress.getValue() + nch );
		}

		// we are done... get rid of progressbar
		/*
		  doc.addUndoableEditListener( undoHandler );
		*/
		status.removeAll();
		status.revalidate();

		resetUndoManager();
	    } catch ( IOException e ) {
		System.err.println( e.toString() );
	    } catch ( BadLocationException e ) {
		System.err.println( e.getMessage() );
	    }
	}

	private File	file;
    }

    //--------------------------------------------------------------------------

    private MyProperties	properties;
    private Resources		resources;
    private String		applicationName;
    private JFrame		frame;

    // Must be static or the initialisation of all actions accessing chessGame
    // must receive it as a parameter.
    private static ChessGame	chessGame ;
    private CopyPaste		copyPaste ;

    private FileDialog		fileDialog  ;
    private JFileChooser	fileChooser ;

    //--------------------------------------------------------------------------

    private static final String DEFAULT_PROPERTIES_FILE
	= "/iNatch/defaultProperties.txt";
    private static final String PROPERTIES_FILE = "iNatch.ini";
    private static final String GUI_FILE
	= "iNatch/resources/iNatch.gui";
    private static final String RESOURCE_FILE
	= "iNatch.resources.iNatch";

    private static final String CURRENT_VERSION  = "v0.6.2";
    private static final String COPYRIGHT
	= "Copyright (c) 2001, 2002, 2003, 2004 Pascal Wassong";
}
