/*
 * MainDiagram.java --- Class to deal with the main diagram.
 *
 * Copyright (c) 2001, 2002, 2003 by Pascal Wassong All Rights Reserved.
 *
 * This file is part of iNatch.
 *
 * iNatch is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * iNatch is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

package iNatch;

import java.awt.Font;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.event.InputEvent;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;

import wassong.gui.tools.Resources;


public class FinalDiagram extends Diagram
    implements KeyListener, MouseListener, MouseMotionListener {

    //--------------------------------------------------------------------------
    // Constructor
    //--------------------------------------------------------------------------

    public FinalDiagram( Resources	resources,
			 MyProperties	properties,
			 ChessGame	chessGame ) {
	super( resources, properties, chessGame );

	this.addKeyListener  ( this );
	this.addMouseListener( this );
	this.addMouseMotionListener( this );
    }

    public void paintComponent( Graphics g ) {

	super.paintComponent( g );

	Graphics2D	g2d = (Graphics2D)g;

	for ( int i = 0; i < 6; i++ ) {
	    g2d.drawImage(
		IMAGES[ i ],
		DX_PIECES ,
		DY_PIECES + ( i * INCY_PIECES ),
		this );
	    g2d.drawImage(
		IMAGES[ i + 6 ],
		DX_PIECES + INCX_PIECES,
		DY_PIECES + ( i * INCY_PIECES ),
		this );
	}

	String s;
	if ( currentSquareX != -1 && currentSquareY != -1 ) {
	    g2d.drawRect( DX_DIAGRAM + currentSquareX * FONT_SIZE ,
			  DY_DIAGRAM + ( 7 - currentSquareY ) * FONT_SIZE ,
			  FONT_SIZE - 1, FONT_SIZE - 1 );
	}
    }

    //--------------------------------------------------------------------------
    // Key events
    //--------------------------------------------------------------------------

    public void keyTyped( KeyEvent e ) {

	char	c	  = e.getKeyChar();
	long	timestamp = e.getWhen();

	if ( c >= KeyEvent.VK_0 && c <= KeyEvent.VK_9 ) {
	    int newValue;
	    if ( timestamp - timestampLastKeyTyped > DELAY_KEY_TYPED ) {
		newValue = c - KeyEvent.VK_0;
	    } else {
		newValue = ( chessGame().finalNumberOfMoves() % 100 ) * 10
		    + c - KeyEvent.VK_0;
	    }
	    chessGame().setFinalNumberOfMoves( newValue );

	} else if ( currentSquareX != -1 && currentSquareY != -1 ) {
	    if ( c == KeyEvent.VK_SPACE ) {
		chessGame().clearSquare( currentSquareX, currentSquareY );
	    } else {
		int colour = Character.isUpperCase( c )
		    ? Piece.WHITE : Piece.BLACK;
		if ( squareXLastKey == currentSquareX
		     && squareYLastKey == currentSquareY
		     && ( timestamp - timestampLastKeyTyped
			  < DELAY_KEY_TYPED ) ) {

		    colour = Character.isUpperCase( c )
			? Piece.BLACK : Piece.WHITE;
		}
		int index = Piece.typeOf( c );
		if ( index != -1 ) {
		    Piece piece = new Piece(
			index,
			colour,
			currentSquareX,
			currentSquareY );
		    chessGame().addPieceAt( piece );
		}
	    }
	}
	timestampLastKeyTyped = timestamp;
	squareXLastKey = currentSquareX;
	squareYLastKey = currentSquareY;
    }

    public void keyPressed( KeyEvent e ) {
    }

    public void keyReleased( KeyEvent e ) {
	switch( e.getKeyCode() ) {
	case KeyEvent.VK_UP:
	    modifyCurrentSquare( 0, 1 );
	    break;
	case KeyEvent.VK_DOWN:
	    modifyCurrentSquare( 0, -1 );
	    break;
	case KeyEvent.VK_LEFT:
	    modifyCurrentSquare( -1, 0 );
	    break;
	case KeyEvent.VK_RIGHT:
	    modifyCurrentSquare( 1, 0 );
	    break;
	default:
	    break;
	}
    }

    private void modifyCurrentSquare( int deltaX, int deltaY ) {
	int	x = currentSquareX;
	int	y = currentSquareY;
	if ( x == -1 || y == -1 ) {
	    x = lastSquareX;
	    y = lastSquareY;
	}

	x += deltaX;
	y += deltaY;

	if ( x < 0 ) {
	    x = 7;
	} else if ( x > 7 ) {
	    x = 0;
	}
	if ( y < 0 ) {
	    y = 7;
	} else if ( y > 7 ) {
	    y = 0;
	}
	setCurrentSquare( x, y );
    }

    //--------------------------------------------------------------------------
    // Mouse events
    //--------------------------------------------------------------------------

    public void mouseClicked( MouseEvent e ) {
	if ( e.getClickCount() == 2 ) {
	    int	x = e.getX() - insetLeft();
	    int	y = e.getY() - insetTop();

	    if ( x >= DX_DIAGRAM && x < DX_DIAGRAM + 8 * FONT_SIZE
		 && y >= DY_DIAGRAM && y < DY_DIAGRAM + 8 * FONT_SIZE ) {

		x = ( x - DX_DIAGRAM ) / FONT_SIZE;
		y = 7 - ( y - DY_DIAGRAM ) / FONT_SIZE;
		chessGame().clearSquare( x, y );
	    }
	}
    }

    public void mouseEntered( MouseEvent e ) {
	requestFocus();
    }

    public void mouseExited( MouseEvent e ) {
	setCurrentSquare( -1, -1 );
    }

    public void mousePressed( MouseEvent e ) {
	if ( e.getClickCount() > 1 ) {
	    return;
	}

	int	x = e.getX() - insetLeft();
	int	y = e.getY() - insetTop();

	if ( x >= DX_PIECES && x < DX_PIECES + 2 * INCX_PIECES
	     && y >= DY_PIECES && y < DY_PIECES + 6 * INCY_PIECES ) {

	    x = ( x - DX_PIECES ) / INCX_PIECES;
	    y = ( y - DY_PIECES ) / INCY_PIECES;

	    setTheCursor( x * 6 + y );

	} else
	if ( x >= DX_DIAGRAM && x < DX_DIAGRAM + 8 * FONT_SIZE
	     && y >= DY_DIAGRAM && y < DY_DIAGRAM + 8 * FONT_SIZE ) {

	    x = ( x - DX_DIAGRAM ) / FONT_SIZE;
	    y = 7 - ( y - DY_DIAGRAM ) / FONT_SIZE;

	    int[][]	chessboard = chessGame().finalPosition().chessboard();
	    int		index      = chessboard[ x ][ y ];
	    if ( index != Position.EMPTY_SQUARE ) {
		Piece	p = chessGame().finalPosition().piece( index );
		int	pieceIndex = ( p.colour() == Piece.WHITE ? 0 : 6 )
		    + p.type();

		if ( ( e.getModifiers() & ( InputEvent.SHIFT_MASK
					    | InputEvent.CTRL_MASK
					    | InputEvent.ALT_MASK
					    | InputEvent.ALT_GRAPH_MASK
					    | InputEvent.META_MASK ) ) != 0 ) {
		    chessGame().clearSquare( x, y );
		}
		setTheCursor( pieceIndex );
	    }
	}
    }

    public void mouseReleased( MouseEvent e ) {
	if ( chessCursorType() != -1 ) {
	    int	x = e.getX() - insetLeft();
	    int	y = e.getY() - insetTop();

	    if ( x >= DX_DIAGRAM && x < DX_DIAGRAM + 8 * FONT_SIZE
		 && y >= DY_DIAGRAM && y < DY_DIAGRAM + 8 * FONT_SIZE ) {

		x = ( x - DX_DIAGRAM ) / FONT_SIZE;
		y = 7 - ( y - DY_DIAGRAM ) / FONT_SIZE;

		Piece	piece = new Piece(
		    chessCursorType() % 6 ,
		    chessCursorType() < 6 ? Piece.WHITE : Piece.BLACK,
		    x, y );
		chessGame().addPieceAt( piece );
	    }

	    setTheCursor( -1 );
	}
    }

    public void mouseDragged( MouseEvent e ) {
	watchMousePosition( e );
    }

    public void mouseMoved( MouseEvent e ) {
	watchMousePosition( e );
    }

    private void watchMousePosition( MouseEvent e ) {
	int	x = e.getX() - insetLeft();
	int	y = e.getY() - insetTop();

	if ( x >= DX_DIAGRAM && x < DX_DIAGRAM + 8 * FONT_SIZE
	     && y >= DY_DIAGRAM && y < DY_DIAGRAM + 8 * FONT_SIZE ) {

	    setCurrentSquare( ( x - DX_DIAGRAM ) / FONT_SIZE ,
			      7 - ( y - DY_DIAGRAM ) / FONT_SIZE );
	} else {
	    setCurrentSquare( -1, -1 );
	}
    }

    private void setCurrentSquare( int x, int y ) {
	if ( ( x != currentSquareX ) || ( y != currentSquareY ) ) {
	    if ( currentSquareX != -1 && currentSquareY != -1 ) {
		lastSquareX = currentSquareX;
		lastSquareY = currentSquareY;

		repaint( DX_DIAGRAM + currentSquareX * FONT_SIZE ,
			 DY_DIAGRAM + ( 7 - currentSquareY ) * FONT_SIZE ,
			 FONT_SIZE, FONT_SIZE );
	    }
	    currentSquareX = x;
	    currentSquareY = y;
	    if ( currentSquareX != -1 && currentSquareY != -1 ) {
		repaint( DX_DIAGRAM + currentSquareX * FONT_SIZE ,
			 DY_DIAGRAM + ( 7 - currentSquareY ) * FONT_SIZE ,
			 FONT_SIZE, FONT_SIZE );
	    }
	}
    }

    //--------------------------------------------------------------------------
    // Attributes
    //--------------------------------------------------------------------------

    private static final long   DELAY_KEY_TYPED = 1000;
    private long		timestampLastKeyTyped = 0;
    private int			squareXLastKey = -1;
    private int			squareYLastKey = -1;

    private int			currentSquareX = -1 ;
    private int			currentSquareY = -1 ;
    private int			lastSquareX = 7 ;
    private int			lastSquareY = 0 ;

    private static final int	DX_PIECES   = DX_DIAGRAM + 8 * FONT_SIZE + 10 ;
    private static final int	DY_PIECES   = FONT_SIZE / 2 ;
    private static final int	INCX_PIECES = FONT_SIZE ;
    private static final int	INCY_PIECES = FONT_SIZE ;

    private static final int	SMALL_FONT_SIZE = 10 ;
    private static final Font	SMALL_FONT      = new Font(
	"Arial", Font.PLAIN, SMALL_FONT_SIZE );

}
