/*
 * MainDiagram.java --- Class to deal with the main diagram.
 *
 * Copyright (c) 2001, 2002, 2003, 2004 by Pascal Wassong All Rights Reserved.
 *
 * This file is part of iNatch.
 *
 * iNatch is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * iNatch is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

package iNatch;

import java.awt.Color;
import java.awt.Font;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.event.ActionEvent;
import java.awt.event.InputEvent;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import javax.swing.AbstractAction;
import javax.swing.JMenuItem;
import javax.swing.JPopupMenu;

import wassong.gui.tools.MyPropertiesListener;
import wassong.gui.tools.Resources;


public class MainDiagram extends Diagram
    implements KeyListener, MouseListener, MyPropertiesListener {

    //--------------------------------------------------------------------------
    // Constructor
    //--------------------------------------------------------------------------

    public MainDiagram( Resources	theResources,
			MyProperties	theProperties,
			ChessGame	chessGame ) {
	super( theResources, theProperties, chessGame );

	this.properties = theProperties;

	this.addKeyListener  ( this );
	this.addMouseListener( this );
	properties.addPropertyListener( this, MyProperties.OUTPUT_LANGUAGE );
    }

    public void paintComponent( Graphics g ) {

	super.paintComponent( g );

	Graphics2D	g2d = (Graphics2D)g;

	g2d.setFont( NORMAL_FONT );
	g2d.drawString( chessGame().nbMovesToString(
			    chessGame().totalNumberOfMoves() ),
			DX_PIECE_NUMBERS ,
			DY_PIECE_NUMBERS );

	g2d.setFont( SMALL_FONT );

	int	currentNumberOfMoves = chessGame().currentNumberOfMoves();
	int	colour = ( currentNumberOfMoves & 1 ) == 0
	    ? Piece.WHITE
	    : Piece.BLACK;
	Move	previousMove = ( currentNumberOfMoves > 0 )
	    ? chessGame().move( currentNumberOfMoves - 1 )
	    : null;
	moves = chessGame().position().moves( colour, previousMove );
	int	x = DX_MOVES;
	int	y = DY_MOVES + INCY_MOVES;
	for ( int i = 0; i < moves.length; i++ ) {
	    g2d.drawString( moves[ i ].toBeautifullString(), x, y );
	    if ( ( i + 1 ) % NUMBER_OF_MOVES_PER_COLUMN == 0 ) {
		x += INCX_MOVES;
		y = DY_MOVES + INCY_MOVES;
	    } else {
		y += INCY_MOVES;
	    }
	}

	boolean useColours = properties.getBooleanProperty(
	    MyProperties.MISC_USE_SPG_HEURISTICS );
	for ( int i = 0; i < nbDestinationSquares; i++ ) {
	    if ( useColours ) {
		switch ( destinationSquares[ i ].status() ) {
		case SPGMove.MOVE_OLD:
		    g2d.setColor( ORANGE );
		    break;
		case SPGMove.MOVE_BAD:
		    g2d.setColor( RED );
		    break;
		default:
		    g2d.setColor( GREEN );
		    break;
		}
	    } else {
		g2d.setColor( GREEN );
	    }
	    g2d.fillOval(
		DX_DIAGRAM + destinationSquares[ i ].x() * FONT_SIZE
		+ FONT_SIZE / 3 + 1,
		DY_DIAGRAM + ( 7 - destinationSquares[ i ].y() ) * FONT_SIZE
		+ FONT_SIZE / 3,
		FONT_SIZE / 3,
		FONT_SIZE / 3 );
	}
	g2d.setColor( Color.black );
    }

    //--------------------------------------------------------------------------
    // Overloaded functions
    //--------------------------------------------------------------------------
    public Position position() {
	return chessGame().position();
    }

    public int numberOfMoves() {
	return chessGame().currentNumberOfMoves();
    }

    //--------------------------------------------------------------------------
    // MyPropertiesListener
    //--------------------------------------------------------------------------

    public void propertyChanged( String property ) {
	repaint();
    }

    //--------------------------------------------------------------------------
    // Key events
    //--------------------------------------------------------------------------

    public void keyTyped( KeyEvent e ) {
	switch( e.getKeyChar() ) {
	case KeyEvent.VK_BACK_SPACE:
	    chessGame().backwardMove();
	    break;
	case KeyEvent.VK_SPACE:
	    chessGame().forwardMove();
	    break;
	default:
	    // Nothing to do
	    break;
	}
    }

    public void keyPressed( KeyEvent e ) {
    }

    public void keyReleased( KeyEvent e ) {
    }

    //--------------------------------------------------------------------------
    // Mouse events
    //--------------------------------------------------------------------------

    public void mouseClicked( MouseEvent e ) {

	if ( hasMouseReleasedActed ) {
	    hasMouseReleasedActed = false;
	    return;
	}

	int	x = e.getX() - insetLeft();
	int	y = e.getY() - insetTop();

	if ( x >= DX_DIAGRAM && x <= DX_DIAGRAM + 8 * FONT_SIZE
	     && y >= DY_DIAGRAM && y <= DY_DIAGRAM + 8 * FONT_SIZE ) {

	    x = ( x - DX_DIAGRAM ) / FONT_SIZE;
	    y = 7 - ( y - DY_DIAGRAM ) / FONT_SIZE;

	    if ( ( e.getModifiers() & InputEvent.SHIFT_MASK ) == 0 ) {
		if ( System.currentTimeMillis() - mousePressedTimeStamp
		     < 500 ) {
		    detectMoveAndPlay( x, y, e );
		}
	    } else {
		replaceMovePopup( x, y, e );
	    }
	} else if ( x >= DX_MOVES && y >= DY_MOVES ) {
	    x = ( x - DX_MOVES ) / INCX_MOVES;
	    y = ( y - DY_MOVES ) / INCY_MOVES;
	    int	index = x * NUMBER_OF_MOVES_PER_COLUMN + y;
	    if ( index >= 0 && index < moves.length ) {
		chessGame().play( moves[ index ] );
	    }
	}
    }

    public void mouseEntered( MouseEvent e ) {
	requestFocus();
    }
    public void mouseExited( MouseEvent e ) {
    }

    public void mousePressed( MouseEvent e ) {
	mousePressedTimeStamp = System.currentTimeMillis();

	if ( e.isPopupTrigger() ) {
	    showLegalMovesPopup( e );
	    return;
	}

	if ( e.getClickCount() > 1
	     || ( e.getModifiers() & InputEvent.BUTTON1_MASK ) == 0 ) {
	    return;
	}

	int	x = e.getX() - insetLeft();
	int	y = e.getY() - insetTop();

	if ( x >= DX_DIAGRAM && x <= DX_DIAGRAM + 8 * FONT_SIZE
	     && y >= DY_DIAGRAM && y <= DY_DIAGRAM + 8 * FONT_SIZE ) {

	    x = ( x - DX_DIAGRAM ) / FONT_SIZE;
	    y = 7 - ( y - DY_DIAGRAM ) / FONT_SIZE;

	    int		index  = chessGame().position().chessboard()[ x ][ y ];
	    if ( index == Position.EMPTY_SQUARE ) {
		return;
	    }
	    Piece	piece = chessGame().position().piece( index );
	    int		colour = ( chessGame().currentNumberOfMoves() & 1 ) == 0
		? Piece.WHITE
		: Piece.BLACK;
	    if ( colour != piece.colour() ) {
		return;
	    }

	    nbDestinationSquares = 0;
	    for ( int i = 0; i < moves.length; i++ ) {
		Piece	p = moves[ i ].origin();
		if ( x == p.column() && y == p.row() ) {
		    nbDestinationSquares++;
		}
	    }
	    if ( nbDestinationSquares > 0 ) {
		destinationSquares =
		    new DestinationSquare[ nbDestinationSquares ];
		Move[]	possibleMoves = new Move[ nbDestinationSquares ];
		nbDestinationSquares = 0;
		for ( int i = 0; i < moves.length; i++ ) {
		    Piece p = moves[ i ].origin();
		    if ( x == p.column() && y == p.row() ) {
			possibleMoves[ nbDestinationSquares ] = moves[ i ];
			p = moves[ i ].destination();
			destinationSquares[ nbDestinationSquares ] =
			    new DestinationSquare( p.column(), p.row(), i );
			nbDestinationSquares++;
		    }
		}

		int		nbMoves = chessGame().currentNumberOfMoves();
		SPGMove[]	orderedMoves = SPGMove.orderMoves(
		    possibleMoves,
		    nbDestinationSquares,
		    chessGame().position(),
		    nbMoves > 0 ? chessGame().move( nbMoves - 1 ) : null,
		    nbMoves > 1 ? chessGame().move( nbMoves - 2 ) : null );
		for ( int i = 0; i < nbDestinationSquares; i++ ) {
		    Piece	p = orderedMoves[ i ].destination();
		    for ( int j = 0; j < nbDestinationSquares; j++ ) {
			if ( p.column() == destinationSquares[ j ].x()
			     && p.row() == destinationSquares[ j ].y() ) {
			    destinationSquares[ j ].setStatus(
				orderedMoves[ i ].status() );
			    break;
			}
		    }
		}

		hideSquare( x, y );
		int	pieceIndex = ( colour == Piece.WHITE ? 0 : 6 )
		    + piece.type();
		setTheCursor( pieceIndex );
		repaint();	// Necessary
	    }
	}
    }

    public void mouseReleased( MouseEvent e ) {
	if ( chessCursorType() != -1 ) {
	    int	x = e.getX() - insetLeft();
	    int	y = e.getY() - insetTop();

	    hideSquare( -1, -1 );
	    setTheCursor( -1 );
	    repaint();		// Necessary if play() isn't called

	    if ( x >= DX_DIAGRAM && x <= DX_DIAGRAM + 8 * FONT_SIZE
		 && y >= DY_DIAGRAM && y <= DY_DIAGRAM + 8 * FONT_SIZE ) {

		x = ( x - DX_DIAGRAM ) / FONT_SIZE;
		y = 7 - ( y - DY_DIAGRAM ) / FONT_SIZE;

		Move[]	possibleMoves   = new Move[ nbDestinationSquares ];
		int	nbPossibleMoves = 0;
		for ( int i = 0; i < nbDestinationSquares; i++ ) {
		    if ( x == destinationSquares[ i ].x()
			 && y == destinationSquares[ i ].y() ) {

			possibleMoves[ nbPossibleMoves++ ] =
			    moves[ destinationSquares[ i ].index() ];
		    }
		}
		if ( nbPossibleMoves == 1 ) {
		    chessGame().play( possibleMoves[ 0 ] );
		    hasMouseReleasedActed = true;
		} else if ( nbPossibleMoves > 1 ) {
		    int	nbMoves = chessGame().currentNumberOfMoves();
		    SPGMove[]	orderedMoves = SPGMove.orderMoves(
			possibleMoves,
			nbPossibleMoves,
			chessGame().position(),
			chessGame().move( nbMoves - 1 ),
			chessGame().move( nbMoves - 2 ) );
		    showOrderedMovesAndPlay( e, "", orderedMoves );
		    hasMouseReleasedActed = true;
		}
	    }
	    nbDestinationSquares = 0;
	}

	if ( e.isPopupTrigger() ) {
	    showLegalMovesPopup( e );
	}
    }

    private void detectMoveAndPlay( int x, int y, MouseEvent e ) {
	int	index        = chessGame().position().chessboard()[ x ][ y ];
	boolean	sourceSquare = false;
	boolean	movePossible = false;
	int	colour = ( chessGame().currentNumberOfMoves() & 1 ) == 0
	    ? Piece.WHITE
	    : Piece.BLACK;
	if ( index != Position.EMPTY_SQUARE ) {
	    if ( colour == chessGame().position().piece( index ).colour() ) {
		sourceSquare = true;
	    }
	}

	Move	m = null;
	for ( int i = 0; i < moves.length; i++ ) {
	    Piece	p;
	    if ( sourceSquare ) {
		p = moves[ i ].origin();
	    } else {
		p = moves[ i ].destination();
	    }
	    if ( x == p.column() && y == p.row() ) {
		movePossible = true;
		if ( m == null ) {
		    m = moves[ i ];
		} else {
		    m = null;
		    break;
		}
	    }
	}
	if ( !movePossible ) {
	    return;
	}
	if ( m != null ) {
	    chessGame().play( m );
	} else {
	    if ( properties.getBooleanProperty(
		     MyProperties.MISC_USE_SPG_HEURISTICS ) ) {
		detectSpgMoveAndPlay( colour, sourceSquare, x, y, e );
	    } else {
		showLegalMovesPopup( e );
	    }
	}
    }

    private void detectSpgMoveAndPlay( int		colour,
				       boolean		sourceSquare,
				       int		x,
				       int		y,
				       MouseEvent	e ) {
	if ( chessGame().currentNumberOfMoves() <= 1 ) {
	    showLegalMovesPopup( e );
	    return;
	}

	Move[]	possibleMoves	= new Move[ moves.length ];
	int	nbPossibleMoves = 0;

	for ( int i = 0; i < moves.length; i++ ) {
	    Piece	p;
	    if ( sourceSquare ) {
		p = moves[ i ].origin();
	    } else {
		p = moves[ i ].destination();
	    }
	    if ( x == p.column() && y == p.row() ) {

		possibleMoves[ nbPossibleMoves++ ] = moves[ i ];
	    }
	}

	int		nbMoves      = chessGame().currentNumberOfMoves();
	SPGMove[]	orderedMoves =
	    SPGMove.orderMoves( possibleMoves,
				nbPossibleMoves,
				chessGame().position(),
				chessGame().move( nbMoves - 1 ),
				chessGame().move( nbMoves - 2 ) );

	// There should be at least 2 possible moves
	if ( orderedMoves[ 0 ].status() == SPGMove.MOVE_OK
	     && orderedMoves[ 1 ].status() != SPGMove.MOVE_OK ) {
	    chessGame().play( orderedMoves[ 0 ] );
	} else if ( orderedMoves[ 0 ].status() == SPGMove.MOVE_OLD
		    && orderedMoves[ 1 ].status() != SPGMove.MOVE_OLD ) {
	    chessGame().play( orderedMoves[ 0 ] );
	} else {
	    showOrderedMovesAndPlay( e,
				     String.valueOf( (char)( 'a' + x ) )
				     + ( y + 1 ),
				     orderedMoves );
	}
    }

    private void showLegalMovesPopup( MouseEvent e ) {
	int	x = e.getX() - insetLeft();
	int	y = e.getY() - insetTop();

	AbstractAction	aa = new AbstractAction() {
		public void actionPerformed( ActionEvent ae ) {
		    for ( int i = 0; i < moves.length; i++ ) {
			if ( moves[ i ].toBeautifullString().equals
			     ( ae.getActionCommand() ) ) {
			    chessGame().play( moves[ i ] );
			    break;
			}
		    }
		}
	    };

	if ( x >= DX_DIAGRAM && x <= DX_DIAGRAM + 8 * FONT_SIZE
	     && y >= DY_DIAGRAM && y <= DY_DIAGRAM + 8 * FONT_SIZE ) {

	    x = ( x - DX_DIAGRAM ) / FONT_SIZE;
	    y = 7 - ( y - DY_DIAGRAM ) / FONT_SIZE;

	    int		index  = chessGame().position().chessboard()[ x ][ y ];
	    boolean	sourceSquare = false;
	    if ( index != Position.EMPTY_SQUARE ) {
		int	colour = ( chessGame().currentNumberOfMoves() & 1 ) == 0
		    ? Piece.WHITE
		    : Piece.BLACK;
		if ( colour
		     == chessGame().position().piece( index ).colour() ) {
		    sourceSquare = true;
		}
	    }

	    boolean	popupEmpty = true;
	    JPopupMenu	popup      = new JPopupMenu(
		String.valueOf( (char)( 'a' + x ) ) + ( y + 1 ) );
	    for ( int i = 0; i < moves.length; i++ ) {
		Piece	p;
		if ( sourceSquare ) {
		    p = moves[ i ].origin();
		} else {
		    p = moves[ i ].destination();
		}
		if ( x == p.column() && y == p.row() ) {
		    JMenuItem   mi = new JMenuItem(
			moves[ i ].toBeautifullString() );
		    mi.addActionListener( aa );
		    popup.add( mi );
		    popupEmpty = false;
		}
	    }
	    if ( !popupEmpty ) {
		popup.show( e.getComponent(), e.getX(), e.getY() );
	    }
	} else if ( x >= DX_MOVES && y >= DY_MOVES ) {
	    x = ( x - DX_MOVES ) / INCX_MOVES;
	    y = ( y - DY_MOVES ) / INCY_MOVES;
	    int	index = x * NUMBER_OF_MOVES_PER_COLUMN + y;
	    if ( index < 0 || index >= moves.length ) {
		return;
	    }
	    x = moves[ index ].origin().column();
	    y = moves[ index ].origin().row();
	    JPopupMenu	popup  = new JPopupMenu(
		String.valueOf( (char)( 'a' + x ) ) + ( y + 1 ) );
	    boolean		popupEmpty = true;
	    for ( int i = 0; i < moves.length; i++ ) {
		Piece	p = moves[ i ].origin();
		if ( x == p.column() && y == p.row() ) {
		    JMenuItem   mi = new JMenuItem(
			moves[ i ].toBeautifullString() );
		    mi.addActionListener( aa );
		    popup.add( mi );
		    popupEmpty = false;
		}
	    }
	    if ( !popupEmpty ) {
		popup.show( e.getComponent(), e.getX(), e.getY() );
	    }
	}
    }

    private void replaceMovePopup( int x, int y, MouseEvent e ) {

	final Move[] alternateMoves = chessGame().alternateMoves( x, y );

	AbstractAction	aa = new AbstractAction() {
		public void actionPerformed( ActionEvent ae ) {
		    for ( int i = 0; i < moves.length; i++ ) {
			if ( alternateMoves[ i ].toBeautifullString().equals
			     ( ae.getActionCommand() ) ) {
			    chessGame().replaceMove( 25,
						   alternateMoves[ i ] );
			    //	    chessGame.play( alternateMoves[ i ] );
			    break;
			}
		    }
		}
	    };

	JPopupMenu	popup  = new JPopupMenu( "Replace move" );
	boolean		popupEmpty = true;
	for ( int i = 0; i < alternateMoves.length; i++ ) {
	    JMenuItem   mi = new JMenuItem(
		alternateMoves[ i ].toBeautifullString() );
		mi.addActionListener( aa );
		popup.add( mi );
		popupEmpty = false;
	}
	if ( !popupEmpty ) {
	    popup.show( e.getComponent(), e.getX(), e.getY() );
	}
    }

    //--------------------------------------------------------------------------
    // Status Move
    //--------------------------------------------------------------------------

    private void showOrderedMovesAndPlay(
	MouseEvent	e,
	String		title,
	SPGMove[]	orderedMoves ) {

	AbstractAction	aa = new AbstractAction() {
		public void actionPerformed( ActionEvent ae ) {
		    for ( int i = 0; i < moves.length; i++ ) {
			if ( moves[ i ].toBeautifullString().equals
			     ( ae.getActionCommand() ) ) {
			    chessGame().play( moves[ i ] );
			    break;
			}
		    }
		}
	    };

	JPopupMenu	popup	     = new JPopupMenu( title );
	boolean		newMoveAdded = false;
	boolean		oldMoveAdded = false;
	for ( int i = 0; i < orderedMoves.length; i++ ) {
	    JMenuItem   mi = new JMenuItem(
		orderedMoves[ i ].toBeautifullString() );
	    mi.addActionListener( aa );
	    switch ( orderedMoves[ i ].status() ) {
	    case SPGMove.MOVE_OK :
		mi.setForeground( GREEN );
		newMoveAdded = true;
		break;
	    case SPGMove.MOVE_OLD :
		if ( newMoveAdded ) {
		    popup.add( new JPopupMenu.Separator() );
		    newMoveAdded = false;
		}
		mi.setForeground( ORANGE );
		oldMoveAdded = true;
		break;
	    case SPGMove.MOVE_BAD :
		if ( newMoveAdded || oldMoveAdded ) {
		    popup.add( new JPopupMenu.Separator() );
		    newMoveAdded = false;
		    oldMoveAdded = false;
		}
		mi.setForeground( RED );
		break;
	    default:
		throw new UnsupportedOperationException();
	    }
	    popup.add( mi );
	}

	if ( orderedMoves.length > 0 ) {
	    popup.show( e.getComponent(), e.getX(), e.getY() );
	}
    }

    //--------------------------------------------------------------------------
    // Attributes
    //--------------------------------------------------------------------------

    private class DestinationSquare {
	public DestinationSquare( int destX, int destY, int destIndex ) {
	    x	   = destX;
	    y	   = destY;
	    index  = destIndex;
	    status = SPGMove.MOVE_OK;
	}

	public int x()      { return x; }
	public int y()      { return y; }
	public int index()  { return index; }
	public int status() { return status; }

	public void setStatus( int aStatus ) { status = aStatus; }

	private	int	x;
	private int	y;
	private int	index;
	private int	status;
    }

    //--------------------------------------------------------------------------
    // Attributes
    //--------------------------------------------------------------------------

    private MyProperties	properties;

    private Move[]		moves;

    private DestinationSquare[]	destinationSquares;
    private int			nbDestinationSquares = 0;

    // This is necessary for a case in Vertical Cylinder board where a piece can
    // play a null move.  If we click on that piece without moving the mouse,
    // mouseClicked should not play the piece that mouseReleased just moved.
    private boolean		hasMouseReleasedActed = false;
    private long		mousePressedTimeStamp;

    private static final int	SMALL_FONT_SIZE = 10;
    private static final Font	SMALL_FONT      = new Font(
	"Arial", Font.PLAIN, SMALL_FONT_SIZE );

    private static final int	DX_PIECE_NUMBERS =
	DX_DIAGRAM + 8 * FONT_SIZE + 5;
    private static final int	DY_PIECE_NUMBERS = 20;

    private static final int	DX_MOVES   = DX_PIECE_NUMBERS;
    private static final int	DY_MOVES   = DY_PIECE_NUMBERS;
    private static final int	INCX_MOVES = 50;
    private static final int	INCY_MOVES = SMALL_FONT_SIZE + 3;

    private static final int	NUMBER_OF_MOVES_PER_COLUMN =
	( FONT_SIZE * 9 - DY_MOVES ) / INCY_MOVES;

    private static final Color  GREEN  = new Color( 0, 192, 0 );
    private static final Color  ORANGE = new Color( 224, 175, 0 );
    private static final Color  RED    = Color.red;
}
