/*
 * MonochromaticCondition.java --- Monochromatic chess game condition.
 *
 * Copyright (c) 2001, 2002, 2003, 2004 by Pascal Wassong All Rights Reserved.
 *
 * This file is part of iNatch.
 *
 * iNatch is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * iNatch is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

package iNatch;

public class MonochromaticCondition implements Condition {

    public String name() {
	return "monochromatic-condition";
    }

    /**
     * Gives all the legal moves for the given colour.
     *
     * @param colour		Colour for which to give the moves.
     * @param previous_move	Used for potential en passant captures. null if
     *				no previous move.
     */
    public Move[] moves( Position position, int colour, Move previousMove ) {

	Move[]	moves   = new Move[ MAX_NUMBER_OF_MOVES ];
	int	nbMoves = 0;

	final int[][]	chessboard = position.chessboard();

	int	otherColour = ( colour == Piece.WHITE )
	    ? Piece.BLACK : Piece.WHITE;
	int	kingRow     = -1;
	int	kingColumn  = -1;

    king_search:
	for ( int column = 0; column < chessboard.length; column++ ) {
	    for ( int row = 0; row < chessboard[ 0 ].length; row++ ) {
		int	index = chessboard[ column ][ row ];
		if ( index != Position.EMPTY_SQUARE ) {
		    Piece p = position.piece( index );
		    if ( p.type() == Piece.KING && p.colour() == colour ) {
			kingRow    = row;
			kingColumn = column;
			break king_search;
		    }
		}
	    }
	}

	for ( int column = 0; column < chessboard.length; column++ ) {
	    for ( int row = 0; row < chessboard[ 0 ].length; row++ ) {
		int pieceIndex = chessboard[ column ][ row ];
		if ( pieceIndex == Position.EMPTY_SQUARE ) {
		    continue;
		}
		Piece p = position.piece( pieceIndex );
		if ( p.colour() != colour ) {
		    continue;
		}

		int type = p.type();

		switch ( type ) {
		case Piece.KING:
		{
		    int	castlingRow = ( colour == Piece.WHITE )
			? 0 : position.board().sizeY() - 1;
		    if ( row == castlingRow && column == 4
			 && numberOfControls( position,
					      4,
					      castlingRow,
					      otherColour ) == 0 ) {
			int rookIndex = chessboard[ 7 ][ castlingRow ];
			if ( rookIndex != Position.EMPTY_SQUARE
			     && position.piece( rookIndex ).type() == Piece.ROOK
			     && position.piece( rookIndex ).colour()
			     == colour ) {

			    int	fIndex = chessboard[ 5 ][ castlingRow ];
			    int	gIndex = chessboard[ 6 ][ castlingRow ];
			    if ( fIndex == Position.EMPTY_SQUARE
				 && gIndex == Position.EMPTY_SQUARE
				 && numberOfControls( position,
						      5,
						      castlingRow,
						      otherColour ) == 0 ) {
				Move m = new Move( Move.KING_CASTLING,
						   colour,
						   7,
						   Piece.ROOK );
				position.play( m );
				if ( numberOfControls( position,
						       6,
						       castlingRow,
						       otherColour ) == 0 ) {
				    moves[ nbMoves ] = m;
				    nbMoves++;
				}
				position.unplay( m );
			    }
			}
			// No Queen castling in Monochromatic
		    }

		    int[]	incX = { -1, -1,  1,  1 };
		    int[]	incY = { -1,  1, -1,  1 };
		    Board.Square newSquare = new Board.Square( -1, -1 );
		    for ( int k = 0; k < incX.length; k++ ) {
			newSquare = position.board().nextSquare(
			    column, row, incX[ k ], incY[ k ], newSquare );
			if ( newSquare == null ) {
			    continue;
			}
			int x = newSquare.x;
			int y = newSquare.y;
			int index = chessboard[ x ][ y ];
			if ( index != Position.EMPTY_SQUARE
			     && position.piece( index ).colour() == colour ) {
			    continue;
			}

			Move m;
			if ( index == Position.EMPTY_SQUARE ) {
			    m = new Move( position.piece( pieceIndex ),
					  new Piece( type, colour, x, y ) );
			} else {
			    m = new Move( position.piece( pieceIndex ),
					  new Piece( type, colour, x, y ),
					  position.piece( index ) );
			}
			position.play( m );
			if ( numberOfControls( position,
					       x,
					       y,
					       otherColour ) == 0 ) {
			    moves[ nbMoves ] = m;
			    nbMoves++;
			}
			position.unplay( m );
		    }
		}
		break;
		case Piece.PAWN:
		{
		    int	incY;
		    int	initialRow;
		    int	promotionRow;
		    int	enPassantRow;
		    if ( colour == Piece.WHITE ) {
			incY         = 1;
			initialRow   = 1;
			promotionRow = position.board().sizeY() - 1;
			enPassantRow = position.board().sizeY() - 4;
		    } else {
			incY         = -1;
			initialRow   = position.board().sizeY() - 2;
			promotionRow = 0;
			enPassantRow = 3;
		    }

		    if ( chessboard[ column ][ row + incY ]
			 == Position.EMPTY_SQUARE ) {

			// No single step pawn move in Monochromatic

			if ( row == initialRow
			     && ( chessboard[ column ][ row + incY + incY ]
				  == Position.EMPTY_SQUARE ) ) {
			    nbMoves = generatePawnMoves(
				position,
				position.piece( pieceIndex ),
				null,
				column,
				row + incY + incY,
				promotionRow,
				kingColumn,
				kingRow,
				otherColour,
				moves,
				nbMoves );
			}
		    }

		    // Left capture
		    Board.Square newLeftSquare = position.board().nextSquare(
			column, row, -1, incY, null );
		    if ( newLeftSquare != null ) {
			int index = chessboard[ newLeftSquare.x ]
			    [ newLeftSquare.y ];
			if ( index != Position.EMPTY_SQUARE
			     && position.piece( index ).colour() != colour ) {

			    nbMoves = generatePawnMoves(
				position,
				position.piece( pieceIndex ),
				position.piece( index ),
				newLeftSquare.x,
				newLeftSquare.y,
				promotionRow,
				kingColumn,
				kingRow,
				otherColour,
				moves,
				nbMoves );
			}
		    }

		    // Right capture
		    Board.Square newRightSquare = position.board().nextSquare(
			column, row, 1, incY, null );
		    if ( newRightSquare != null ) {
			int index = chessboard[ newRightSquare.x ]
			    [ newRightSquare.y ];
			if ( index != Position.EMPTY_SQUARE
			     && position.piece( index ).colour() != colour ) {

			    nbMoves = generatePawnMoves(
				position,
				position.piece( pieceIndex ),
				position.piece( index ),
				newRightSquare.x,
				newRightSquare.y,
				promotionRow,
				kingColumn,
				kingRow,
				otherColour,
				moves,
				nbMoves );
			}
		    }

		    // en passant capture
		    if ( enPassantRow == row
			 && previousMove != null
			 && !previousMove.isNullMove()
			 && previousMove.origin().type() == Piece.PAWN
			 && previousMove.origin().row() == promotionRow - incY
			 && previousMove.destination().row() == row ) {

			Board.Square square = null;
			if ( newRightSquare != null
			     && ( previousMove.destination().column()
				  == newRightSquare.x ) ) {
			    square = newRightSquare;
			} else if ( newLeftSquare != null
				    && ( previousMove.destination().column()
					 == newLeftSquare.x ) ) {
			    square = newLeftSquare;
			}

			if ( square != null ) {
			    int	indexPawn = chessboard[ square.x ][ row ];
			    int	indexEmpty
				= chessboard[ square.x ][ square.y ];
			    if ( indexEmpty == Position.EMPTY_SQUARE
				 && indexPawn != Position.EMPTY_SQUARE
				 && ( position.piece( indexPawn ).type()
				      == Piece.PAWN )
				 && ( position.piece( indexPawn ).colour()
				      != colour ) ) {

				nbMoves = generatePawnMoves(
				    position,
				    position.piece( pieceIndex ),
				    position.piece( indexPawn ),
				    square.x,
				    square.y,
				    promotionRow,
				    kingColumn,
				    kingRow,
				    otherColour,
				    moves,
				    nbMoves );
			    }
			}
		    }
		}
		break;
		case Piece.QUEEN:
		{
		    int[]	incX = { -1, -1, -1,  0,  0,  1,  1,  1 };
		    int[]	incY = { -1,  0,  1, -1,  1, -1,  0,  1 };

		    nbMoves = generateMoves(
			position,
			position.piece( pieceIndex ),
			kingColumn,
			kingRow,
			otherColour,
			incX,
			incY,
			moves,
			nbMoves );
		}
		break;
		case Piece.ROOK:
		{
		    int[]	incX = { -1,  0,  0,  1 };
		    int[]	incY = {  0, -1,  1,  0 };

		    nbMoves = generateMoves(
			position,
			position.piece( pieceIndex ),
			kingColumn,
			kingRow,
			otherColour,
			incX,
			incY,
			moves,
			nbMoves );
		}
		break;
		case Piece.BISHOP:
		{
		    int[]	incX = { -1, -1,  1,  1 };
		    int[]	incY = { -1,  1, -1,  1 };

		    nbMoves = generateMoves(
			position,
			position.piece( pieceIndex ),
			kingColumn,
			kingRow,
			otherColour,
			incX,
			incY,
			moves,
			nbMoves );
		}
		break;
		case Piece.KNIGHT:
		    // No knight moves in Monochromatic
		break;
		default:
		    throw new RuntimeException();
		}
	    }
	}

	Move[]	theMoves = new Move[ nbMoves ];
	for ( int i = 0; i < nbMoves; i++ ) {
	    theMoves[ i ] = moves[ i ];
	}
	return theMoves;
    }

    public int checkStatus( Position	position,
			    int		colour,
			    Move	previousMove ) {

	Move[]	moves = this.moves( position, colour, previousMove );
	if ( moves.length == 0 ) {
	    return Position.CHECK_MATE;
	}

	int[][]	chessboard = position.chessboard();
	Piece	king = null;
    king_search:
	for ( int column = 0; column < chessboard.length; column++ ) {
	    for ( int row = 0; row < chessboard[ 0 ].length; row++ ) {
		int index = chessboard[ column ][ row ];
		if ( index != Position.EMPTY_SQUARE ) {
		    Piece p = position.piece( index );
		    if ( p.type() == Piece.KING && p.colour() == colour ) {
			king = p;
			break king_search;
		    }
		}
	    }
	}
	colour = ( colour == Piece.WHITE ) ? Piece.BLACK : Piece.WHITE;

	int	check;
	switch( numberOfControls( position,
				  king.column(),
				  king.row(),
				  colour ) ) {
	case 0:
	    check = Position.NO_CHECK;
	    break;
	case 1:
	    check = Position.CHECK;
	    break;
	default:
	    check = Position.DOUBLE_CHECK;
	    break;
	}
	return check;
    }

    /**
     * Counts how many times a square is controlled by the given color.
     *
     * @param column	Column of the square to check.
     * @param row	Row of the square to check.
     * @param colour	Colour for which to verify if the square is
     *			controlled. Must be <code>Piece.WHITE</code>
     *			or <code>Piece.BLACK</code>.
     */
    private int numberOfControls( Position	position,
				  int		column,
				  int		row,
				  int		colour ) {

	int		nb	   = 0;
	int		evenSquare = ( column + row ) & 1;
	Board.Square	aSquare    = new Board.Square( -1, -1 );
	Board.Square	newSquare;

	// King
	int[]	incX = { -1, -1,  1,  1 };
	int[]	incY = { -1,  1, -1,  1 };
	for ( int i = 0; i < incX.length; i++ ) {
	    newSquare = position.board().nextSquare(
		column, row, incX[ i ], incY[ i ], aSquare );
	    if ( newSquare == null ) {
		continue;
	    }
	    int	index = position.chessboard()[ newSquare.x ][ newSquare.y ];
	    if ( index == Position.EMPTY_SQUARE ) {
		continue;
	    }
	    if ( position.piece( index ).type() == Piece.KING
		 && position.piece( index ).colour() == colour ) {
		nb++;
	    }
	}

	// Pawn
	int	pawnIncY = ( colour == Piece.WHITE ) ? -1 : 1;
	if ( ( colour == Piece.WHITE && row > 0 )
	     || ( colour == Piece.BLACK
		  && row < position.board().sizeY() - 1 ) ) {
	    newSquare = position.board().nextSquare(
		column, row, -1, pawnIncY, aSquare );
	    if ( newSquare != null ) {
		int index = position.chessboard()[ newSquare.x ][ newSquare.y ];
		if ( index != Position.EMPTY_SQUARE ) {
		    if ( position.piece( index ).type() == Piece.PAWN
			 && position.piece( index ).colour() == colour ) {
			nb++;
		    }
		}
	    }
	    newSquare = position.board().nextSquare(
		column, row, 1, pawnIncY, aSquare );
	    if ( newSquare != null ) {
		int index = position.chessboard()[ newSquare.x ][ newSquare.y ];
		if ( index != Position.EMPTY_SQUARE ) {
		    if ( position.piece( index ).type() == Piece.PAWN
			 && position.piece( index ).colour() == colour ) {
			nb++;
		    }
		}
	    }
	}

	// Orthogonal and Diagonal
	int[]	lineX = { 0,  0,  1, -1,  1,  1, -1, -1 };
	int[]	lineY = { 1, -1,  0,  0,  1, -1,  1, -1 };
	int	type  = Piece.ROOK;
	for ( int i = 0; i < lineX.length; i++ ) {
	    newSquare = position.board().nextSquare(
		column, row, lineX[ i ], lineY[ i ], aSquare );
	    if ( i == 4 ) {
		type = Piece.BISHOP;
	    }
	    while ( newSquare != null ) {
		int index = position.chessboard()[ newSquare.x ][ newSquare.y ];
		if ( index != Position.EMPTY_SQUARE ) {
		    if ( ( ( newSquare.x + newSquare.y ) & 1 ) == evenSquare
			 && position.piece( index ).colour() == colour
			 && ( position.piece( index ).type() == Piece.QUEEN
			      || position.piece( index ).type() == type ) ) {
			nb++;
		    }
		    break;
		}
		newSquare = position.board().nextSquare(
		    newSquare.x, newSquare.y, lineX[ i ], lineY[ i ], aSquare );
	    }
	}

	// No Knight moves in Monochromatic

	return nb;
    }

    private int generateMoves( Position	position,
			       Piece	piece,
			       int	kingColumn,
			       int	kingRow,
			       int	otherColour,
			       int[]	incX,
			       int[]	incY,
			       Move[]	moves,
			       int	nbMoves ) {
	int		column	   = piece.column();
	int		row	   = piece.row();
	int		type	   = piece.type();
	int		colour	   = piece.colour();
	int		evenSquare = ( column + row ) & 1;
	Board.Square	aSquare    = new Board.Square( -1, -1 );

	for ( int i = 0; i < incX.length; i++ ) {
	    for ( Board.Square newSquare = position.board().nextSquare(
		      column, row, incX[ i ], incY[ i ], aSquare );
		  newSquare != null;
		  newSquare = position.board().nextSquare(
		      newSquare.x, newSquare.y,
		      incX[ i ], incY[ i ],
		      aSquare ) ) {

		int	x     = newSquare.x;
		int	y     = newSquare.y;
		int	index = position.chessboard()[ x ][ y ];

		if ( ( ( x + y ) & 1 ) != evenSquare ) {
		    if ( index == position.EMPTY_SQUARE ) {
			continue;
		    } else {
			break;
		    }
		}

		if ( index != Position.EMPTY_SQUARE
		     && ( x != column || y != row )
		     && position.piece( index ).colour() == colour ) {
		    break;
		}

		Move m;
		if ( index == Position.EMPTY_SQUARE
		     || ( x == column && y == row ) ) {
		    m = new Move( piece, new Piece( type, colour, x, y ) );
		} else {
		    m = new Move( piece,
				  new Piece( type, colour, x, y ),
				  position.piece( index ) );
		}

		// Possible in vertical cylinder
		boolean	alreadyFound = false;
		for ( int j = 0; j < nbMoves; j++ ) {
		    if ( m.equals( moves[ j ] ) ) {
			alreadyFound = true;
			break;
		    }
		}
		if ( !alreadyFound ) {
		    position.play( m );
		    if ( numberOfControls( position,
					   kingColumn,
					   kingRow,
					   otherColour ) == 0 ) {
			moves[ nbMoves ] = m;
			nbMoves++;
		    }
		    position.unplay( m );
		}

		if ( index != Position.EMPTY_SQUARE ) {
		    break;
		}
		if ( newSquare.x == column && newSquare.y == row ) {
		    break;	// Possible in  vertical cylinder
		}
	    }
	}
	return nbMoves;
    }

    private int generatePawnMoves( Position	position,
				   Piece	piece,
				   Piece	captured,
				   int		newColumn,
				   int		newRow,
				   int		promotionRow,
				   int		kingColumn,
				   int		kingRow,
				   int		otherColour,
				   Move[]	moves,
				   int		nbMoves ) {

	int[]	promotionTypes = { Piece.QUEEN,
				   Piece.ROOK,
				   Piece.BISHOP,
				   Piece.KNIGHT };

	if ( newRow != promotionRow ) {
	    Move m = new Move( piece,
			       new Piece( Piece.PAWN,
					  piece.colour(),
					  newColumn,
					  newRow ),
			       captured );
	    position.play( m );
	    if ( numberOfControls( position,
				   kingColumn,
				   kingRow,
				   otherColour ) == 0 ) {
		moves[ nbMoves ] = m;
		nbMoves++;
	    }
	    position.unplay( m );
	} else {
	    for ( int promotionIndex = 0;
		  promotionIndex < promotionTypes.length;
		  promotionIndex++ ) {
		Move m = new Move( piece,
				   new Piece( promotionTypes[ promotionIndex ],
					      piece.colour(),
					      newColumn,
					      newRow ),
				   captured );
		position.play( m );
		if ( numberOfControls( position,
				       kingColumn,
				       kingRow,
				       otherColour ) == 0 ) {
		    moves[ nbMoves ] = m;
		    nbMoves++;
		}
		position.unplay( m );
	    }
	}
	return nbMoves;
    }
}
