/*
 * Move.java --- Store a chess move.
 *
 * Copyright (c) 2001, 2002, 2003 by Pascal Wassong All Rights Reserved.
 *
 * This file is part of iNatch.
 *
 * iNatch is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * iNatch is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

package iNatch;

public class Move {

    public static final int	NO_CASTLING    = 0;
    public static final int	KING_CASTLING  = 1;
    public static final int	QUEEN_CASTLING = 2;

    public class IllegalMoveException extends Exception { }
    public class EmptyMoveException   extends Exception { }

    //--------------------------------------------------------------------------
    // Constructors
    //--------------------------------------------------------------------------

    public Move( Piece anOrigin, Piece aDestination ) {
	origin       = anOrigin;
	destination  = aDestination;
	captured     = null;
	capturedDest = null;
	castling     = NO_CASTLING;
	isNullMove   = false;
    }

    public Move( Piece anOrigin, Piece aDestination, Piece aCaptured ) {
	origin       = anOrigin;
	destination  = aDestination;
	captured     = aCaptured;
	capturedDest = null;
	castling     = NO_CASTLING;
	isNullMove   = false;
    }

    public Move( int aCastling,
		 int colour,
		 int originColumn,
		 int newRookType ) {

	int	row = ( colour == Piece.WHITE ) ? 0 : 7;
	int	colDestK;
	int	colDestR;
	if ( aCastling == KING_CASTLING ) {
	    colDestK = 6;
	    colDestR = 5;
	} else if ( aCastling == QUEEN_CASTLING ) {
	    colDestK = 2;
	    colDestR = 3;
	} else {
	    throw new IllegalArgumentException();
	}
	origin       = new Piece( Piece.KING,  colour, 4, row );
	destination  = new Piece( Piece.KING,  colour, colDestK, row );
	captured     = new Piece( Piece.ROOK,  colour, originColumn, row );
	capturedDest = new Piece( newRookType, colour, colDestR, row );
	castling     = aCastling;
	isNullMove   = false;
    }

    public Move( int colour ) {
	this.origin       = new Piece( Piece.KING, colour, 0, 0 );
	this.destination  = null;
	this.captured     = null;
	this.capturedDest = null;
	this.castling     = NO_CASTLING;
	this.isNullMove   = true;
    }

    protected Move( Move m ) {
	castling     = m.castling;
	origin	     = m.origin;
	destination  = m.destination;
	captured     = m.captured;
	capturedDest = m.capturedDest;
	isNullMove   = m.isNullMove;
     }

    //--------------------------------------------------------------------------
    // Selectors
    //--------------------------------------------------------------------------

    public Piece origin() {
	return this.origin;
    }

    public Piece destination() {
	return this.destination;
    }

    public Piece captured() {
	return this.captured;
    }

    public Piece capturedDestination() {
	return this.capturedDest;
    }

    public boolean hasCaptured() {
	return castling == NO_CASTLING && captured != null;
    }

    public int castling() {
	return this.castling;
    }

    public boolean isNullMove() {
	return this.isNullMove;
    }

    //--------------------------------------------------------------------------
    // String operations
    //--------------------------------------------------------------------------

    private String toLongString( boolean withCapture ) {
	if ( isNullMove ) {
	    return "  ...   ";
	}
	switch ( castling ) {
	case KING_CASTLING:
	    return longKingCastling;
	case QUEEN_CASTLING:
	    return longQueenCastling;
	default:
	    StringBuffer	s = new StringBuffer();
	    if ( origin.type() == Piece.PAWN ) {
		s.append( ' ' );
	    } else {
		s.append( origin.typeChar() );
	    }
	    s.append( origin.columnChar() );
	    s.append( origin.rowChar() );
	    s.append( ( captured == null )
		      ? '-' : miscCharacters.charAt( CAPTURE_INDEX ) );
	    if ( captured != null && withCapture ) {
		s.append( captured.typeChar() );
	    }
	    s.append( destination.columnChar() );
	    s.append( destination.rowChar() );

	    if ( origin.type() != destination.type() ) {
		s.append( '=' );
		s.append( destination.typeChar() );
	    }

	    if ( origin.type() == Piece.PAWN
		 && captured != null
		 && origin.row() == captured.row() ) {
		s.append( "ep" );
	    }
	    return s.toString();
	}
    }

    public String toBeautifullString() {
	return toLongString( false );
    }

    public String toStringWithCapture() {
	return toLongString( true );
    }

    public String toShortString( Position position ) {
	if ( isNullMove ) {
	    return "...";
	}
	switch ( castling ) {
	case KING_CASTLING:
	    return shortKingCastling;
	case QUEEN_CASTLING:
	    return shortQueenCastling;
	default:
	    StringBuffer	s = new StringBuffer();
	    if ( origin.type() != Piece.PAWN ) {
		s.append( origin.typeChar() );
		if ( position.originPiece( origin.type(),
					   destination,
					   -1,
					   -1 ) == null ) {
		    if ( position.originPiece( origin.type(),
					       destination,
					       origin.column(),
					       -1 ) != null ) {
			s.append( origin.columnChar() );
		    } else {
			if ( position.originPiece( origin.type(),
						   destination,
						   -1,
						   origin.row() ) == null ) {
			    s.append( origin.columnChar() );
			}
			s.append( origin.rowChar() );
		    }
		}
	    }
	    if ( captured != null ) {
		if ( origin.type() == Piece.PAWN ) {
		    s.append( origin.columnChar() );
		}
		s.append( miscCharacters.charAt( CAPTURE_INDEX ) );
	    }
	    s.append( destination.columnChar() );
	    s.append( destination.rowChar() );

	    if ( origin.type() != destination.type() ) {
		s.append( '=' );
		s.append( destination.typeChar() );
	    }
	    if ( origin.type() == Piece.PAWN
		 && captured != null
		 && origin.row() == captured.row() ) {
		s.append( "ep" );
	    }
	    return s.toString();
	}
    }

    //--------------------------------------------------------------------------
    // Overloaded Object functions
    //--------------------------------------------------------------------------

    public String toString() {
	if ( isNullMove ) {
	    return "...";
	}
	switch ( castling ) {
	case KING_CASTLING:
	    return "O-O";
	case QUEEN_CASTLING:
	    return "O-O-O";
	default:
	    StringBuffer s = new StringBuffer( origin.toString() );
	    s = s.append( captured == null ? '-' : 'x' );
	    s = s.append( destination.toString() );
	    if ( captured != null ) {
		s = s.append( '(' + captured.toString() + ')' );
	    }
	    return s.toString();
	}
    }

    public boolean equals( Object o ) {
	if ( o instanceof Move ) {
	    Move m = (Move)o;
	    if ( ( castling != NO_CASTLING && castling == m.castling )
		 || ( isNullMove && m.isNullMove ) ) {
		return true;
	    }
	    if ( !isNullMove
		 && !m.isNullMove
		 && origin.equals( m.origin )
		 && destination.equals( m.destination )
		 && ( ( captured == null && m.captured == null )
		      || ( captured != null && captured.equals( m.captured ) ) )
		 && ( ( capturedDest == null && m.capturedDest == null )
		      || ( capturedDest != null
			   && capturedDest.equals( m.capturedDest ) ) ) ) {

		return true;
	    }
	}
	return false;
    }

    public int hashCode() {
	return toString().hashCode();
    }

    //--------------------------------------------------------------------------

    public static void setCastlingNotationUseZeroes( boolean set ) {
	if ( set ) {
	    shortKingCastling  = SHORT_00;
	    shortQueenCastling = SHORT_000;
	    longKingCastling   = LONG_00;
	    longQueenCastling  = LONG_000;
	} else {
	    shortKingCastling  = SHORT_OO;
	    shortQueenCastling = SHORT_OOO;
	    longKingCastling   = LONG_OO;
	    longQueenCastling  = LONG_OOO;
	}
    }

    public static void setSpecialNotation( boolean set ) {
	if ( set ) {
	    miscCharacters = MISC_CHARACTERS_SPECIAL;
	} else {
	    miscCharacters = MISC_CHARACTERS_NORMAL;
	}
    }

    //--------------------------------------------------------------------------

    public static final Move NULL_WHITE_MOVE = new Move( Piece.WHITE );
    public static final Move NULL_BLACK_MOVE = new Move( Piece.BLACK );

    //--------------------------------------------------------------------------

    private static final String	MISC_CHARACTERS_NORMAL  = "x#" ;
    private static final String	MISC_CHARACTERS_SPECIAL = "\u2260" ;
    private static String	miscCharacters = MISC_CHARACTERS_NORMAL ;
    private static final int	CAPTURE_INDEX = 0 ;
    private static final int	MATE_INDEX    = 1 ;

    private static final String SHORT_OO  = "OO"  ;
    private static final String SHORT_OOO = "OOO" ;
    private static final String SHORT_00  = "00"  ;
    private static final String SHORT_000 = "000" ;

    private static final String LONG_OO   = "  O-O   ";
    private static final String LONG_OOO  = " O-O-O  ";
    private static final String LONG_00   = "  0-0   ";
    private static final String LONG_000  = " 0-0-0  ";

    private static String	shortKingCastling  = SHORT_OO ;
    private static String	shortQueenCastling = SHORT_OOO ;
    private static String	longKingCastling   = LONG_OO ;
    private static String	longQueenCastling  = LONG_OOO ;

    //--------------------------------------------------------------------------

    private int		castling;
    private Piece	origin;
    private Piece	destination;
    private Piece	captured;
    private Piece	capturedDest;
    private boolean	isNullMove;
}
