/*
 * MoveFactory.java --- Create Move from String (code was in Move.java).
 *
 * Copyright (c) 2002, 2003 by Pascal Wassong All Rights Reserved.
 *
 * This file is part of iNatch.
 *
 * iNatch is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * iNatch is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

package iNatch;

/**
 * @invariant ( status() != CORRECT_MOVE ) || ( move() != null )
 */
public final class MoveFactory {

    public static final int	CORRECT_MOVE = 0 ;
    public static final int	ILLEGAL_MOVE = 1 ;
    public static final int	EMPTY_MOVE   = 2 ;

    public static int status() {
	return status;
    }

    public static Move move() {
	return move;
    }

    /**
     * Create Move from String.  Creation status can be checked via
     * `status'.  Created move can be retrieved via `move'.
     */
    public static void make(
	String		oneMove,
	int		colour,
	String		forsytheInitials,
	Position	position ) {

	status = ILLEGAL_MOVE;
	move   = null;

	final String	captureChars = "x:";
	final String	punctuation  = "<>+#!?.,;";

	// Remove leading move number (except if 00 or 0-0)
	// Detect null move
	// Replace '' by 'c' and '' by 'e'
	// Analyze from the end :
	//	ignore all '+', '#', '!', '?' and punctuations
	//	test for castling
	//	ignore ep and e.p (final '.' has been removed)
	//	transform axb into ab
	//	if ':' , a piece has been captured
	//	if a letter, this is the type of promoted piece
	//		ignore '=' if present
	//	if <2 characters => error
	//	if only 2 characters => ab or a4 and end
	//	2 characters => destination square
	//	if a letter preceded by 'x', type of catured piece
	//	ignore '-' or 'x'
	//	if 1 letter  => type of piece or pawn captures
	//	if 1 letter and 1 number => origin square of piece
	//		or origin column and type of piece
	//	if 2 letters => origin column and type of piece
	//	if >2 characters, origin square and type of piece
	//		if not empty => error

	int[]	types = {
	    Piece.KING,
	    Piece.PAWN,
	    Piece.QUEEN,
	    Piece.ROOK,
	    Piece.BISHOP,
	    Piece.KNIGHT,
	    Piece.KNIGHT
	};

	int	type             = -1;
	int	destinationType  = -1;
	int	capturedType     = -1;
	boolean	hasCaptured      = false;
	boolean	enPassantCapture = false;

	if ( oneMove.length() == 0 ) {
	    status = EMPTY_MOVE;
	    return;
	}

	// Remove leading move number (except if 00 or 0-0)
	int	index;
	char	c;
	if ( !oneMove.startsWith( "00" ) && !oneMove.startsWith( "0-0" ) ) {
	    index = 0;
	    c     = oneMove.charAt( 0 );
	    while ( index < oneMove.length()
		   && ( c >= '0' && c <= '9' ) ) {
		index++;
		if ( index < oneMove.length() ) {
		    c = oneMove.charAt( index );
		}
	    }
	    if ( index < oneMove.length() && oneMove.charAt( index ) == '.' ) {
		index++;
	    }
	    if ( index == oneMove.length() ) {
		status = EMPTY_MOVE;
		return;
	    }
	    oneMove = oneMove.substring( index );
	}

	// Detect null move
	if ( oneMove.startsWith( ".." ) ) {
	    move = ( colour == Piece.WHITE )
		? Move.NULL_WHITE_MOVE : Move.NULL_BLACK_MOVE;
	    status = CORRECT_MOVE;
	    return;
	}

	// Replace '' by 'c' and '' by 'e'
	StringBuffer	sb = new StringBuffer( oneMove );
	for ( index = 0; index < sb.length(); index++ ) {
	    if ( sb.charAt( index ) == '' ) {
		sb.setCharAt( index, 'c' );
	    } else if ( sb.charAt( index ) == '' ) {
		sb.setCharAt( index, 'e' );
	    }
	}
	oneMove = sb.toString();

	String	s = oneMove.toLowerCase();

	//	ignore all '+', '#', '!', '?' and punctuations
	index = s.length() - 1;
	while ( index >= 0 && punctuation.indexOf( s.charAt( index ) ) != -1 ) {
	    index--;
	}
	s = s.substring( 0, index + 1 );

	//	test for castling
	if (    s.equals( "oo" ) || s.equals( "o-o" )
	     || s.equals( "00" ) || s.equals( "0-0" ) ) {

	    int	row	  = ( colour == Piece.WHITE ) ? 0 : 7;
	    int	kingIndex = position.chessboard()[ 4 ][ row ];
	    int	fIndex	  = position.chessboard()[ 5 ][ row ];
	    int	gIndex	  = position.chessboard()[ 6 ][ row ];
	    int	rookIndex = position.chessboard()[ 7 ][ row ];
	    int rookColumn = 7;
	    if ( rookIndex == Position.EMPTY_SQUARE ) {
		rookIndex = position.chessboard()[ 0 ][ row ];
		rookColumn = 0;
	    }
	    if ( ( kingIndex != Position.EMPTY_SQUARE )
		 && ( rookIndex != Position.EMPTY_SQUARE )
		 && ( fIndex    == Position.EMPTY_SQUARE )
		 && ( gIndex    == Position.EMPTY_SQUARE ) ) {

		move = new Move( Move.KING_CASTLING,
				 row == 0 ? Piece.WHITE : Piece.BLACK,
				 rookColumn,
				 position.piece( rookIndex ).type() );
		status = CORRECT_MOVE;
	    }
	    return;
	} else if (    s.equals( "ooo" ) || s.equals( "o-o-o" )
		    || s.equals( "000" ) || s.equals( "0-0-0" ) ) {

	    int	row	  = ( colour == Piece.WHITE ) ? 0 : 7;
	    int	kingIndex = position.chessboard()[ 4 ][ row ];
	    int	dIndex	  = position.chessboard()[ 3 ][ row ];
	    int	cIndex	  = position.chessboard()[ 2 ][ row ];
	    int	bIndex	  = position.chessboard()[ 1 ][ row ];
	    int	rookIndex = position.chessboard()[ 0 ][ row ];
	    int rookColumn = 0;
	    if ( rookIndex == Position.EMPTY_SQUARE ) {
		rookIndex = position.chessboard()[ 7 ][ row ];
		rookColumn = 7;
	    }
	    if ( ( kingIndex != Position.EMPTY_SQUARE )
		 && ( rookIndex != Position.EMPTY_SQUARE )
		 && ( dIndex    == Position.EMPTY_SQUARE )
		 && ( cIndex    == Position.EMPTY_SQUARE )
		 && ( bIndex    == Position.EMPTY_SQUARE ) ) {

		move = new Move( Move.QUEEN_CASTLING,
				 row == 0 ? Piece.WHITE : Piece.BLACK,
				 rookColumn,
				 Piece.ROOK );
		status = CORRECT_MOVE;
	    }
	    return;
	}

	//	ignore ep and e.p (final '.' has been removed)
	if ( index > 0 ) {
	    int		epIndex = s.lastIndexOf( "ep" );
	    if ( epIndex != -1 && epIndex == index - 1 ) {
		index -= 2;
		type             = Piece.PAWN;
		enPassantCapture = true;
		capturedType     = Piece.PAWN;
		hasCaptured      = true;
	    }

	    epIndex = s.lastIndexOf( "e.p" );
	    if ( epIndex != -1 && epIndex == index - 2 ) {
		index -= 3;
		type             = Piece.PAWN;
		enPassantCapture = true;
		capturedType     = Piece.PAWN;
		hasCaptured      = true;
	    }
	}

	//	transform axb into ab
	if ( index == 2 && captureChars.indexOf( s.charAt( 1 ) ) != -1 ) {
	    char[]	t = { s.charAt( 0 ), s.charAt( 2 ) };
	    s = new String( t );
	    index = 1;
	}

	if ( index > 1 ) {
	    //	if ':' , a piece has been captured
	    if ( s.charAt( index ) == ':' ) {
		hasCaptured = true;
		index--;
	    }

	    if ( index > 1 ) {
		//	if a letter, this is the type of promoted piece
		//		ignore '=' if present
		int promotionIndex = forsytheInitials.indexOf(
		    s.charAt( index ) );
		if ( promotionIndex != -1 ) {
		    index--;
		    destinationType = types[ promotionIndex ];
		    if ( s.charAt( index ) == '=' ) {
			index--;
		    }
		}
	    }
	}

	//	if <2 characters => error
	if ( index == -1 ) {
	    status = EMPTY_MOVE;
	    return;
	}
	if ( index < 1 ) {
	    return;
	}
	if ( index == 1 ) {
	    //	if only 2 characters => ab or a4 and end
	    char	c1 = s.charAt( 0 );
	    char	c2 = s.charAt( 1 );
	    if ( c1 < 'a' || c1 > 'h' ) {
		return;
	    }
	    if ( destinationType == -1 ) {
		destinationType = Piece.PAWN;
	    }
	    if ( c2 >= '1' && c2 <= '8' ) {
		Piece destination = new Piece( destinationType ,
					       colour ,
					       s.substring( 0, 2 ) );
		Piece origin = position.originPiece( Piece.PAWN,
						     destination,
						     -1,
						     -1 );
		if ( origin != null ) {
		    move = new Move( origin, destination );
		    status = CORRECT_MOVE;
		}
	    } else if ( c2 >= 'a' && c2 <= 'h' ) {
		Piece	origin	    = null;
		Piece	destination = null;
		Piece	captured    = null;
		int	x	    = c2 - 'a';
		for ( int y = 0; y <= 7; y++ ) {
		    index = position.chessboard()[ x ][ y ];
		    if ( index == Position.EMPTY_SQUARE ) {
			continue;
		    }
		    Piece p = position.piece( index );
		    if ( p.colour() == colour ) {
			continue;
		    }
		    int		row = y;
		    Piece	p1 = position.originPiece(
			Piece.PAWN,
			new Piece( Piece.PAWN, colour, x, y ),
			c1 - 'a',
			-1 );
		    if ( p1 == null ) {
			// Maybe en passant capture
			if ( ( colour == Piece.WHITE )
			     && ( y == 4 )
			     && ( position.chessboard()[ x ][ 5 ]
				  == Position.EMPTY_SQUARE ) ) {
			    row = 5;
			} else if ( ( colour == Piece.BLACK )
				    && ( y == 3 )
				    && ( position.chessboard()[ x ][ 2 ]
					 == Position.EMPTY_SQUARE ) ) {
			    row = 2;
			} else {
			    continue;
			}
			p1 = position.originPiece(
			    Piece.PAWN,
			    new Piece( Piece.PAWN, colour, x, row ),
			    c1 - 'a',
			    -1 );
			if ( p1 == null ) {
			    continue;
			}
		    }
		    if ( move != null ) {
			move = null;
			return;
		    }
		    move = new Move( p1,
				     new Piece( destinationType,
						colour,
						x,
						row ),
				     p );
		}
		if ( move != null ) {
		    status = CORRECT_MOVE;
		}
	    }
	    return;
	}

	//	2 characters => destination square
	char	c1 = s.charAt( index - 1 );
	char	c2 = s.charAt( index );
	if ( c1 < 'a' || c1 > 'h' || c2 < '1' || c2 > '8' ) {
	    return;
	}
	String	destinationSquare = s.substring( index - 1, index + 1 );
	index -= 2;

	//	if a letter preceded by 'x', type of catured piece
	if ( index > 1
	     && captureChars.indexOf( s.charAt( index - 1 ) ) != -1 ) {

	    int	capturedIndex = forsytheInitials.indexOf( s.charAt( index ) );
	    if ( capturedIndex == -1 ) {
		return;
	    }
	    capturedType = types[ capturedIndex ];
	    hasCaptured = true;
	    index -= 2;
	}

	//	ignore '-' or 'x'
	c = s.charAt( index );
	if ( c == '-' ) {
	    index--;
	} else if ( captureChars.indexOf( c ) != -1 ) {
	    hasCaptured = true;
	    index--;
	}

	if ( index == 0 ) {
	    //	if 1 letter  => type of piece or pawn captures
	    c = s.charAt( 0 );
	    index = forsytheInitials.indexOf( c );

	    Piece	destinationColumn = null;
	    Piece	originColumn	  = null;
	    Piece	destinationIndex  = null;
	    Piece	originIndex	  = null;

	    boolean pawnCapture = false;
	    if ( c >= 'a' && c <= 'h' ) {
		int newDestinationType = destinationType;
		if ( destinationType == -1 ) {
		    newDestinationType = Piece.PAWN;
		}
		destinationColumn =
		    new Piece( newDestinationType, colour, destinationSquare );
		originColumn = position.originPiece( Piece.PAWN,
						     destinationColumn,
						     c - 'a',
						     -1 );
		if ( originColumn != null
		     && position.capturedPiece(
			 true,
			 capturedType,
			 enPassantCapture,
			 originColumn ,
			 new Piece( Piece.PAWN,
				    colour,
				    destinationSquare ) ) != null ) {
		    pawnCapture = true;
		}
	    }
	    if ( index != -1 ) {
		int newDestinationType = destinationType;
		if ( destinationType == -1 ) {
		    newDestinationType = types[ index ];
		}
		destinationIndex =
		    new Piece( newDestinationType, colour, destinationSquare );
		originIndex = position.originPiece( types[ index ],
						    destinationIndex,
						    -1,
						    -1 );
		if ( originIndex == null ) {
		    index = -1;
		}
	    }

	    if ( index == -1 && !pawnCapture ) {
		return;
	    }

	    if ( index != -1 && pawnCapture ) {
		if ( Character.isLowerCase( oneMove.charAt( 0 ) ) ) {
		    index = -1;
		} else {
		    pawnCapture = false;
		}
	    }

	    Piece origin;
	    Piece destination;
	    if ( pawnCapture ) {
		origin      = originColumn;
		destination = destinationColumn;
		hasCaptured = true;
	    } else {
		origin      = originIndex;
		destination = destinationIndex;
	    }

	    Piece captured = position.capturedPiece( hasCaptured,
						     capturedType,
						     enPassantCapture,
						     origin ,
						     destination );
	    if ( !hasCaptured || captured != null ) {
		move   = new Move( origin, destination, captured );
		status = CORRECT_MOVE;
	    }
	    return;
	} else if ( index == 1 ) {
	    // Rd-d7, R3-d7 or a7-d7
	    c1 = s.charAt( 0 );
	    int	pieceIndex = forsytheInitials.indexOf( c1 );
	    c2 = s.charAt( 1 );
	    if ( c2 >= '1' && c2 <= '8' ) {
		//	if 1 letter and 1 number => origin square of piece
		//	or type of piece and origin column
		boolean isColumn = ( c1 >= 'a' && c1 <= 'h' );

		Piece	destinationColumn = null;
		Piece	originColumn	  = null;
		Piece	destinationIndex  = null;
		Piece	originIndex	  = null;

		if ( isColumn ) {
		    index = position.chessboard()[ c1 - 'a' ][ c2 - '1' ];
		    if ( index == -1 ) {
			isColumn = false;
		    } else {
			originColumn = position.piece( index );
			int newDestinationType = destinationType;
			if ( destinationType == -1 ) {
			    newDestinationType = originColumn.type();
			}
			destinationColumn = new Piece( newDestinationType,
						       originColumn.colour(),
						       destinationSquare );
		    }
		}

		if ( pieceIndex != -1 ) {
		    int	newDestinationType = destinationType;
		    if ( destinationType == -1 ) {
			newDestinationType = types[ pieceIndex ];
		    }
		    destinationIndex = new Piece(
			newDestinationType, colour, destinationSquare );
		    originIndex = position.originPiece( types[ pieceIndex ],
							destinationIndex,
							-1,
							c2 - '1' );
		    if ( originIndex == null ) {
			pieceIndex = -1;
		    }
		}

		if ( pieceIndex == -1 && !isColumn ) {
		    return;
		}

		if ( pieceIndex != -1 && isColumn ) {
		    if ( Character.isLowerCase( oneMove.charAt( 0 ) ) ) {
			pieceIndex = -1;
		    } else {
			isColumn = false;
		    }
		}

		Piece origin;
		Piece destination;
		if ( isColumn ) {
		    origin      = originColumn;
		    destination = destinationColumn;
		} else {
		    origin      = originIndex;
		    destination = destinationIndex;
		}

		Piece captured = position.capturedPiece( hasCaptured,
							 capturedType,
							 enPassantCapture,
							 origin ,
							 destination );
		if ( !hasCaptured || captured != null ) {
		    move   = new Move( origin, destination, captured );
		    status = CORRECT_MOVE;
		}
	    } else if ( c2 >= 'a' && c2 <= 'h' ) {
		//	if 2 letters => origin column and type of piece
		if ( destinationType == -1 ) {
		    destinationType = types[ pieceIndex ];
		}
		Piece destination = new Piece( destinationType ,
					       colour ,
					       destinationSquare );
		Piece origin = position.originPiece( types[ pieceIndex ],
						     destination,
						     c2 - 'a',
						     -1 );
		if ( origin != null ) {
		    Piece captured = position.capturedPiece( hasCaptured,
							     capturedType,
							     enPassantCapture,
							     origin ,
							     destination );
		    if ( !hasCaptured || captured != null ) {
			move   = new Move( origin, destination, captured );
			status = CORRECT_MOVE;
		    }
		}
	    }
	    return;
	}

	//	if >2 characters, origin square and type of piece
	c1 = s.charAt( index - 1 );
	c2 = s.charAt( index );
	if ( c1 < 'a' || c1 > 'h' || c2 < '1' || c2 > '8' ) {
	    return;
	}
	String	originSquare = s.substring( index - 1, index + 1 );
	index -= 2;

	if ( index != 0 ) {
	    //		if not empty => error
	    return;
	}
	c = s.charAt( 0 );
	index = forsytheInitials.indexOf( c );
	if ( index == -1 ) {
	    return;
	}
	Piece origin = new Piece( types[ index ], colour, originSquare );
	if ( destinationType == -1 ) {
	    destinationType = types[ index ];
	}
	Piece destination
	    = new Piece( destinationType, colour, destinationSquare );
	Piece captured = position.capturedPiece( hasCaptured,
						 capturedType,
						 enPassantCapture,
						 origin ,
						 destination );

	if ( origin.type() == Piece.PAWN ) {
	    if ( captured != null ) {
		int	deltaX = origin.column() - destination.column();
		if ( deltaX != 1 && deltaX != -1 ) {
		    return;
		}
	    } else {
		Piece p = position.originPiece( Piece.PAWN,
						destination,
						-1,
						-1 );
		if ( p == null
		     || p.type()   != origin.type()
		     || p.colour() != origin.colour()
		     || p.column() != origin.column()
		     || p.row()    != origin.row() ) {
		    return;
		}
	    }
	}

	if ( !hasCaptured || captured != null ) {
	    move   = new Move( origin, destination, captured );
	    status = CORRECT_MOVE;
	}
    }

    //--------------------------------------------------------------------------

    private MoveFactory() { }  // Avoid instanciation

    //--------------------------------------------------------------------------

    private static Move move   = null ;
    private static int	status = ILLEGAL_MOVE ;

}
