/*
 * MovesPanel.java --- Draw the moves.
 *
 * Copyright (c) 2001, 2002, 2003 by Pascal Wassong All Rights Reserved.
 *
 * This file is part of iNatch.
 *
 * iNatch is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * iNatch is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

package iNatch;

import java.awt.Color;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.Font;
import javax.swing.DefaultCellEditor;
import javax.swing.JLabel;
import javax.swing.JScrollPane;
import javax.swing.JTable;
import javax.swing.JTextField;
import javax.swing.ListSelectionModel;
import javax.swing.event.ListSelectionEvent;
import javax.swing.event.ListSelectionListener;
import javax.swing.table.AbstractTableModel;
import javax.swing.table.TableCellRenderer;
import javax.swing.table.TableColumn;

import wassong.gui.tools.MyPropertiesListener;
import wassong.gui.tools.Resources;


public class MovesPanel extends JScrollPane
    implements ChessGameListener, MyPropertiesListener {

    // Table model
    class MyTableModel extends AbstractTableModel {

	public int getColumnCount() {
	    return 3;
	}

	public int getRowCount() {
	    // The BasicTableUI has a strange behaviour. Adding 3 is neccesary
	    // even if this adds an unnecessary row in some circumstances.
	    return ( chessGame.totalNumberOfMoves() + 3 ) / 2;
	}

	public String getColumnName( int column ) {
	    if ( column == 1 ) {
		return resources.getString( "whiteStr" );
	    } else if ( column == 2 ) {
		return resources.getString( "blackStr" );
	    } else {
		return "";
	    }
	}

	public Object getValueAt( int row, int column ) {
	    if ( column == 0 ) {
		return new Integer( row + 1 );
	    } else {
		int	n = row * 2 + column - 1;
		if ( n < chessGame.totalNumberOfMoves() ) {
		    return chessGame.move( n );
		} else {
		    return null;
		}
	    }
	}

	public Class getColumnClass( int c ) {
	    if ( c == 0 ) {
		return new Integer( 0 ).getClass();
	    } else {
		return new Move( null, null ).getClass();
	    }
	}

	public boolean isCellEditable( int row, int column ) {
	    return ( column >= 1 );
	}

	public void setValueAt( Object value, int row, int column ) {
	    MoveFactory.make( value.toString(),
			      ( ( chessGame.currentNumberOfMoves() & 1 ) == 0 )
			      ? Piece.WHITE : Piece.BLACK,
			      Piece.initials(),
			      chessGame.position() );
	    if ( MoveFactory.status() == MoveFactory.CORRECT_MOVE ) {
		chessGame.play( MoveFactory.move() );

		// See comments and code in changeSelection
		table.changeSelection( -1, -1, false, false );

		fireTableCellUpdated( row, column );
	    }
	}
    }

    // Cell renderer
    class MoveRenderer extends JLabel
	implements TableCellRenderer {

	public MoveRenderer() {
	    super();
	    // setOpaque(true); //MUST do this for background to show up.
	    Font f = new Font( "Courier", Font.BOLD, 14 );
	    this.setFont( f );
	}

	public Component getTableCellRendererComponent( JTable	aTable,
							Object	move,
							boolean	isSelected,
							boolean	hasFocus,
							int	row,
							int	column ) {

	    this.setBackground( Color.white );
	    if ( row * 2 + column - 1 >= chessGame.currentNumberOfMoves() ) {
		this.setForeground( Color.gray );
	    } else {
		this.setForeground( Color.black );
	    }

	    if ( move != null ) {
		setText( " " + chessGame.longMove( row * 2 + column - 1 ) );
	    } else {
		setText( "" );
	    }
	    return this;
	}
    }

    // ListSelectionListener
    class MyListSelectionListener implements ListSelectionListener {
	public void valueChanged( ListSelectionEvent e ) {
	    if ( e.getValueIsAdjusting() ) {
		return;
	    }
	    for ( int row = 0; row < table.getRowCount(); row++ ) {
		for ( int col = 0; col < 3; col++ ) {
		    if ( table.isCellSelected( row, col ) ) {
			if ( col > 0 ) {
			    chessGame.gotoMove( row * 2 + col - 1 );
			    return;
			}
		    }
		}
	    }
	}
    }

    class MyTable extends JTable {
	public MyTable( AbstractTableModel aTableModel ) {
	    super( aTableModel );
	}

	public void changeSelection( int	row,
				     int	column,
				     boolean	toggle,
				     boolean	extend ) {
	    // The BasicTableUI has a strange behaviour. If the selected square
	    // is out of the normal bounds, we put it at the end of the moves.
	    int n  = chessGame.totalNumberOfMoves();
	    if ( ( column < 1 ) || ( row * 2 + column - 1 > n ) ) {
		column = ( n & 1 ) + 1;
		row    = n / 2;
	    }
	    super.changeSelection( row, column, toggle, extend );
	}
    }

    public MovesPanel( Resources	theResources,
		       MyProperties	theProperties,
		       ChessGame	aChessGame ) {

	resources = theResources;
	chessGame = aChessGame;

	tableModel = new MyTableModel();
	table = new MyTable( tableModel );
	setViewportView( table );
	table.setPreferredScrollableViewportSize( new Dimension( 250, 500 ) );

	table.setDefaultRenderer( Move.class, new MoveRenderer() );

	table.setSelectionMode( ListSelectionModel.SINGLE_SELECTION );
	table.setColumnSelectionAllowed( true );
	table.setRowSelectionAllowed( true );

	//	table.setShowGrid( false );

	ListSelectionModel rowSM = table.getSelectionModel();
	rowSM.addListSelectionListener( new MyListSelectionListener() );

	ListSelectionModel colSM =
	    table.getColumnModel().getSelectionModel();
	colSM.addListSelectionListener( new MyListSelectionListener() );

	final JTextField	textField = new JTextField( 8 );
	DefaultCellEditor	moveCellEditor = new
	    DefaultCellEditor( textField );
	table.getColumnModel().getColumn( 1 ).setCellEditor( moveCellEditor );
	table.getColumnModel().getColumn( 2 ).setCellEditor( moveCellEditor );

	TableColumn	column = table.getColumnModel().getColumn( 0 );
	column.setPreferredWidth( 20 );
	column.setMinWidth( 20 );
	column.setMaxWidth( 20 );

	chessGame.addChessPositionListener( this );
	theProperties.addPropertyListener(
	    this, MyProperties.NOTATION_USE_FRENCH_C_AND_E );
	theProperties.addPropertyListener(
	    this, MyProperties.NOTATION_USE_SPECIAL_CHARACTERS );
	theProperties.addPropertyListener(
	    this, MyProperties.NOTATION_CASTLING_USE_ZEROS );
	theProperties.addPropertyListener(
	    this, MyProperties.OUTPUT_LANGUAGE );

    }

    public void chessGameChanged() {
	chessPositionChanged();
    }

    public void chessPositionChanged() {
	tableModel.fireTableDataChanged();
    }

    public void propertyChanged( String property ) {
	chessPositionChanged();
    }

    //--------------------------------------------------------------------------

    private Resources		resources;
    private final ChessGame	chessGame;
    private JTable		table;
    private MyTableModel	tableModel;

}
