/*
 * OrthodoxCondition.java --- Orthodox chess game condition.
 *
 * Copyright (c) 2001, 2002, 2003, 2004 by Pascal Wassong All Rights Reserved.
 *
 * This file is part of iNatch.
 *
 * iNatch is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * iNatch is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

package iNatch;

public class OrthodoxCondition implements Condition {

    public String name() {
	return "orthodox-condition";
    }

    /**
     * Gives all the legal moves for the given colour.
     *
     * @param colour		Colour for which to give the moves.
     * @param previous_move	Used for potential en passant captures. null if
     *				no previous move.
     */
    public Move[] moves( Position position, int colour, Move previousMove ) {

	Move[]	moves    = new Move[ MAX_NUMBER_OF_MOVES ];
	int	nbMoves = 0;

	final int[][]	chessboard  = position.chessboard();

	int		otherColour = ( colour == Piece.WHITE )
	    ? Piece.BLACK : Piece.WHITE;
	int		kingRow     = -1;
	int		kingColumn  = -1;
	Board.Square	aSquare     = new Board.Square( -1, -1 );

    king_search:
	for ( int column = 0; column < chessboard.length; column++ ) {
	    for ( int row = 0; row < chessboard[ 0 ].length; row++ ) {
		int	index = chessboard[ column ][ row ];
		if ( index != Position.EMPTY_SQUARE ) {
		    Piece p = position.piece( index );
		    if ( p.type() == Piece.KING && p.colour() == colour ) {
			kingRow    = row;
			kingColumn = column;
			break king_search;
		    }
		}
	    }
	}

	for ( int column = 0; column < chessboard.length; column++ ) {
	    for ( int row = 0; row < chessboard[ 0 ].length; row++ ) {
		int	pieceIndex = chessboard[ column ][ row ];
		if ( pieceIndex == Position.EMPTY_SQUARE ) {
		    continue;
		}
		Piece p = position.piece( pieceIndex );
		if ( p.colour() != colour ) {
		    continue;
		}

		int type = p.type();

		switch ( type ) {
		case Piece.KING:
		{
		    int	castlingRow = ( colour == Piece.WHITE )
			? 0 : position.board().sizeY() - 1;
		    if ( row == castlingRow && column == 4
			 && numberOfControls( position,
					      4,
					      castlingRow,
					      otherColour,
					      1 ) == 0 ) {
			int rookIndex = chessboard[ 7 ][ castlingRow ];
			int rookColumn = 7;
			if ( rookIndex == Position.EMPTY_SQUARE
			     && !position.board().hasVerticalBorders() ) {
			    rookIndex = chessboard[ 0 ][ castlingRow ];
			    rookColumn = 0;
			}
			if ( rookIndex != Position.EMPTY_SQUARE
			     && position.piece( rookIndex ).type() == Piece.ROOK
			     && position.piece( rookIndex ).colour()
			     == colour ) {

			    int	fIndex = chessboard[ 5 ][ castlingRow ];
			    int	gIndex = chessboard[ 6 ][ castlingRow ];
			    if ( fIndex == Position.EMPTY_SQUARE
				 && gIndex == Position.EMPTY_SQUARE
				 && numberOfControls( position,
						      5,
						      castlingRow,
						      otherColour,
						      1 ) == 0 ) {
				Move m = new Move( Move.KING_CASTLING,
						   colour,
						   rookColumn,
						   newPieceTypeMove(
						       Piece.ROOK ) );
				position.play( m );
				if ( numberOfControls( position,
						       6,
						       castlingRow,
						       otherColour,
						       1 ) == 0 ) {
				    moves[ nbMoves ] = m;
				    nbMoves++;
				}
				position.unplay( m );
			    }
			}
			rookIndex = chessboard[ 0 ][ castlingRow ];
			rookColumn = 0;
			if ( rookIndex == Position.EMPTY_SQUARE
			     && !position.board().hasVerticalBorders() ) {
			    rookIndex = chessboard[ 7 ][ castlingRow ];
			    rookColumn = 7;
			}
			if ( rookIndex != Position.EMPTY_SQUARE
			     && position.piece( rookIndex ).type() == Piece.ROOK
			     && position.piece( rookIndex ).colour()
			     == colour ) {

			    int	bIndex = chessboard[ 1 ][ castlingRow ];
			    int	cIndex = chessboard[ 2 ][ castlingRow ];
			    int	dIndex = chessboard[ 3 ][ castlingRow ];
			    if ( bIndex == Position.EMPTY_SQUARE
				 && cIndex == Position.EMPTY_SQUARE
				 && dIndex == Position.EMPTY_SQUARE
				 && numberOfControls( position,
						      3,
						      castlingRow,
						      otherColour,
						      1 ) == 0 ) {
				Move m = new Move( Move.QUEEN_CASTLING,
						   colour,
						   rookColumn,
						   newPieceTypeMove(
						       Piece.ROOK ) );
				position.play( m );
				if ( numberOfControls( position,
						       2,
						       castlingRow,
						       otherColour,
						       1 ) == 0 ) {
				    moves[ nbMoves ] = m;
				    nbMoves++;
				}
				position.unplay( m );
			    }
			}
		    }

		    int[]	incX = { -1,  0,  1, -1,  1, -1,  0,  1 };
		    int[]	incY = {  1,  1,  1,  0,  0, -1, -1, -1 };
		    for ( int i = 0; i < incX.length; i++ ) {
			Board.Square newSquare = position.board().nextSquare(
			    column, row, incX[ i ], incY[ i ], aSquare );
			if ( newSquare == null ) {
			    continue;
			}
			int	x     = newSquare.x;
			int	y     = newSquare.y;
			int	index = chessboard[ x ][ y ];
			if ( index != Position.EMPTY_SQUARE
			     && position.piece( index ).colour() == colour ) {
			    continue;
			}

			Move m;
			if ( index == Position.EMPTY_SQUARE ) {
			    m = new Move( p,
					  new Piece( type, colour, x, y ) );
			} else {
			    m = new Move( p,
					  new Piece( type, colour, x, y ),
					  position.piece( index ) );
			}
			position.play( m );
			if ( numberOfControls( position,
					       x,
					       y,
					       otherColour,
					       1 ) == 0 ) {
			    moves[ nbMoves ] = m;
			    nbMoves++;
			}
			position.unplay( m );
		    }
		}
		break;
		case Piece.PAWN:
		{
		    int	incY;
		    int	initialRow;
		    int	promotionRow;
		    int	enPassantRow;
		    int	limitDoubleStep;
		    if ( colour == Piece.WHITE ) {
			incY	= 1;
			initialRow	= 1;
			promotionRow	= position.board().sizeY() - 1;
			enPassantRow	= position.board().sizeY() - 4;
			limitDoubleStep = 3;
		    } else {
			incY	= -1;
			initialRow	= position.board().sizeY() - 2;
			promotionRow	= 0;
			enPassantRow	= 3;
			limitDoubleStep = position.board().sizeY() - 4;
		    }

		    Board.Square square = position.board().nextSquare(
			column, row, 0, incY, aSquare );
		    if ( square != null && chessboard[ column ][ row + incY ]
			 == Position.EMPTY_SQUARE ) {

			// Single step
			nbMoves = generatePawnMoves(
			    position,
			    p,
			    null,
			    column,
			    row + incY,
			    promotionRow,
			    kingColumn,
			    kingRow,
			    otherColour,
			    moves,
			    nbMoves );

			// Multiple steps
			if ( ( colour == Piece.WHITE && row < 2 )
			     || ( colour == Piece.BLACK
				  && row > position.board().sizeY() - 3 ) ) {

			    int	currentRow = row + incY;
			    while ( currentRow != limitDoubleStep ) {
				currentRow += incY;
				if ( chessboard[ column ][ currentRow ]
				     == Position.EMPTY_SQUARE ) {

				    nbMoves = generatePawnMoves(
					position,
					p,
					null,
					column,
					currentRow,
					promotionRow,
					kingColumn,
					kingRow,
					otherColour,
					moves,
					nbMoves );
				} else {
				    break;
				}
			    }
			}
		    }

		    // Left capture
		    Board.Square newLeftSquare = position.board().nextSquare(
			column, row, -1, incY, null );
		    if ( newLeftSquare != null ) {
			int index = chessboard[ newLeftSquare.x ]
			    [ newLeftSquare.y ];
			if ( index != Position.EMPTY_SQUARE
			     && position.piece( index ).colour() != colour ) {

			    nbMoves = generatePawnMoves(
				position,
				p,
				position.piece( index ),
				newLeftSquare.x,
				newLeftSquare.y,
				promotionRow,
				kingColumn,
				kingRow,
				otherColour,
				moves,
				nbMoves );
			}
		    }

		    // Right capture
		    Board.Square newRightSquare = position.board().nextSquare(
			column, row, 1, incY, null ); // Not aSquare which is
						      // used by newLeftSquare
		    if ( newRightSquare != null ) {
			int index = chessboard[ newRightSquare.x ]
			    [ newRightSquare.y ];
			if ( index != Position.EMPTY_SQUARE
			     && position.piece( index ).colour() != colour ) {

			    nbMoves = generatePawnMoves(
				position,
				p,
				position.piece( index ),
				newRightSquare.x,
				newRightSquare.y,
				promotionRow,
				kingColumn,
				kingRow,
				otherColour,
				moves,
				nbMoves );
			}
		    }

		    // en passant capture
		    if ( previousMove != null
			 && !previousMove.isNullMove()
			 && previousMove.origin().type() == Piece.PAWN
			 && ( previousMove.destination().row() == row
			      || ( previousMove.destination().row()
				   == row - incY ) ) ) {

			square = null;
			if ( newRightSquare != null
			     && ( previousMove.destination().column()
				  == newRightSquare.x ) ) {
			    square = newRightSquare;
			} else if ( newLeftSquare != null
				    && ( previousMove.destination().column()
					 == newLeftSquare.x ) ) {
			    square = newLeftSquare;
			}

			int deltaY = previousMove.origin().row()
			    - previousMove.destination().row();
			if ( ( deltaY > 1 || deltaY < -1 )
			     && square != null
			     && ( square.x != previousMove.origin().column()
				  || ( square.y
				       != previousMove.origin().row() ) ) ) {

			    nbMoves = generatePawnMoves(
				position,
				p,
				previousMove.destination(),
				square.x,
				square.y,
				promotionRow,
				kingColumn,
				kingRow,
				otherColour,
				moves,
				nbMoves );
			}
		    }
		}
		break;
		case Piece.QUEEN:
		{
		    int[]	incX = { -1, -1, -1,  0,  0,  1,  1,  1 };
		    int[]	incY = { -1,  0,  1, -1,  1, -1,  0,  1 };

		    nbMoves = generateMoves(
			position,
			position.piece( pieceIndex ),
			kingColumn,
			kingRow,
			otherColour,
			incX,
			incY,
			moves,
			nbMoves );
		}
		break;
		case Piece.ROOK:
		{
		    int[]	incX = { -1,  0,  0,  1 };
		    int[]	incY = {  0, -1,  1,  0 };

		    nbMoves = generateMoves(
			position,
			position.piece( pieceIndex ),
			kingColumn,
			kingRow,
			otherColour,
			incX,
			incY,
			moves,
			nbMoves );
		}
		break;
		case Piece.BISHOP:
		{
		    int[]	incX = { -1, -1,  1,  1 };
		    int[]	incY = { -1,  1, -1,  1 };

		    nbMoves = generateMoves(
			position,
			position.piece( pieceIndex ),
			kingColumn,
			kingRow,
			otherColour,
			incX,
			incY,
			moves,
			nbMoves );
		}
		break;
		case Piece.KNIGHT:
		{
		    int[]	incX = { -2, -2, -1, -1,  1,  1,  2,  2 };
		    int[]	incY = { -1,  1, -2,  2, -2,  2, -1,  1 };
		    for ( int i = 0; i < incX.length; i++ ) {
			Board.Square newSquare = position.board().nextSquare(
			    column, row, incX[ i ], incY[ i ], aSquare );
			if ( newSquare == null ) {
			    continue;
			}
			int x = newSquare.x;
			int y = newSquare.y;
			int index = chessboard[ x ][ y ];
			if ( index != Position.EMPTY_SQUARE
			     && position.piece( index ).colour() == colour ) {
			    continue;
			}

			Move m;
			if ( index == Position.EMPTY_SQUARE ) {
			    m = new Move( position.piece( pieceIndex ),
					  new Piece( newPieceTypeMove(
					      Piece.KNIGHT ),
						     colour,
						     x,
						     y ) );
			} else {
			    m = new Move( position.piece( pieceIndex ),
					  new Piece( newPieceTypeCapture(
					      Piece.KNIGHT ),
						     colour,
						     x,
						     y ),
					  position.piece( index ) );
			}
			position.play( m );
			if ( numberOfControls( position,
					       kingColumn,
					       kingRow,
					       otherColour,
					       1 ) == 0 ) {
			    moves[ nbMoves ] = m;
			    nbMoves++;
			}
			position.unplay( m );
		    }
		}
		break;
		default:
		    throw new RuntimeException();
		}
	    }
	}

	Move[]	theMoves = new Move[ nbMoves ];
	for ( int i = 0; i < nbMoves; i++ ) {
	    theMoves[ i ] = moves[ i ];
	}
	return theMoves;
    }

    public int checkStatus( Position	position,
			    int		colour,
			    Move	previousMove ) {

	Move[]	moves = this.moves( position, colour, previousMove );
	if ( moves.length == 0 ) {
	    return Position.CHECK_MATE;
	}

	int[][]	chessboard = position.chessboard();
	Piece	king = null;
    king_search:
	for ( int column = 0; column < chessboard.length; column++ ) {
	    for ( int row = 0; row < chessboard[ 0 ].length; row++ ) {
		int	index = chessboard[ column ][ row ];
		if ( index != Position.EMPTY_SQUARE ) {
		    Piece p = position.piece( index );
		    if ( p.type() == Piece.KING && p.colour() == colour ) {
			king = p;
			break king_search;
		    }
		}
	    }
	}
	colour = ( colour == Piece.WHITE ) ? Piece.BLACK : Piece.WHITE;

	int	check;
	switch( numberOfControls(
		    position, king.column(), king.row(), colour, 2 ) ) {
	case 0:
	    check = Position.NO_CHECK;
	    break;
	case 1:
	    check = Position.CHECK;
	    break;
	default:
	    check = Position.DOUBLE_CHECK;
	    break;
	}
	return check;
    }

    //------------------------------------------------------------------------

    protected int[] listOfPromotionTypes() {
	final int[]	promotionTypes = {
	    Piece.QUEEN,
	    Piece.ROOK,
	    Piece.BISHOP,
	    Piece.KNIGHT
	};
	return promotionTypes;
    }

    protected int newPieceTypeMove( int type ) {
	final int[]	newPieceTypeMove = {
	    Piece.KING,
	    Piece.PAWN,
	    Piece.QUEEN,
	    Piece.ROOK,
	    Piece.BISHOP,
	    Piece.KNIGHT
	};
	return newPieceTypeMove[ type ];
    }

    protected int newPieceTypeCapture( int type ) {
	final int[]	newPieceTypeCapture = {
	    Piece.KING,
	    Piece.PAWN,
	    Piece.QUEEN,
	    Piece.ROOK,
	    Piece.BISHOP,
	    Piece.KNIGHT
	};
	return newPieceTypeCapture[ type ];
    }

    //------------------------------------------------------------------------

    /**
     * Counts how many times a square is controlled by the given color.
     *
     * @param column	Column of the square to check.
     * @param row	Row of the square to check.
     * @param colour	Colour for which to verify if the square is
     *			controlled. Must be <code>Piece.WHITE</code>
     *			or <code>Piece.BLACK</code>.
     * @param maxControl maximum number of control.  When reached, returns.
     */
    private int numberOfControls( Position	position,
				  int		column,
				  int		row,
				  int		colour,
				  int		maxControl ) {

	int		nb	= 0;
	Board.Square	aSquare = new Board.Square( -1, -1 );

	// Orthogonal and Diagonal
	int[]	lineX = { 0,  0,  1, -1,  1,  1, -1, -1 };
	int[]	lineY = { 1, -1,  0,  0,  1, -1,  1, -1 };
	int	type  = Piece.ROOK;
	for ( int i = 0; i < lineX.length; i++ ) {
	    Board.Square newSquare = position.board().nextSquare(
		column, row, lineX[ i ], lineY[ i ], aSquare );
	    if ( i == 4 ) {
		type = Piece.BISHOP;
	    }
	    while ( newSquare != null ) {
		int index = position.chessboard()[ newSquare.x ][ newSquare.y ];
		if ( index != Position.EMPTY_SQUARE ) {
		    if ( position.piece( index ).colour() == colour
			 && ( position.piece( index ).type() == Piece.QUEEN
			      || position.piece( index ).type() == type ) ) {
			nb++;
			if ( nb == maxControl ) {
			    return nb;
			}
		    }
		    break;
		}
		newSquare = position.board().nextSquare( newSquare.x,
							 newSquare.y,
							 lineX[ i ],
							 lineY[ i ],
							 newSquare );
	    }
	}

	// Knight
	int[]	knightX = { -2, -2, -1, -1,  1,  1,  2,  2 };
	int[]	knightY = { -1,  1, -2,  2, -2,  2, -1,  1 };
	for ( int i = knightX.length - 1; i >= 0; i-- ) {
	    Board.Square newSquare = position.board().nextSquare(
		column, row, knightX[ i ], knightY[ i ], aSquare );
	    if ( newSquare == null ) {
		continue;
	    }
	    int index = position.chessboard()[ newSquare.x ][ newSquare.y ];
	    if ( index == Position.EMPTY_SQUARE ) {
		continue;
	    }
	    if ( position.piece( index ).type() == Piece.KNIGHT
		 && position.piece( index ).colour() == colour ) {
		nb++;
		if ( nb == maxControl ) {
		    return nb;
		}
	    }
	}

	// King
	boolean found = false;
	for ( int i = -1; !found && i < 2; i++ ) {
	    for ( int j = -1; !found && j < 2; j++ ) {
		if ( i == 0 && j == 0 ) {
		    continue;
		}
		Board.Square newSquare
		    = position.board().nextSquare( column, row, i, j, aSquare );
		if ( newSquare == null ) {
		    continue;
		}
		int index = position.chessboard()[ newSquare.x ][ newSquare.y ];
		if ( index == Position.EMPTY_SQUARE ) {
		    continue;
		}
		if ( position.piece( index ).type() == Piece.KING
		     && position.piece( index ).colour() == colour ) {
		    nb++;
		    if ( nb == maxControl ) {
			return nb;
		    }
		    found = true;
		}
	    }
	}

	// Pawn
	found = false;
	int pawnIncY = ( colour == Piece.WHITE ) ? -1 : 1;
	if ( ( colour == Piece.WHITE && row > 0 )
	     || ( colour == Piece.BLACK
		  && row < position.board().sizeY() - 1 ) ) {
	    Board.Square newSquare = position.board().nextSquare(
		column, row, -1, pawnIncY, aSquare );
	    if ( newSquare != null ) {
		int index = position.chessboard()[ newSquare.x ][ newSquare.y ];
		if ( index != Position.EMPTY_SQUARE ) {
		    if ( position.piece( index ).type() == Piece.PAWN
			 && position.piece( index ).colour() == colour ) {
			nb++;
			if ( nb == maxControl ) {
			    return nb;
			}
			found = true;
		    }
		}
	    }
	    if ( !found ) {
		newSquare = position.board().nextSquare(
		    column, row, 1, pawnIncY, aSquare );
		if ( newSquare != null ) {
		    int index
			= position.chessboard()[ newSquare.x ][ newSquare.y ];
		    if ( index != Position.EMPTY_SQUARE ) {
			if ( position.piece( index ).type() == Piece.PAWN
			     && position.piece( index ).colour() == colour ) {
			    nb++;
			    if ( nb == maxControl ) {
				return nb;
			    }
			}
		    }
		}
	    }
	}

	return nb;
    }

    private int generateMoves( Position	position,
			       Piece	piece,
			       int	kingColumn,
			       int	kingRow,
			       int	otherColour,
			       int[]	incX,
			       int[]	incY,
			       Move[]	moves,
			       int	nbMoves ) {
	int column = piece.column();
	int row    = piece.row();
	int type   = piece.type();
	int colour = piece.colour();

	for ( int i = 0; i < incX.length; i++ ) {
	    Board.Square newSquare = position.board().nextSquare(
		column, row, incX[ i ], incY[ i ], null );
	    while ( newSquare != null ) {
		int	x     = newSquare.x;
		int	y     = newSquare.y;
		int	index = position.chessboard()[ x ][ y ];

		if ( index != Position.EMPTY_SQUARE
		     && ( x != column || y != row )
		     && position.piece( index ).colour() == colour ) {
		    break;
		}

		Move m;
		if ( index == Position.EMPTY_SQUARE
		     || ( x == column && y == row ) ) {
		    m = new Move(
			piece,
			new Piece( newPieceTypeMove( type ),
				   colour, x, y ) );
		} else {
		    m = new Move(
			piece,
			new Piece( newPieceTypeCapture( type ),
				   colour, x, y ),
			position.piece( index ) );
		}

		// Possible in vertical cylinder
		boolean	alreadyFound = false;
		for ( int j = 0; j < nbMoves; j++ ) {
		    if ( m.equals( moves[ j ] ) ) {
			alreadyFound = true;
			break;
		    }
		}
		if ( !alreadyFound ) {
		    position.play( m );
		    if ( numberOfControls( position,
					   kingColumn,
					   kingRow,
					   otherColour,
					   1 ) == 0 ) {
			moves[ nbMoves ] = m;
			nbMoves++;
		    }
		    position.unplay( m );
		}

		if ( index != Position.EMPTY_SQUARE ) {
		    break;
		}
		if ( newSquare.x == column && newSquare.y == row ) {
		    break;	// Possible in  vertical cylinder
		}
		newSquare = position.board().nextSquare(
		    newSquare.x, newSquare.y, incX[ i ], incY[ i ], newSquare );
	    }
	}
	return nbMoves;
    }

    private int generatePawnMoves( Position	position,
				   Piece	piece,
				   Piece	captured,
				   int		newColumn,
				   int		newRow,
				   int		promotionRow,
				   int		kingColumn,
				   int		kingRow,
				   int		otherColour,
				   Move[]	moves,
				   int		nbMoves ) {

	int[]	promotionTypes = listOfPromotionTypes();

	if ( newRow != promotionRow || promotionTypes == null ) {
	    Move m;
	    if ( captured == null ) {
		m = new Move( piece,
			      new Piece( newPieceTypeMove( Piece.PAWN ),
					 piece.colour(),
					 newColumn,
					 newRow ) );
	    } else {
		m = new Move( piece,
			      new Piece( newPieceTypeCapture( Piece.PAWN ),
					 piece.colour(),
					 newColumn,
					 newRow ),
			      captured );
	    }
	    position.play( m );
	    if ( numberOfControls(
		     position, kingColumn, kingRow, otherColour, 1 ) == 0 ) {
		moves[ nbMoves ] = m;
		nbMoves++;
	    }
	    position.unplay( m );
	} else {
	    for ( int promotionIndex = 0;
		  promotionIndex < promotionTypes.length;
		  promotionIndex++ ) {
		Move m = new Move( piece,
				   new Piece( promotionTypes[ promotionIndex ],
					      piece.colour(),
					      newColumn,
					      newRow ),
				   captured );
		position.play( m );
		if ( numberOfControls( position,
				       kingColumn,
				       kingRow,
				       otherColour,
				       1 ) == 0 ) {
		    moves[ nbMoves ] = m;
		    nbMoves++;
		}
		position.unplay( m );
	    }
	}
	return nbMoves;
    }
}
