/*
 * Piece.java --- Store a piece.
 *
 * Copyright (c) 2001, 2002, 2003 by Pascal Wassong All Rights Reserved.
 *
 * This file is part of iNatch.
 *
 * iNatch is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * iNatch is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

package iNatch;

public class Piece {
    public static final int	KING   = 0;
    public static final int	PAWN   = 1;
    public static final int	QUEEN  = 2;
    public static final int	ROOK   = 3;
    public static final int	BISHOP = 4;
    public static final int	KNIGHT = 5;

    public static final int	WHITE  = 0;
    public static final int	BLACK  = 1;

    //--------------------------------------------------------------------------
    // Constructors
    //--------------------------------------------------------------------------

    public Piece( int aType, int aColour, String aSquare ) {

	if ( aType < KING || aType > KNIGHT
	     || ( aColour != WHITE && aColour != BLACK )
	     || aSquare.length() != 2 ) {
	    throw new IllegalArgumentException();
	}
	int aColumn = Character.toLowerCase( aSquare.charAt( 0 ) ) - 'a';
	int aRow    = aSquare.charAt( 1 ) - '1';
	if ( aColumn < 0 || aColumn > 7
	     || aRow < 0 || aRow > 7 ) {
	    throw new IllegalArgumentException();
	}

	type   = aType;
	colour = aColour;
	row    = aRow;
	column = aColumn;
    }

    public Piece( int aType, int aColour, int aColumn, int aRow ) {

	if ( aType < KING || aType > KNIGHT
	     || ( aColour != WHITE && aColour != BLACK )
	     || aColumn < 0 || aColumn > 7
	     || aRow < 0   || aRow > 7 ) {
	    throw new IllegalArgumentException();
	}
	type   = aType;
	colour = aColour;
	row    = aRow;
	column = aColumn;
    }

    //--------------------------------------------------------------------------
    // Selectors
    //--------------------------------------------------------------------------

    public int type() {
	return this.type;
    }

    public char typeChar() {
	return pieceInitials.charAt( type );
    }

    public int colour() {
	return this.colour;
    }

    public int row() {
	return this.row;
    }

    public char rowChar() {
	return Character.forDigit( row + 1, 10 );
    }

    public int column() {
	return this.column;
    }

    public char columnChar() {
	return columnInitials.charAt( column );
    }

    //--------------------------------------------------------------------------
    // Change the language for toString
    //--------------------------------------------------------------------------

    public static void setInitials( String  pieces,
				    String  colours ) {

	pieceInitials  = pieces == null
	    ? DEFAULT_PIECE_INITIALS : pieces.toUpperCase();
	colourInitials = colours == null
	    ? DEFAULT_COLOUR_INITIALS : colours.toLowerCase();
    }

    public static void setColumnsFrenchNames( boolean french ) {
	columnInitials = french
	    ? COLUMN_INITIALS_FRENCH : COLUMN_INITIALS_NORMAL;
    }

    //--------------------------------------------------------------------------
    // Overloaded Object functions
    //--------------------------------------------------------------------------

    public String toString() {
	return "" + colourInitials.charAt( colour )
	    + pieceInitials.charAt( type )
	    + columnInitials.charAt( column )
	    + ( row + 1 );
    }

    public boolean equals( Object o ) {
	if ( o instanceof Piece ) {
	    Piece	p = (Piece)o;
	    return
		type == p.type
		&& colour == p.colour
		&& row == p.row
		&& column == p.column;
	}
	return false;
    }

    public int hashCode() {
	return toString().hashCode();
    }

    //--------------------------------------------------------------------------
    // Helpers
    //--------------------------------------------------------------------------

    public static int typeOf( char c ) {
	return pieceInitials.indexOf( Character.toUpperCase( c ) );
    }

    public static String initials() {
	return pieceInitials;
    }

    //--------------------------------------------------------------------------
    // Attributes
    //--------------------------------------------------------------------------

    private int		type;
    private int		colour;
    private int		row;
    private int		column;

    private static final String	DEFAULT_PIECE_INITIALS  = "KPQRBN";
    private static final String	DEFAULT_COLOUR_INITIALS = "wb";

    private static final String	COLUMN_INITIALS_NORMAL = "abcdefgh";
    private static final String	COLUMN_INITIALS_FRENCH = "abdfgh";

    private static String	pieceInitials  = DEFAULT_PIECE_INITIALS;
    private static String	colourInitials = DEFAULT_COLOUR_INITIALS;
    private static String	columnInitials = COLUMN_INITIALS_NORMAL;
}
