/*
 * Position.java --- Store a position.
 *
 * Copyright (c) 2001, 2002, 2003, 2004 by Pascal Wassong All Rights Reserved.
 *
 * This file is part of iNatch.
 *
 * iNatch is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * iNatch is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

package iNatch;

public class Position {

    public static final int	EMPTY_SQUARE = -1;

    public static final int	NO_CHECK     = 0;
    public static final int	CHECK        = 1;
    public static final int	DOUBLE_CHECK = 2;
    public static final int	CHECK_MATE   = 3;

    //--------------------------------------------------------------------------
    // Constructors
    //--------------------------------------------------------------------------

    public Position( Board	aBoard,
		     Condition	aCondition,
		     boolean	setInitialPosition ) {

	board     = aBoard;
	condition = aCondition;

	if ( setInitialPosition ) {
	    Piece[] initialPieces = {
		new Piece( Piece.KING,   Piece.WHITE, "e1" ),
		new Piece( Piece.ROOK,   Piece.WHITE, "a1" ),
		new Piece( Piece.KNIGHT, Piece.WHITE, "b1" ),
		new Piece( Piece.BISHOP, Piece.WHITE, "c1" ),
		new Piece( Piece.QUEEN,  Piece.WHITE, "d1" ),
		new Piece( Piece.BISHOP, Piece.WHITE, "f1" ),
		new Piece( Piece.KNIGHT, Piece.WHITE, "g1" ),
		new Piece( Piece.ROOK,   Piece.WHITE, "h1" ),

		new Piece( Piece.PAWN,   Piece.WHITE, "a2" ),
		new Piece( Piece.PAWN,   Piece.WHITE, "b2" ),
		new Piece( Piece.PAWN,   Piece.WHITE, "c2" ),
		new Piece( Piece.PAWN,   Piece.WHITE, "d2" ),
		new Piece( Piece.PAWN,   Piece.WHITE, "e2" ),
		new Piece( Piece.PAWN,   Piece.WHITE, "f2" ),
		new Piece( Piece.PAWN,   Piece.WHITE, "g2" ),
		new Piece( Piece.PAWN,   Piece.WHITE, "h2" ),

		new Piece( Piece.KING,   Piece.BLACK, "e8" ),
		new Piece( Piece.ROOK,   Piece.BLACK, "a8" ),
		new Piece( Piece.KNIGHT, Piece.BLACK, "b8" ),
		new Piece( Piece.BISHOP, Piece.BLACK, "c8" ),
		new Piece( Piece.QUEEN,  Piece.BLACK, "d8" ),
		new Piece( Piece.BISHOP, Piece.BLACK, "f8" ),
		new Piece( Piece.KNIGHT, Piece.BLACK, "g8" ),
		new Piece( Piece.ROOK,   Piece.BLACK, "h8" ),

		new Piece( Piece.PAWN,   Piece.BLACK, "a7" ),
		new Piece( Piece.PAWN,   Piece.BLACK, "b7" ),
		new Piece( Piece.PAWN,   Piece.BLACK, "c7" ),
		new Piece( Piece.PAWN,   Piece.BLACK, "d7" ),
		new Piece( Piece.PAWN,   Piece.BLACK, "e7" ),
		new Piece( Piece.PAWN,   Piece.BLACK, "f7" ),
		new Piece( Piece.PAWN,   Piece.BLACK, "g7" ),
		new Piece( Piece.PAWN,   Piece.BLACK, "h7" ),
	    };

	    pieces     = initialPieces;
	    nbPieces   = pieces.length;
	} else {
	    pieces     = new Piece[ 0 ];
	    nbPieces   = 0;
	}

	isCaptured = new boolean[ nbPieces ];
	for ( int i = 0; i < nbPieces; i++ ) {
	    isCaptured[ i ] = false;
	}

	chessboard = new int[ board.sizeX() ][ board.sizeY() ];
	for ( int i = 0; i < chessboard.length; i++ ) {
	    for ( int j = 0; j < chessboard[ 0 ].length; j++ ) {
		chessboard[ i ][ j ] = EMPTY_SQUARE;
	    }
	}
	for ( int i = 0; i < nbPieces; i++ ) {
	    chessboard[ pieces[ i ].column() ][ pieces[ i ].row() ] = i;
	}
    }

    private Position( Position p ) {

	board	   = p.board;
	condition  = p.condition;
	chessboard = new int[ board.sizeX() ][ board.sizeY() ];
	for ( int i = 0; i < chessboard.length; i++ ) {
	    for ( int j = 0; j < chessboard[ 0 ].length; j++ ) {
		chessboard[ i ][ j ] = p.chessboard[ i ][ j ];
	    }
	}

	nbPieces  = p.nbPieces;
	pieces     = (Piece[])p.pieces.clone();
	isCaptured = (boolean[])p.isCaptured.clone();
    }

    public Object clone() {
	return new Position( this );
    }

    //--------------------------------------------------------------------------
    // Selectors
    //--------------------------------------------------------------------------

    public int[][] chessboard() {
	return this.chessboard;
    }

    public Board board() {
	return this.board;
    }

    public Piece piece( int index ) {
	return pieces[ index ];
    }

    public int numberOfWhitePieces() {
	int	nb = 0;
	for ( int i = 0; i < nbPieces; i++ ) {
	    if ( pieces[ i ].colour() == Piece.WHITE && !isCaptured[ i ] ) {
		nb++;
	    }
	}
	return nb;
    }

    public int numberOfBlackPieces() {
	int	nb = 0;
	for ( int i = 0; i < nbPieces; i++ ) {
	    if ( pieces[ i ].colour() == Piece.BLACK && !isCaptured[ i ] ) {
		nb++;
	    }
	}
	return nb;
    }

    public String conditionName() {
	return condition.name();
    }

    //--------------------------------------------------------------------------
    // play / unplay
    //--------------------------------------------------------------------------

    public void play( Move m ) {
	if ( m.isNullMove() ) {
	    return;
	}

	Piece	origin      = m.origin();
	int	indexOrigin = chessboard[ origin.column() ][ origin.row() ];
	Piece	destination = m.destination();

	// Destination must be different from origin for the null moves in
	// vertical cylinder chess.
	boolean	hasCaptured =
	    ( m.castling() == Move.NO_CASTLING )
	    && ( ( origin.row() != destination.row() )
		 || ( origin.column() != destination.column() ) )
	    && m.captured() != null;
	if ( indexOrigin == EMPTY_SQUARE
	     || pieces[ indexOrigin ].type() != origin.type()
	     || pieces[ indexOrigin ].colour() != origin.colour()
	     || m.hasCaptured() != hasCaptured ) {
	    throw new IllegalArgumentException();
	}

	chessboard[ origin.column() ][ origin.row() ] = EMPTY_SQUARE;

	if ( m.captured() != null ) {
	    Piece	captured      = m.captured();
	    int		indexCaptured =
		chessboard[ captured.column() ][ captured.row() ];
	    chessboard[ captured.column() ][ captured.row() ] = EMPTY_SQUARE;

	    captured = m.capturedDestination();
	    if ( captured != null ) {
		chessboard[ captured.column() ][ captured.row() ] =
		    indexCaptured;
		pieces[ indexCaptured ] = captured;
	    } else {
		isCaptured[ indexCaptured ] = true;
	    }
	}

	chessboard[ destination.column() ][ destination.row() ] = indexOrigin;
	pieces[ indexOrigin ] = destination;
    }

    public void unplay( Move m ) {

	if ( m.isNullMove() ) {
	    return;
	}

	Piece	destination      = m.destination();
	int	indexDestination = chessboard
	    [ destination.column() ][ destination.row() ];

	if ( pieces[ indexDestination ].type() != destination.type()
	     || pieces[ indexDestination ].colour() != destination.colour() ) {
	    throw new IllegalArgumentException();
	}

	chessboard[ destination.column() ][ destination.row() ] = EMPTY_SQUARE;

	Piece	origin = m.origin();
	chessboard[ origin.column() ][ origin.row() ] = indexDestination;
	pieces[ indexDestination ] = origin;

	if ( m.captured() != null ) {
	    Piece	captured = m.captured();
	    Piece	capturedDest = m.capturedDestination();
	    if ( capturedDest != null ) {
		int	indexCapt = chessboard[ capturedDest.column() ]
		    [ capturedDest.row() ];
		chessboard[ capturedDest.column() ][ capturedDest.row() ] =
		    EMPTY_SQUARE;
		chessboard[ captured.column() ][ captured.row() ] = indexCapt;
		pieces[ indexCapt ] = captured;
	    } else {
		int i;
		for ( i = 0; i < nbPieces; i++ ) {
		    if ( isCaptured[ i ]
			 && pieces[ i ].type() == captured.type()
			 && pieces[ i ].colour() == captured.colour()
			 && pieces[ i ].row() == captured.row()
			 && pieces[ i ].column() == captured.column() ) {

			chessboard[ captured.column() ][ captured.row() ] = i;
			isCaptured[ i ] = false;
			break;
		    }
		}
		if ( i == nbPieces ) {
		    addPieceAt( captured );
		}
	    }
	}
    }

    //--------------------------------------------------------------------------
    // Edit the position
    //--------------------------------------------------------------------------

    /**
     * @return true if position has been modified.
     */
    public boolean clearSquare( int column, int row ) {
	int	index = chessboard[ column ][ row ];
	if ( index != EMPTY_SQUARE ) {
	    chessboard[ column ][ row ] = EMPTY_SQUARE;

	    nbPieces--;
	    if ( index < nbPieces ) {
		pieces[ index ]     = pieces[ nbPieces ];
		isCaptured[ index ] = isCaptured[ nbPieces ];
		chessboard[ pieces[ index ].column() ]
		    [ pieces[ index ].row() ] = index;
	    }
	    return true;
	} else {
	    return false;
	}
    }

    /**
     * @return true if position has been modified.
     */
    public boolean addPieceAt( Piece piece ) {
	int	column = piece.column();
	int	row    = piece.row();
	int	index  = chessboard[ column ][ row ];
	if ( index == EMPTY_SQUARE ) {
	    Piece[]	newPieces = new Piece[ nbPieces + 1 ];
	    boolean[]	newIsCapt = new boolean[ nbPieces + 1 ];
	    for ( int i = 0; i < nbPieces; i++ ) {
		newPieces[ i ] = pieces[ i ];
		newIsCapt[ i ] = isCaptured[ i ];
	    }
	    pieces     = newPieces;
	    isCaptured = newIsCapt;

	    index = nbPieces;
	    chessboard[ column ][ row ] = index;
	    nbPieces++;
	} else if ( piece.type() == pieces[ index ].type()
		    && piece.colour() == pieces[ index ].colour() ) {
	    return false;
	}
	pieces[ index ]     = piece;
	isCaptured[ index ] = false;

	return true;
    }

    //--------------------------------------------------------------------------
    // Utilities
    //--------------------------------------------------------------------------

    public void setBoard( Board aBoard ) {

	if ( aBoard.sizeX() != board.sizeX()
	     || aBoard.sizeY() != board.sizeY() ) {
	    throw new IllegalArgumentException();
	}
	board = aBoard;
    }

    public void setCondition( Condition aCondition ) {
	condition = aCondition;
    }

    public Move[] moves( int colour, Move previousMove ) {
	return condition.moves( this, colour, previousMove );
    }

    public int checkStatus( int colour, Move previousMove ) {
	return condition.checkStatus( this, colour, previousMove );
    }

    public Piece originPiece( int	type,
			      Piece	piece,
			      int	columnConstraint,
			      int	rowConstraint ) {

	Piece		pieceFound = null;

	switch ( type ) {
	case Piece.KING:
	case Piece.KNIGHT:
	{
	    int[]	incKingX   = { -1, -1, -1,  0,  0,  1,  1,  1 };
	    int[]	incKingY   = { -1,  0,  1, -1,  1, -1,  0,  1 };
	    int[]	incKnightX = { -2, -2, -1, -1,  1,  1,  2,  2 };
	    int[]	incKnightY = { -1,  1, -2,  2, -2,  2, -1,  1 };

	    int[]	incX = ( type == Piece.KING ) ? incKingX : incKnightX;
	    int[]	incY = ( type == Piece.KING ) ? incKingY : incKnightY;

	    Board.Square aSquare = new Board.Square( -1, -1 );
	    for ( int i = incX.length; i-- > 0; ) {
		Board.Square newSquare = board.nextSquare( piece.column(),
							   piece.row(),
							   incX[ i ],
							   incY[ i ],
							   aSquare );
		if ( newSquare != null ) {
		    int index = chessboard[ newSquare.x ][ newSquare.y ];
		    if ( index != Position.EMPTY_SQUARE ) {
			Piece p = piece( index );
			if ( p.type() == type
			     && p.colour() == piece.colour()
			     && ( columnConstraint == -1
				  || newSquare.x == columnConstraint )
			     && ( rowConstraint == -1
				  || newSquare.y == rowConstraint ) ) {
			    if ( pieceFound != null ) {
				return null;
			    }
			    pieceFound = p;
			}
		    }
		}
	    }
	}
	break;
	case Piece.PAWN:
	{
	    if ( columnConstraint != -1 ) {
		int	rowInc  = ( piece.colour() == Piece.WHITE ) ? -1 : 1;
		int	destX   = piece.column();
		int	destY   = piece.row();
		int	sourceX = columnConstraint;
		int	sourceY = destY + rowInc;

		int deltaX = destX - sourceX;
		if ( deltaX != 1 && deltaX != -1 ) {
		    return null;
		}

		if ( sourceY < 0 || sourceY >= board.sizeY() ) {
		    // This may happen if we have an english white move
		    // like Bc1 which is intepreted in german as Pawn c1.
		    return null;
		}

		int index = chessboard[ sourceX ][ sourceY ];
		if ( index == Position.EMPTY_SQUARE ) {
		    return null;
		}
		Piece p = piece( index );
		if ( p.type() == Piece.PAWN && p.colour() == piece.colour() ) {
		    return p;
		}
		return null;
	    }

	    int rowIncrement = ( piece.colour() == Piece.WHITE ) ? -1 : 1;
	    int	row	     = piece.row() + rowIncrement;
	    if ( row < 0 || row >= board.sizeY() ) {
		return null;
	    }
	    int		index = chessboard[ piece.column() ][ row ];
	    if ( index == -1
		 && ( ( row == 2 && rowIncrement == -1 )
		      || ( row == 5 && rowIncrement == 1 ) ) ) {
		row += rowIncrement;
		index = chessboard[ piece.column() ][ row ];
	    }
	    if ( index == Position.EMPTY_SQUARE ) {
		return null;
	    }
	    Piece p = piece( index );
	    if ( p.type() == Piece.PAWN && p.colour() == piece.colour() ) {
		pieceFound = p;
	    }
	}
	break;
	case Piece.QUEEN:
	case Piece.ROOK:
	{
	    int		x = piece.column() - 1;
	    int		y = piece.row();
	    while ( x >= 0 ) {
		int		index = chessboard[ x ][ y ];
		if ( index != Position.EMPTY_SQUARE ) {
		    Piece	p = piece( index );
		    if ( p.type() == type
			 && p.colour() == piece.colour()
			 && ( columnConstraint == -1
			      || x == columnConstraint )
			 && ( rowConstraint == -1
			      || y == rowConstraint ) ) {
			pieceFound = p;
		    }
		    break;
		}
		x--;
	    }

	    x = piece.column() + 1;
	    while ( x < board.sizeX() ) {
		int		index = chessboard[ x ][ y ];
		if ( index != Position.EMPTY_SQUARE ) {
		    Piece	p = piece( index );
		    if ( p.type() == type
			 && p.colour() == piece.colour()
			 && ( columnConstraint == -1
			      || x == columnConstraint )
			 && ( rowConstraint == -1
			      || y == rowConstraint ) ) {
			if ( pieceFound != null ) {
			    return null;
			}
			pieceFound = p;
		    }
		    break;
		}
		x++;
	    }

	    x = piece.column();
	    y = piece.row() - 1;
	    while ( y >= 0 ) {
		int		index = chessboard[ x ][ y ];
		if ( index != Position.EMPTY_SQUARE ) {
		    Piece	p = piece( index );
		    if ( p.type() == type
			 && p.colour() == piece.colour()
			 && ( columnConstraint == -1
			      || x == columnConstraint )
			 && ( rowConstraint == -1
			      || y == rowConstraint ) ) {
			if ( pieceFound != null ) {
			    return null;
			}
			pieceFound = p;
		    }
		    break;
		}
		y--;
	    }

	    y = piece.row() + 1;
	    while ( y < board.sizeY() ) {
		int		index = chessboard[ x ][ y ];
		if ( index != Position.EMPTY_SQUARE ) {
		    Piece	p = piece( index );
		    if ( p.type() == type
			 && p.colour() == piece.colour()
			 && ( columnConstraint == -1
			      || x == columnConstraint )
			 && ( rowConstraint == -1
			      || y == rowConstraint ) ) {
			if ( pieceFound != null ) {
			    return null;
			}
			pieceFound = p;
		    }
		    break;
		}
		y++;
	    }
	}
	if ( type != Piece.QUEEN ) {
	    break;
	}
	// Fall throu if QUEEN
	case Piece.BISHOP:
	{
	    int		x = piece.column() - 1;
	    int		y = piece.row() - 1;
	    while ( x >= 0 && y >= 0 ) {
		int	index = chessboard[ x ][ y ];
		if ( index != Position.EMPTY_SQUARE ) {
		    Piece	p = piece( index );
		    if ( p.type() == type
			 && p.colour() == piece.colour()
			 && ( columnConstraint == -1
			      || x == columnConstraint )
			 && ( rowConstraint == -1
			      || y == rowConstraint ) ) {
			pieceFound = p;
		    }
		    break;
		}
		x--;
		y--;
	    }

	    x = piece.column() + 1;
	    y = piece.row() - 1;
	    while ( x < board.sizeX() && y >= 0 ) {
		int	index = chessboard[ x ][ y ];
		if ( index != Position.EMPTY_SQUARE ) {
		    Piece	p = piece( index );
		    if ( p.type() == type
			 && p.colour() == piece.colour()
			 && ( columnConstraint == -1
			      || x == columnConstraint )
			 && ( rowConstraint == -1
			      || y == rowConstraint ) ) {
			if ( pieceFound != null ) {
			    return null;
			}
			pieceFound = p;
		    }
		    break;
		}
		x++;
		y--;
	    }

	    x = piece.column() - 1;
	    y = piece.row() + 1;
	    while ( x >= 0 && y < board.sizeY() ) {
		int		index = chessboard[ x ][ y ];
		if ( index != Position.EMPTY_SQUARE ) {
		    Piece	p = piece( index );
		    if ( p.type() == type
			 && p.colour() == piece.colour()
			 && ( columnConstraint == -1
			      || x == columnConstraint )
			 && ( rowConstraint == -1
			      || y == rowConstraint ) ) {
			if ( pieceFound != null ) {
			    return null;
			}
			pieceFound = p;
		    }
		    break;
		}
		x--;
		y++;
	    }

	    x = piece.column() + 1;
	    y = piece.row() + 1;
	    while ( x < board.sizeX() && y < board.sizeY() ) {
		int index = chessboard[ x ][ y ];
		if ( index != Position.EMPTY_SQUARE ) {
		    Piece p = piece( index );
		    if ( p.type() == type
			 && p.colour() == piece.colour()
			 && ( columnConstraint == -1
			      || x == columnConstraint )
			 && ( rowConstraint == -1
			      || y == rowConstraint ) ) {
			if ( pieceFound != null ) {
			    return null;
			}
			pieceFound = p;
		    }
		    break;
		}
		x++;
		y++;
	    }
	}
	break;
	default:
	    break;
	}

	return pieceFound;
    }

    public Piece capturedPiece(
	boolean		hasCaptured      ,
	int		capturedType     ,
	boolean		enPassantCapture ,
	Piece		exceptionPiece   ,
	Piece		p                ) {

	if ( exceptionPiece.row() == p.row()
	     && exceptionPiece.column() == p.column() ) {
	    // A piece cannot capture itself. This kind of move is possible in
	    // vertical cylinder chess when playing a null move suh as Ra3-a3.
	    return null;
	}

	int	index = chessboard[ p.column() ][ p.row() ];

	// Potential en Passant capture
	if ( ( index == Position.EMPTY_SQUARE )
	     && ( p.type() == Piece.PAWN && hasCaptured ) ) {
	    if ( p.colour() == Piece.WHITE && p.row() == 5 ) {
		index = chessboard[ p.column() ][ 4 ];
		capturedType = Piece.PAWN;
	    } else if ( p.colour() == Piece.BLACK && p.row() == 2 ) {
		index = chessboard[ p.column() ][ 3 ];
		capturedType = Piece.PAWN;
	    }
	}

	if ( index == Position.EMPTY_SQUARE ) {
	    return null;
	}
	Piece	pc = piece( index );
	if ( capturedType != -1 && capturedType != pc.type() ) {
	    return null;
	}

	if ( enPassantCapture ) {
	    if ( capturedType != Piece.PAWN || p.type() != Piece.PAWN ) {
		return null;
	    }
	}

	return pc;
    }

    //--------------------------------------------------------------------------
    // Conversion to strings
    //--------------------------------------------------------------------------

    public String toString() {
	StringBuffer	s = new StringBuffer( 100 );
	for ( int j = chessboard[0].length - 1; j >= 0; j-- ) {
	    int		nbEmpty = 0;
	    for ( int i = 0; i < chessboard.length; i++ ) {
		int	pieceIndex = chessboard[ i ][ j ];
		if ( pieceIndex == EMPTY_SQUARE ) {
		    nbEmpty++;
		} else {
		    if ( nbEmpty > 0 ) {
			s.append( (char)( nbEmpty + '0' ) );
			nbEmpty = 0;
		    }
		    char c = pieces[ pieceIndex ].typeChar();
		    if ( pieces[ pieceIndex ].colour() == Piece.BLACK ) {
			c = Character.toLowerCase( c );
		    }
		    s.append( c );
		}
	    }
	    if ( nbEmpty > 0 ) {
		s.append( (char)( nbEmpty + '0' ) );
		nbEmpty = 0;
	    }
	    if ( j != 0 ) {
		s.append( '/' );
	    }
	}
	return s.toString();
    }

    public String diagram() {
	StringBuffer	s        = new StringBuffer( 600 );
	String		lineSep  = "+---+---+---+---+---+---+---+---+\n";
	s.append( lineSep );
	for ( int j = chessboard[0].length - 1; j >= 0; j-- ) {
	    s.append( '|' );
	    for ( int i = 0; i < chessboard.length; i++ ) {
		int	pieceIndex = chessboard[ i ][ j ];
		if ( pieceIndex == EMPTY_SQUARE ) {
		    if ( ( ( i + j ) & 1 ) == 1 ) {
			s.append( "  " );
		    } else {
			s.append( " ." );
		    }
		} else {
		    char c = pieces[ pieceIndex ].typeChar();
		    s.append( pieces[ pieceIndex ].colour() == Piece.WHITE
			      ? ' ' : '*' );
		    s.append( Character.toUpperCase( c ) );
		}
		s.append( " |" );
	    }
	    s.append( '\n' + lineSep );
	}
	return s.toString();
    }

    public String font1Echecs() {

	char[][]	piecesGlyphs = {
	    // On white squares
	    { 'r' },		// wK
	    { 'p' },		// wP
	    { 'd', 'e' },	// wQ
	    { 't' },		// wR
	    { 'f' },		// wB
	    { 'c' },		// wN

	    { 'R' },		// bK
	    { 'P' },		// bP
	    { 'D' },		// bQ
	    { 'T' },		// bR
	    { 'F' },		// bB
	    { 'C' },		// bN

	    // On black squares
	    { 48, 49, 'r' },		// wK
	    { 40, 41, 'p' },		// wP
	    { 50, 51, 'd', 'e' },	// wQ
	    { 52, 53, 't' },		// wR
	    { 54, 55, 'f' },		// wB
	    { 56, 57, 'c' },		// wN

	    { 48, 49, 'R' },		// bK
	    { 40, 41, 'P' },		// bP
	    { 50, 51, 'D' },		// bQ
	    { 52, 53, 'T' },		// bR
	    { 54, 55, 'F' },		// bB
	    { 56, 57, 'C' },		// bN
	};
	char[]	whiteSquareGlyphs    = { 32 };
	char[]	blackSquareGlyphs    = { 58 };
	char	verticalBarGlyph     = 47;
	char	horizontalLowerGlyph = 45;
	char	horizontalUpperGlyph = 95;
	char	lowerPointGlyph      = 33;
	char	upperPointGlyph      = 36;

	StringBuffer	s = new StringBuffer();

	s.append( lowerPointGlyph );
	for ( int i = 0; i < chessboard.length; i++ ) {
	    s.append( horizontalLowerGlyph );
	}
	s.append( lowerPointGlyph );
	s.append( '\n' );

	boolean		hasVerticalBorders = board.hasVerticalBorders();

	char[]		glyphs;
	for ( int j = chessboard[ 0 ].length - 1; j >= 0; j-- ) {

	    if ( hasVerticalBorders ) {
		s.append( verticalBarGlyph );
	    }

	    for ( int i = 0; i < chessboard.length; i++ ) {

		int		p = chessboard[ i ][ j ];
		boolean	isBlackSquare = ( ( i + j ) % 2 ) == 0;
		if ( p == -1 ) {
		    if ( isBlackSquare ) {
			glyphs = blackSquareGlyphs;
		    } else {
			glyphs = whiteSquareGlyphs;
		    }
		} else {
		    Piece	pi    = piece( p );
		    int	index = pi.type();
		    if ( pi.colour() == Piece.BLACK ) {
			index += 6;
		    }
		    if ( isBlackSquare ) {
			index += 12;
		    }
		    glyphs = piecesGlyphs[ index ];
		}
		s.append( glyphs );
	    }
	    if ( hasVerticalBorders ) {
		s.append( verticalBarGlyph );
	    }
	    s.append( '\n' );
	}

	s.append( upperPointGlyph );
	for ( int i = 0; i < chessboard.length; i++ ) {
	    s.append( horizontalUpperGlyph );
	}
	s.append( upperPointGlyph );
	s.append( '\n' );

	return s.toString();
    }

    //--------------------------------------------------------------------------
    // Attributes
    //--------------------------------------------------------------------------

    private Board	board	   ;
    private Condition	condition  ;
    private int[][]	chessboard ;
    private Piece[]	pieces	   ;
    private boolean[]	isCaptured ;
    private int		nbPieces   ;
}
