/*
 * Preferences.java --- Preferences dialog box.
 *
 * Copyright (c) 2001, 2002, 2003 by Pascal Wassong All Rights Reserved.
 *
 * This file is part of iNatch.
 *
 * iNatch is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * iNatch is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

package iNatch;

import java.awt.BorderLayout;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.FileDialog;
import java.awt.Frame;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.io.File;
import java.io.FilenameFilter;
import java.util.StringTokenizer;
import javax.swing.Box;
import javax.swing.BoxLayout;
import javax.swing.DefaultListModel;
import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JComboBox;
import javax.swing.JDialog;
import javax.swing.JLabel;
import javax.swing.JList;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTabbedPane;
import javax.swing.JTextField;
import javax.swing.ListSelectionModel;

import wassong.gui.tools.Resources;


public final class Preferences {

    public static boolean showPreferences( Resources	resources,
					   MyProperties	properties,
					   Frame	aFrame ) {

	nbUpdatables = 0;

	JTabbedPane tabbedPane = new JTabbedPane();
	tabbedPane.addTab( resources.getString( "tabTitleCopy" ),
			   null,
			   createCopyPanel( resources, properties, aFrame ),
			   resources.getString( "tabTooltipCopy" ) );
	tabbedPane.addTab( resources.getString( "tabTitleNatch" ),
			   null,
			   createNatchPanel( resources, properties, aFrame ),
			   resources.getString( "tabTooltipNatch" ) );
	tabbedPane.addTab( resources.getString( "tabTitleNotation" ),
			   null,
			   createNotationPanel( resources, properties, aFrame ),
			   resources.getString( "tabTooltipNotation" ) );
	tabbedPane.addTab( resources.getString( "tabTitleMisc" ),
			   null,
			   createMiscPanel( resources, properties, aFrame ),
			   resources.getString( "tabTooltipMisc" ) );

	final String	okButton     = resources.getString( "ok-string" );
	final String	cancelButton = resources.getString( "cancel-string" );
	final Object[]	options      = { okButton , cancelButton };

	optionPane = new JOptionPane(
	    tabbedPane                   ,
	    JOptionPane.PLAIN_MESSAGE    ,
	    JOptionPane.OK_CANCEL_OPTION ,
	    null                         ,
	    options                      ,
	    options[ 0 ]                 );

	dialog = optionPane.createDialog(
	    aFrame ,
	    resources.getString( "titlePreferencesDialog" ) );

	dialog.show();

	Object	selectedValue = optionPane.getValue();
	boolean	ok = okButton.equals( selectedValue );
	if ( ok ) {
	    updateProperties( resources, properties );
	}
	return ok;
    }

    //--------------------------------------------------------------------------

    private interface Updatable {
	void	updateProperty();
    }

    private static final int	MAX_UPDATABLE = 10 ;
    private static Updatable[]	updatables    = new Updatable[ MAX_UPDATABLE ];
    private static int		nbUpdatables  = 0 ;
    private static void addToUpdatableList( Updatable u ) {
	updatables[ nbUpdatables ] = u;
	nbUpdatables++;
    }

    private static void updateProperties( Resources	resources,
					  MyProperties	properties ) {
	for ( int i = 0; i < nbUpdatables; i++ ) {
	    updatables[ i ].updateProperty();
	}

	// Specific updates
	StringBuffer defaultCopyAction = new StringBuffer();
	for ( int i = 0; i < currentCopyOptions.getSize(); i++ ) {
	    if ( i > 0 ) {
		defaultCopyAction.append( ';' );
	    }
	    String	name = (String)currentCopyOptions.get( i );
	    int		j;
	    for ( j = 0; j < copyNames.length; j++ ) {
		if ( name.equals( copyNames[ j ] ) ) {
		    break;
		}
	    }
	    if ( j < copyNames.length ) {
		defaultCopyAction.append( copyTypes[ j ] );
	    } else {
		defaultCopyAction.append( name );
	    }
	}
	properties.setProperty( MyProperties.DEFAULT_COPY_ACTION,
				defaultCopyAction.toString() );

	String newLanguage = (String )( languageCombo.getSelectedItem() );
	properties.setProperty(
	    MyProperties.OUTPUT_LANGUAGE,
	    newLanguage.substring( 0, newLanguage.indexOf( ' ' ) ) );

    }

    //--------------------------------------------------------------------------

    private static class MyCheckBox extends JCheckBox implements Updatable {
	public MyCheckBox( Resources	theResources,
			   MyProperties	theProperties,
			   String	aProperty ) {

	    super( theResources.getString( aProperty ) );

	    properties = theProperties;
	    property   = aProperty;

	    setSelected( properties.getBooleanProperty( property ) );

	    addToUpdatableList( this );
	}

	public void updateProperty() {
	    properties.setProperty(
		property,
		( isSelected() ? Boolean.TRUE : Boolean.FALSE ).toString() );
	}

	private MyProperties	properties;
	private String		property;
    }

    private static class MyTextField extends JTextField implements Updatable {
	public MyTextField( MyProperties theProperties, String aProperty ) {

	    super( 20 );

	    properties = theProperties;
	    property   = aProperty;

	    setText( properties.getProperty( property ) );

	    addToUpdatableList( this );
	}

	public void updateProperty() {
	    properties.setProperty( property, getText() );
	}

	private MyProperties	properties;
	private String		property;
    }

    //--------------------------------------------------------------------------

    private static JPanel createNatchPanel( Resources		resources,
					    MyProperties	properties,
					    final Frame		aFrame ) {

	JLabel pathLabel   = new JLabel( resources.getString( "Natch.path" ) );

	final MyTextField pathText
	    = new MyTextField( properties, MyProperties.NATCH_PATH );

	JButton		pathBrowse  =
	    new JButton( resources.getString( "Browse-button" ) );
	pathBrowse.addActionListener( new ActionListener() {
	    public void actionPerformed( ActionEvent e ) {

		FileDialog fileDialog = new FileDialog( aFrame );

		fileDialog.setMode( FileDialog.LOAD );
		fileDialog.setFilenameFilter( new FilenameFilter() {
			public boolean accept( File dir, String name ) {
			    String	execName = Natch.executableName();
			    boolean	acceptIt = name.equals( execName );
			    if ( !acceptIt ) {
				// For an unknown reason, on Mac OS X, the file
				// name has a '/' prefix.
				execName = "/" + execName;
				acceptIt = name.equals( execName );
			    }
			    return acceptIt;
			} } );

		fileDialog.setDirectory( pathText.getText() );
		fileDialog.show();

		String file = fileDialog.getFile();
		if ( file != null ) {
		    pathText.setText( fileDialog.getDirectory() );
		}
	    }
	} );

	JPanel pathBox = new JPanel();
	pathBox.setLayout( new BorderLayout() );
	pathBox.add( pathLabel  , BorderLayout.WEST   );
	pathBox.add( pathText   , BorderLayout.CENTER );
	pathBox.add( pathBrowse , BorderLayout.EAST   );

	JLabel argumentsLabel
	    = new JLabel( resources.getString( "Natch.arguments" ) );

	MyTextField	argumentsText
	    = new MyTextField( properties, MyProperties.NATCH_PARAMETERS );
	JPanel		argumentBox    = new JPanel();
	argumentBox.setLayout( new BorderLayout() );
	argumentBox.add( argumentsLabel , BorderLayout.WEST   );
	argumentBox.add( argumentsText  , BorderLayout.CENTER );

	JPanel box = new JPanel();
	box.setLayout( new BoxLayout( box, BoxLayout.Y_AXIS ) );
	box.add( pathBox       );
	box.add( argumentBox   );

	JPanel pane = new JPanel();
	pane.setLayout( new BorderLayout() );
	pane.add( box, BorderLayout.NORTH );

	return pane;
    }

    //--------------------------------------------------------------------------

    private static JPanel createCopyPanel( Resources	resources,
					   MyProperties	properties,
					   final Frame	aFrame ) {

	if ( copyTypes == null ) {
	    copyTypes = ChessGame.formatTypes();
	    copyNames = new String[ copyTypes.length ];
	    for ( int i = 0; i < copyTypes.length; i++ ) {
		copyNames[ i ]
		    = resources.getString( "Copy-" + copyTypes[ i ] );
	    }
	}

	final JList	left = new JList( copyNames );
	left.setSelectionMode( ListSelectionModel.SINGLE_SELECTION );
	left.setSelectedIndex( 0 );
	JScrollPane	leftSP = new JScrollPane( left );

	JPanel	leftBox = new JPanel();
	leftBox.setLayout( new BorderLayout() );
	leftBox.add( new JLabel(
	    resources.getString( "Copy_Possible_Options" ) ),
		     BorderLayout.NORTH );
	leftBox.add( leftSP , BorderLayout.CENTER );

	currentCopyOptions.clear();
	StringTokenizer	t = new StringTokenizer(
	    properties.getProperty( MyProperties.DEFAULT_COPY_ACTION ), ";" );
	while ( t.hasMoreTokens() ) {
	    String	type = t.nextToken();
	    int		i;
	    for ( i = 0; i < copyTypes.length; i++ ) {
		if ( type.equals( copyTypes[ i ] ) ) {
		    break;
		}
	    }
	    if ( i < copyTypes.length ) {
		currentCopyOptions.addElement( copyNames[ i ] );
	    } else {
		currentCopyOptions.addElement( type );
	    }
	}

	final JList	right = new JList( currentCopyOptions );
	right.setSelectionMode( ListSelectionModel.SINGLE_SELECTION );
	right.setSelectedIndex( currentCopyOptions.getSize() - 1 );
	JScrollPane	rightSP = new JScrollPane( right );

	JPanel	rightBox = new JPanel();
	rightBox.setLayout( new BorderLayout() );
	rightBox.add( new JLabel(
	    resources.getString( "Copy_Current_Options" ) ),
		      BorderLayout.NORTH );
	rightBox.add( rightSP , BorderLayout.CENTER );

	Dimension minimumSize = new Dimension( 100, 140 );
	leftSP.setPreferredSize ( minimumSize );
	rightSP.setPreferredSize( minimumSize );

	JButton		addButton    = new JButton( "==>" );
	final JButton	removeButton = new JButton( "<==" );
	addButton.setAlignmentX   ( Component.CENTER_ALIGNMENT );
	removeButton.setAlignmentX( Component.CENTER_ALIGNMENT );
	removeButton.setEnabled( currentCopyOptions.getSize() > 0 );

	addButton.addActionListener( new ActionListener() {
	    public void actionPerformed( ActionEvent e ) {
		int	selIndex = left.getSelectedIndex();
		if ( selIndex != -1 ) {
		    currentCopyOptions.addElement( copyNames[ selIndex ] );
		    right.setSelectedIndex( currentCopyOptions.getSize() - 1 );
		    removeButton.setEnabled( true );
		}
	    }
	} );
	MouseListener leftMouseListener = new MouseAdapter() {
		public void mouseClicked( MouseEvent e ) {
		    if ( e.getClickCount() == 2 ) {
			int selIndex =
			    left.locationToIndex( e.getPoint() );
			if ( selIndex != -1 ) {
			    currentCopyOptions.addElement(
				copyNames[ selIndex ] );
			    right.setSelectedIndex(
				currentCopyOptions.getSize() - 1 );
			    removeButton.setEnabled( true );
			}
		    }
		}
	    };
	left.addMouseListener( leftMouseListener );
	removeButton.addActionListener( new ActionListener() {
	    public void actionPerformed( ActionEvent e ) {
		int	index = right.getSelectedIndex();
		if ( index == -1 ) {
		    return;
		}
		currentCopyOptions.remove( index );

		int	size = currentCopyOptions.getSize();
		if ( size == 0 ) {
		    removeButton.setEnabled( false );
		} else {
		    if ( index == currentCopyOptions.getSize() ) {
			index--;
		    }
		    right.setSelectedIndex( index );
		}
	    }
	} );
	MouseListener rightMouseListener = new MouseAdapter() {
		public void mouseClicked( MouseEvent e ) {
		    if ( e.getClickCount() == 2 ) {
			int index =
			    right.locationToIndex( e.getPoint() );
			if ( index == -1 ) {
			    return;
			}
			currentCopyOptions.remove( index );

			int	size = currentCopyOptions.getSize();
			if ( size == 0 ) {
			    removeButton.setEnabled( false );
			} else {
			    if ( index == currentCopyOptions.getSize() ) {
				index--;
			    }
			    right.setSelectedIndex( index );
			}
		    }
		}
	    };
	right.addMouseListener( rightMouseListener );

	JPanel	buttonPanel = new JPanel();
	buttonPanel.setLayout( new BoxLayout( buttonPanel, BoxLayout.Y_AXIS ) );
	buttonPanel.add( addButton );
	buttonPanel.add( removeButton );

	JPanel	box = new JPanel();
	box.setLayout( new BoxLayout( box, BoxLayout.X_AXIS ) );
	box.add( leftBox     );
	box.add( buttonPanel );
	box.add( rightBox    );

	JPanel pane = new JPanel();
	pane.setLayout( new BorderLayout() );
	pane.add( box, BorderLayout.NORTH );

	return pane;
    }

    //--------------------------------------------------------------------------

    private static JPanel createNotationPanel( Resources	resources,
					       MyProperties	properties,
					       final Frame	aFrame ) {

	Box box = Box.createVerticalBox();

	MyCheckBox	button;
	Box		inbox;

	inbox = Box.createHorizontalBox();
	button = new MyCheckBox(
	    resources, properties, MyProperties.NOTATION_USE_FRENCH_C_AND_E );
	inbox.add( button );
	inbox.add( Box.createHorizontalGlue() );
	box.add( inbox );

	inbox = Box.createHorizontalBox();
	inbox.add( Box.createHorizontalStrut( 40 ) );
	button = new MyCheckBox(
	    resources, properties,
	    MyProperties.NOTATION_USE_FRENCH_C_AND_E_IF_FRENCH );
	inbox.add( button );
	inbox.add( Box.createHorizontalGlue() );
	box.add( inbox );

	inbox = Box.createHorizontalBox();
	button = new MyCheckBox( resources,
				 properties,
				 MyProperties.NOTATION_USE_SPECIAL_CHARACTERS );
	inbox.add( button );
	inbox.add( Box.createHorizontalGlue() );
	box.add( inbox );

	inbox = Box.createHorizontalBox();
	button = new MyCheckBox(
	    resources, properties, MyProperties.NOTATION_CASTLING_USE_ZEROS );
	inbox.add( button );
	inbox.add( Box.createHorizontalGlue() );
	box.add( inbox );

	inbox = Box.createHorizontalBox();
	button = new MyCheckBox(
	    resources, properties, MyProperties.NOTATION_SHOW_SINGLE_MOVES );
	inbox.add( button );
	inbox.add( Box.createHorizontalGlue() );
	box.add( inbox );

	JLabel languageLabel   = new JLabel(
	    resources.getString( "Notation_Pieces_Output_Language" ) );

	languageCombo.removeAllItems();
	StringTokenizer	t = new StringTokenizer(
	    resources.getString( "outputLanguages" ), "," );
	int	currentChoice = -1;
	int	counter       = 0;
	String	language
	    = properties.getProperty( MyProperties.OUTPUT_LANGUAGE );
	while ( t.hasMoreTokens() ) {
	    String s = t.nextToken().trim();
	    languageCombo.addItem( s );
	    if ( s.startsWith( language ) ) {
		currentChoice = counter;
	    }
	    counter++;
	}
	if ( currentChoice != -1 ) {
	    languageCombo.setSelectedIndex( currentChoice );
	}
	languageCombo.setMaximumSize( languageCombo.getMinimumSize() );

	Box	languageBox = Box.createHorizontalBox();
	languageBox.add( languageLabel );
	languageBox.add( languageCombo );
	languageBox.add( Box.createHorizontalGlue() );
	box.add( languageBox );

	JPanel pane = new JPanel();
	pane.setLayout( new BorderLayout() );
	pane.add( box, BorderLayout.NORTH );

	return pane;
    }

    //--------------------------------------------------------------------------

    private static JPanel createMiscPanel( Resources	resources,
					   MyProperties	properties,
					   final Frame	aFrame ) {

	JPanel box = new JPanel();
	box.setLayout( new BoxLayout( box, BoxLayout.Y_AXIS ) );

	MyCheckBox button;

	button = new MyCheckBox(
	    resources, properties, MyProperties.MISC_USE_SPG_HEURISTICS );
	box.add( button );

	JPanel pane = new JPanel();
	pane.setLayout( new BorderLayout() );
	pane.add( box, BorderLayout.NORTH );

	return pane;
    }

    //--------------------------------------------------------------------------

    private Preferences() { } // Avoid instanciation

    //--------------------------------------------------------------------------

    private static JOptionPane	optionPane = null ;
    private static JDialog	dialog     = null ;

    private static String[]	copyTypes ;
    private static String[]	copyNames ;
    private static DefaultListModel currentCopyOptions = new DefaultListModel();
    private static JComboBox	languageCombo = new JComboBox();

}
