/*
 * SPGMove.java --- Store a chess move which has a status.
 *
 * Copyright (c) 2002, 2003 by Pascal Wassong All Rights Reserved.
 *
 * This file is part of iNatch.
 *
 * iNatch is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * iNatch is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

package iNatch;

import java.util.Arrays;


public final class SPGMove extends Move implements Comparable {

    public static final int	MOVE_OK  = 0;
    public static final int	MOVE_OLD = 1;
    public static final int	MOVE_BAD = 2;

    public int status() {
	return this.status;
    }

    public static SPGMove[] orderMoves(
	Move[]		moves,
	int		nbMoves,
	Position	position,
	Move		ennemiMove,
	Move		previousMove ) {

	SPGMove[]	orderedMoves = new SPGMove[ nbMoves ];

	if ( ennemiMove == null || previousMove == null ) {
	    for ( int i = 0; i < nbMoves; i++ ) {
		orderedMoves[ i ] = new SPGMove( moves[ i ], MOVE_OK );
	    }
	    return orderedMoves;
	}

	Position previousPosition = (Position)position.clone();
	previousPosition.unplay( ennemiMove );
	previousPosition.unplay( previousMove );
	// null is OK, because if an ep was possible, it is no more now.
	Move[]	previousMoves = previousPosition.moves(
	    previousMove.origin().colour(), null );

	next_move:
	for ( int i = 0; i < nbMoves; i++ ) {

	    if ( moves[ i ].origin().equals(
		     previousMove.destination() )
		 && moves[ i ].destination().equals(
		     previousMove.origin() ) ) {

		orderedMoves[ i ] = new SPGMove( moves[ i ], MOVE_BAD );
		continue;
	    }

	    Position p = (Position)previousPosition.clone();

	    if ( !previousMove.isNullMove()
		 && previousMove.destination().equals( moves[ i ].origin() ) ) {

		Move m =
		    new Move( previousMove.origin(), moves[ i ].destination() );

		for ( int j = 0; j < previousMoves.length; j++ ) {
		    if ( previousMoves[ j ].equals( m ) ) {
			orderedMoves[ i ] = new SPGMove( moves[ i ], MOVE_BAD );
			continue next_move;
		    }
		}
	    }

	    try {
		boolean	ok = false;
		for ( int j = 0; j < previousMoves.length; j++ ) {
		    if ( previousMoves[ j ].equals( moves[ i ] ) ) {
			ok = true;
			break;
		    }
		}
		if ( !ok ) {
		    orderedMoves[ i ] = new SPGMove( moves[ i ], MOVE_OK );
		    continue;
		}
		p.play( moves[ i ] );

		Move[]	ennemiMoves =  p.moves(
		    ennemiMove.origin().colour(), moves[ i ] );

		ok = false;
		for ( int j = 0; j < ennemiMoves.length; j++ ) {
		    if ( ennemiMoves[ j ].equals( ennemiMove ) ) {
			ok = true;
			break;
		    }
		}
		if ( !ok ) {
		    orderedMoves[ i ] = new SPGMove( moves[ i ], MOVE_OK );
		    continue;
		}

		p.play( ennemiMove );

		Move[]	myMoves =  p.moves(
		    previousMove.origin().colour(), ennemiMove );

		ok = false;
		for ( int j = 0; j < myMoves.length; j++ ) {
		    if ( myMoves[ j ].equals( previousMove ) ) {
			ok = true;
			break;
		    }
		}
		if ( !ok ) {
		    orderedMoves[ i ] = new SPGMove( moves[ i ], MOVE_OK );
		    continue;
		}

		orderedMoves[ i ] = new SPGMove( moves[ i ], MOVE_OLD );
	    } catch ( IllegalArgumentException iae ) {
		// This should not happen, but if it happens, this means that
		// p.play has thrown the exception and the answer is :
		orderedMoves[ i ] = new SPGMove( moves[ i ], MOVE_OK );
	    }
	}

	Arrays.sort( orderedMoves );

	return orderedMoves;
    }

    //--------------------------------------------------------------------------

    public int compareTo( Object o ) {
	SPGMove	m = (SPGMove)o;

	if ( status < m.status() ) {
	    return -1;
	} else if ( status == m.status() ) {
	    return 0;
	} else {
	    return 1;
	}
    }

    //--------------------------------------------------------------------------

    private SPGMove( Move m, int aStatus ) {
	super( m );
	status = aStatus;
    }

    private int status;
}
