/*
 * ChessGameTest.java --- JUnit tests for ChessGame class.
 *
 * Copyright (c) 2001, 2002, 2003 by Pascal Wassong All Rights Reserved.
 *
 * This file is part of iNatch.
 *
 * iNatch is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * iNatch is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

package iNatch.tests;

import java.util.ArrayList;
import java.util.List;

import iNatch.ChessGame;
import iNatch.EinsteinCondition;
import iNatch.Move;
import iNatch.OrthodoxCondition;
import iNatch.Piece;
import iNatch.Position;
import iNatch.StandardBoard;
import iNatch.VerticalCylinderBoard;
import junit.framework.Test;
import junit.framework.TestCase;
import junit.framework.TestSuite;


public class ChessGameTest extends TestCase {

    public ChessGameTest( String name ) {
	super( name );
    }

    public static Test suite() {
	return new TestSuite( ChessGameTest.class );
    }

    public static void main( String[] args ) {
	junit.textui.TestRunner.run ( suite() );
    }

    //-------------------------------------------------------------------------

    public void testConstructors() {
	Position  p = g.position();
	assertEquals( "rnbqkbnr/pppppppp/8/8/8/8/PPPPPPPP/RNBQKBNR",
		      p.toString() );
	assertEquals( new StandardBoard().name(), p.board().name() );
	assertEquals( new OrthodoxCondition().name(), p.conditionName() );

	assertEquals( "rnbqkbnr/pppppppp/8/8/8/8/PPPPPPPP/RNBQKBNR",
		      g.finalPosition().toString() );
    }

    public void testSelectors() {
	assertEquals( 0, g.totalNumberOfMoves() );
	assertEquals( 0, g.currentNumberOfMoves() );
	assertEquals( 0, g.finalNumberOfMoves() );
	assertTrue(   g.comments().isEmpty() );
	assertNull(   g.file() );
	assertNull(   g.filename() );
	assertTrue(   !g.hasChanged() );
    }

    // Changing the game functions

    public void testPlay() {
	g.play( new Move( new Piece( Piece.PAWN, Piece.WHITE, "e2" ),
			  new Piece( Piece.PAWN, Piece.WHITE, "e4" ) ) );
	assertEquals( "rnbqkbnr/pppppppp/8/8/4P3/8/PPPP1PPP/RNBQKBNR",
		      g.position().toString() );
	assertEquals( 1, g.totalNumberOfMoves() );
	assertEquals( 1, g.currentNumberOfMoves() );
	assertTrue(   g.hasChanged() );
    }

    public void testKillEndOfGame() {
	g.play( new Move( new Piece( Piece.PAWN, Piece.WHITE, "e2" ),
			  new Piece( Piece.PAWN, Piece.WHITE, "e4" ) ) );
	g.backwardMove();
	g.killEndOfGame();
	assertEquals( "rnbqkbnr/pppppppp/8/8/8/8/PPPPPPPP/RNBQKBNR",
		      g.position().toString() );
	assertEquals( 0, g.totalNumberOfMoves() );
	assertEquals( 0, g.currentNumberOfMoves() );
	assertTrue(   g.hasChanged() );
    }

    public void testUseFinalPositionAsInitialPosition() {
	assertEquals( "rnbqkbnr/pppppppp/8/8/8/8/PPPPPPPP/RNBQKBNR",
		      g.position().toString() );

	g.clearFinalBoard();
	g.useFinalPositionAsInitialPosition();
	assertEquals( "4k3/8/8/8/8/8/8/4K3", g.finalPosition().toString() );

	g.setInitialPositionInFinalBoard();
	g.useFinalPositionAsInitialPosition();
	assertEquals( "rnbqkbnr/pppppppp/8/8/8/8/PPPPPPPP/RNBQKBNR",
		      g.position().toString() );
    }

    public void testSetBoard() {
	assertEquals( new StandardBoard().name(), g.position().board().name() );
	g.setBoard( new VerticalCylinderBoard() );
	assertEquals( new VerticalCylinderBoard().name(),
		      g.position().board().name() );
    }

    public void testSetCondition() {
	assertEquals( new OrthodoxCondition().name(),
		      g.position().conditionName() );
	g.setCondition( new EinsteinCondition() );
	assertEquals( new EinsteinCondition().name(),
		      g.position().conditionName() );
    }

    // Changing the final position

    public void testClearFinalBoard() {
	assertEquals( "rnbqkbnr/pppppppp/8/8/8/8/PPPPPPPP/RNBQKBNR",
		      g.finalPosition().toString() );
	g.clearFinalBoard();
	assertEquals( "4k3/8/8/8/8/8/8/4K3", g.finalPosition().toString() );
    }

    public void testSetInitialPositionInFinalBoard() {
	g.clearFinalBoard();
	assertEquals( "4k3/8/8/8/8/8/8/4K3", g.finalPosition().toString() );
	g.setInitialPositionInFinalBoard();
	assertEquals( "rnbqkbnr/pppppppp/8/8/8/8/PPPPPPPP/RNBQKBNR",
		      g.finalPosition().toString() );
    }

    public void testClearSquare() {
	assertEquals( "rnbqkbnr/pppppppp/8/8/8/8/PPPPPPPP/RNBQKBNR",
		      g.finalPosition().toString() );
	g.clearSquare( 0, 0 );
	assertEquals( "rnbqkbnr/pppppppp/8/8/8/8/PPPPPPPP/1NBQKBNR",
		      g.finalPosition().toString() );
    }

    public void testAddPieceAt() {
	assertEquals( "rnbqkbnr/pppppppp/8/8/8/8/PPPPPPPP/RNBQKBNR",
		      g.finalPosition().toString() );
	g.addPieceAt( new Piece( Piece.KNIGHT, Piece.BLACK, "a3" ) );
	assertEquals( "rnbqkbnr/pppppppp/8/8/8/n7/PPPPPPPP/RNBQKBNR",
		      g.finalPosition().toString() );
    }

    public void testSetFinalNumberOfMoves() {
	assertEquals( 0, g.finalNumberOfMoves() );
	g.setFinalNumberOfMoves( 45 );
	assertEquals( 45, g.finalNumberOfMoves() );
    }

    public void testCopyPositionOfGameToFinalPosition() {
	g.play( new Move( new Piece( Piece.PAWN, Piece.WHITE, "e2" ),
			  new Piece( Piece.PAWN, Piece.WHITE, "e4" ) ) );

	assertEquals( "rnbqkbnr/pppppppp/8/8/8/8/PPPPPPPP/RNBQKBNR",
		      g.finalPosition().toString() );
	g.copyPositionOfGameToFinalPosition();
	assertEquals( "rnbqkbnr/pppppppp/8/8/4P3/8/PPPP1PPP/RNBQKBNR",
		      g.finalPosition().toString() );
    }

    public void testMovingInTheGame() {
	g.play( new Move( new Piece( Piece.PAWN,   Piece.WHITE, "e2" ),
			  new Piece( Piece.PAWN,   Piece.WHITE, "e4" ) ) );
	g.play( new Move( new Piece( Piece.PAWN,   Piece.BLACK, "e7" ),
			  new Piece( Piece.PAWN,   Piece.BLACK, "e5" ) ) );
	g.play( new Move( new Piece( Piece.BISHOP, Piece.WHITE, "f1" ),
			  new Piece( Piece.BISHOP, Piece.WHITE, "b5" ) ) );

	assertEquals( "rnbqkbnr/pppp1ppp/8/1B2p3/4P3/8/PPPP1PPP/RNBQK1NR",
		      g.position().toString() );
	g.backwardMove();
	g.backwardMove();
	assertEquals( "rnbqkbnr/pppppppp/8/8/4P3/8/PPPP1PPP/RNBQKBNR",
		      g.position().toString() );

	g.forwardMove();
	assertEquals( "rnbqkbnr/pppp1ppp/8/4p3/4P3/8/PPPP1PPP/RNBQKBNR",
		      g.position().toString() );

	g.beginningOfGame();
	assertEquals( "rnbqkbnr/pppppppp/8/8/8/8/PPPPPPPP/RNBQKBNR",
		      g.position().toString() );

	g.endOfGame();
	assertEquals( "rnbqkbnr/pppp1ppp/8/1B2p3/4P3/8/PPPP1PPP/RNBQK1NR",
		      g.position().toString() );

	g.gotoMove( 1 );
	assertEquals( "rnbqkbnr/pppppppp/8/8/4P3/8/PPPP1PPP/RNBQKBNR",
		      g.position().toString() );
    }

    public void testComments() {
	List l = new ArrayList( 3 );
	l.add( "1" );
	l.add( "2" );
	l.add( "3" );
	g.setComments( l );

	List l2 = g.comments();
	assertEquals( l.size(), l2.size() );

	for ( int i = 0; i < 3; i++ ) {
	    assertEquals( "Error at list index " + i, l.get( i ), l2.get( i ) );
	}
    }

    public void testFormatOutput() {
	g.play( new Move( new Piece( Piece.PAWN,   Piece.WHITE, "e2" ),
			  new Piece( Piece.PAWN,   Piece.WHITE, "e4" ) ) );
	g.play( new Move( new Piece( Piece.PAWN,   Piece.BLACK, "e7" ),
			  new Piece( Piece.PAWN,   Piece.BLACK, "e5" ) ) );
	g.play( new Move( new Piece( Piece.BISHOP, Piece.WHITE, "f1" ),
			  new Piece( Piece.BISHOP, Piece.WHITE, "b5" ) ) );

	assertEquals(
	    "rnbqkbnr/pppp1ppp/8/1B2p3/4P3/8/PPPP1PPP/RNBQK1NR\n\n"
	    + "1.e4 e5 2.Bb5\n",
	    g.formatOutput( "board-forsythe;empty-line;moves-short" ) );
    }

    public void testToString() {
	g.play( new Move( new Piece( Piece.PAWN,   Piece.WHITE, "e2" ),
			  new Piece( Piece.PAWN,   Piece.WHITE, "e4" ) ) );
	g.play( new Move( new Piece( Piece.PAWN,   Piece.BLACK, "e7" ),
			  new Piece( Piece.PAWN,   Piece.BLACK, "e5" ) ) );
	g.play( new Move( new Piece( Piece.BISHOP, Piece.WHITE, "f1" ),
			  new Piece( Piece.BISHOP, Piece.WHITE, "b5" ) ) );

	assertEquals( "1.e4 e5 2.Bb5",          g.shortMoves()   );
	assertEquals( "1.e2-e4 e7-e5 2.Bf1-b5", g.longMoves()    );
	assertEquals( " e2-e4",                 g.longMove( 0 )  );
	assertEquals( " e7-e5",                 g.longMove( 1 )  );
	assertEquals( "Bf1-b5",                 g.longMove( 2 )  );
	assertEquals( "1.Bf1-b5 e7-e5 2.e2-e4", g.reverseMoves() );
    }

    //-------------------------------------------------------------------------

    protected void setUp() {
	g = new ChessGame( null, null );
    }

    protected void tearDown() {
	g = null;
    }

    protected ChessGame g;
}
