/*
 * PositionTest.java --- JUnit tests for the Position class.
 *
 * Copyright (c) 2001, 2002, 2003 by Pascal Wassong All Rights Reserved.
 *
 * This file is part of iNatch.
 *
 * iNatch is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * iNatch is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

package iNatch.tests;

import iNatch.Move;
import iNatch.OrthodoxCondition;
import iNatch.Piece;
import iNatch.Position;
import iNatch.PositionFactory;
import iNatch.StandardBoard;
import junit.framework.Test;
import junit.framework.TestCase;
import junit.framework.TestSuite;


public class PositionTest extends TestCase {

    public PositionTest( String name ) {
	super( name );
    }

    public static Test suite() {
	return new TestSuite( PositionTest.class );
    }

    public static void main( String[] args ) {
	junit.textui.TestRunner.run ( suite() );
    }

    //-------------------------------------------------------------------------

    public void testConstructor() {
	Position p = new Position( new StandardBoard(),
				   new OrthodoxCondition(),
				   true );
	assertEquals( initialPosition.toString(), p.toString() );
    }

    public void testStringConstructor() {
	// Minimal input should be correct
	PositionFactory.make( new StandardBoard(),
			      new OrthodoxCondition(),
			      "k1K" );
	if ( PositionFactory.status() != PositionFactory.CORRECT_POSITION ) {
	    fail( "Minimal input not accepted" );
	    return;
	}

	// Known foreign languages are recognised
	PositionFactory.setForsytheInitials( "kpqrbns rpdtfcc" );
	PositionFactory.make( new StandardBoard(),
			      new OrthodoxCondition(),
			      "tcfdrfct/pppppppp/8/8/8/8/PPPPPPPP/TCFDRFCT" );
	if ( PositionFactory.status() != PositionFactory.CORRECT_POSITION ) {
	    fail( "French input not accepted" );
	    return;
	}
	PositionFactory.setForsytheInitials( null ); // Reset to default.

	// Only one king of each side
	PositionFactory.make( new StandardBoard(),
			      new OrthodoxCondition(),
			      "k1k1K3/8/8/8/8/8/8/8" );
	if ( PositionFactory.status() == PositionFactory.CORRECT_POSITION ) {
	    fail( "Many kings accepted" );
	    return;
	}
    }

    public void testClone() {
	Position p = (Position)initialPosition.clone();
	assertEquals( initialPosition.toString(), p.toString() );
    }

    public void testSelectors() {
	assertTrue( ( initialPosition.chessboard()[ 0 ][ 0 ]
		      != Position.EMPTY_SQUARE )
		    && initialPosition.board().name() == "standard"
		    && initialPosition.piece( 31 ) != null
		    && initialPosition.numberOfWhitePieces() == 16
		    && initialPosition.numberOfBlackPieces() == 16
		    && ( initialPosition.conditionName()
			 == "orthodox-condition" ) );
    }

    public void testPlay() {
	Position p = (Position)initialPosition.clone();
	p.play( new Move( new Piece( Piece.PAWN,   Piece.WHITE, "e2" ),
			  new Piece( Piece.PAWN,   Piece.WHITE, "e4" ) ) );
	p.play( new Move( new Piece( Piece.KNIGHT, Piece.BLACK, "b8" ),
			  new Piece( Piece.KNIGHT, Piece.BLACK, "a6" ) ) );
	p.play( new Move( new Piece( Piece.BISHOP, Piece.WHITE, "f1" ),
			  new Piece( Piece.BISHOP, Piece.WHITE, "a6" ),
			  new Piece( Piece.KNIGHT, Piece.BLACK, "a6" ) ) );
	p.play( new Move( new Piece( Piece.PAWN,   Piece.BLACK, "b7" ),
			  new Piece( Piece.PAWN,   Piece.BLACK, "a6" ),
			  new Piece( Piece.BISHOP, Piece.WHITE, "a6" ) ) );
	p.play( new Move( new Piece( Piece.PAWN,   Piece.WHITE, "e4" ),
			  new Piece( Piece.PAWN,   Piece.WHITE, "e5" ) ) );
	p.play( new Move( new Piece( Piece.PAWN,   Piece.BLACK, "d7" ),
			  new Piece( Piece.PAWN,   Piece.BLACK, "d5" ) ) );
	p.play( new Move( new Piece( Piece.PAWN,   Piece.WHITE, "e5" ),
			  new Piece( Piece.PAWN,   Piece.WHITE, "d6" ),
			  new Piece( Piece.PAWN,   Piece.BLACK, "d5" ) ) );
	p.play( new Move( new Piece( Piece.QUEEN,  Piece.BLACK, "d8" ),
			  new Piece( Piece.QUEEN,  Piece.BLACK, "d7" ) ) );
	p.play( new Move( new Piece( Piece.KNIGHT, Piece.WHITE, "g1" ),
			  new Piece( Piece.KNIGHT, Piece.WHITE, "e2" ) ) );
	p.play( new Move( new Piece( Piece.KING,   Piece.BLACK, "e8" ),
			  new Piece( Piece.KING,   Piece.BLACK, "d8" ) ) );
	p.play( new Move( Move.KING_CASTLING, Piece.WHITE, 7, Piece.ROOK ) );

	assertEquals( "r1bk1bnr/p1pqpppp/p2P4/8/8/8/PPPPNPPP/RNBQ1RK1",
		      p.toString() );
    }

    public void testUnplay() {
	PositionFactory.make(
	    new StandardBoard(),
	    new OrthodoxCondition(),
	    "r1bk1bnr/p1pqpppp/p2P4/8/8/8/PPPPNPPP/RNBQ1RK1" );
	if ( PositionFactory.status() != PositionFactory.CORRECT_POSITION ) {
	    fail( "Illegal position" );
	    return;
	}

	Position p = PositionFactory.position();

	p.unplay( new Move( Move.KING_CASTLING, Piece.WHITE, 7, Piece.ROOK ) );
	p.unplay( new Move( new Piece( Piece.KING,   Piece.BLACK, "e8" ),
			    new Piece( Piece.KING,   Piece.BLACK, "d8" ) ) );
	p.unplay( new Move( new Piece( Piece.KNIGHT, Piece.WHITE, "g1" ),
			    new Piece( Piece.KNIGHT, Piece.WHITE, "e2" ) ) );
	p.unplay( new Move( new Piece( Piece.QUEEN,  Piece.BLACK, "d8" ),
			    new Piece( Piece.QUEEN,  Piece.BLACK, "d7" ) ) );
	p.unplay( new Move( new Piece( Piece.PAWN,   Piece.WHITE, "e5" ),
			    new Piece( Piece.PAWN,   Piece.WHITE, "d6" ),
			    new Piece( Piece.PAWN,   Piece.BLACK, "d5" ) ) );
	p.unplay( new Move( new Piece( Piece.PAWN,   Piece.BLACK, "d7" ),
			    new Piece( Piece.PAWN,   Piece.BLACK, "d5" ) ) );
	p.unplay( new Move( new Piece( Piece.PAWN,   Piece.WHITE, "e4" ),
			    new Piece( Piece.PAWN,   Piece.WHITE, "e5" ) ) );
	p.unplay( new Move( new Piece( Piece.PAWN,   Piece.BLACK, "b7" ),
			    new Piece( Piece.PAWN,   Piece.BLACK, "a6" ),
			    new Piece( Piece.BISHOP, Piece.WHITE, "a6" ) ) );
	p.unplay( new Move( new Piece( Piece.BISHOP, Piece.WHITE, "f1" ),
			    new Piece( Piece.BISHOP, Piece.WHITE, "a6" ),
			    new Piece( Piece.KNIGHT, Piece.BLACK, "a6" ) ) );
	p.unplay( new Move( new Piece( Piece.KNIGHT, Piece.BLACK, "b8" ),
			    new Piece( Piece.KNIGHT, Piece.BLACK, "a6" ) ) );
	p.unplay( new Move( new Piece( Piece.PAWN,   Piece.WHITE, "e2" ),
			    new Piece( Piece.PAWN,   Piece.WHITE, "e4" ) ) );

	assertEquals( initialPosition.toString(), p.toString() );
    }

    public void testClearSquare() {
	Position p = (Position)initialPosition.clone();
	p.clearSquare( 0, 0 );
	assertEquals( "rnbqkbnr/pppppppp/8/8/8/8/PPPPPPPP/1NBQKBNR",
		      p.toString() );
    }

    public void testAddPieceAt() {
	Position p = (Position)initialPosition.clone();
	p.addPieceAt( new Piece( Piece.PAWN, Piece.WHITE, "a3" ) );
	assertEquals( "rnbqkbnr/pppppppp/8/8/8/P7/PPPPPPPP/RNBQKBNR",
		      p.toString() );
    }

    public void testListOfMoves() {
	assertTrue( initialPosition.moves( Piece.WHITE, null ).length == 20 );
    }

    public void testToString() {
	assertEquals( "rnbqkbnr/pppppppp/8/8/8/8/PPPPPPPP/RNBQKBNR",
		      initialPosition.toString() );
    }

    public void testToDiagram() {
	assertEquals( "+---+---+---+---+---+---+---+---+\n"
		      + "|*R |*N |*B |*Q |*K |*B |*N |*R |\n"
		      + "+---+---+---+---+---+---+---+---+\n"
		      + "|*P |*P |*P |*P |*P |*P |*P |*P |\n"
		      + "+---+---+---+---+---+---+---+---+\n"
		      + "|   | . |   | . |   | . |   | . |\n"
		      + "+---+---+---+---+---+---+---+---+\n"
		      + "| . |   | . |   | . |   | . |   |\n"
		      + "+---+---+---+---+---+---+---+---+\n"
		      + "|   | . |   | . |   | . |   | . |\n"
		      + "+---+---+---+---+---+---+---+---+\n"
		      + "| . |   | . |   | . |   | . |   |\n"
		      + "+---+---+---+---+---+---+---+---+\n"
		      + "| P | P | P | P | P | P | P | P |\n"
		      + "+---+---+---+---+---+---+---+---+\n"
		      + "| R | N | B | Q | K | B | N | R |\n"
		      + "+---+---+---+---+---+---+---+---+\n",
		      initialPosition.diagram() );
    }

    public void testToFont1Echecs() {
	assertEquals(
	    "!--------!\n"
	    + "/T89CF23DR67FC45T/\n"
	    + "/()PP()PP()PP()PP/\n"
	    + "/ : : : :/\n"
	    + "/: : : : /\n"
	    + "/ : : : :/\n"
	    + "/: : : : /\n"
	    + "/p()pp()pp()pp()p/\n"
	    + "/45tc67fde01rf89ct/\n"
	    + "$________$\n",
	    initialPosition.font1Echecs() );
    }

    //-------------------------------------------------------------------------

    protected void setUp() {
	initialPosition = new Position( new StandardBoard(),
					new OrthodoxCondition(),
					true );
    }

    protected void tearDown() {
	initialPosition = null;
    }

    protected Position	initialPosition;
}
